#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-15-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np

class EpiAbsValue:
    """
    This class computes the projection onto the epigrapgh of:
        
                phi(y)= w * |y|
        
    When the inputs are arrays, the outputs are computed element-wise
     INPUTS
    ========
     y    - ND array
     ksi  - ND array with the same size as 'y'
     w    - positive, scalar or ND array with the same size as 'ksi'
     """
    
    def __init__(self, ksi, w=1):

        if np.any( w <= 0 ) :
            raise Exception("'w' (or all of its components if it is an array) must be positive")   
        if np.size(w)>1 and ( np.size(w) != np.size(ksi)):
           raise Exception(" 'w' must be a scalar or must have the same size as 'ksi'")
        if np.size(ksi)<=1:
            ksi=np.array([ksi])
        self.w=w
        self.ksi=ksi
        
    # proximal operator (i.e the projetion on the constraint set)   
    def prox(self, y):
        self._check(y)
        ksi=self.ksi
        w=self.w
        if np.size(y)<=1:
            y=np.array([y])
            
        # 2nd branch
        y_abs = np.abs(y)
        c= np.maximum(0, y_abs + w*ksi)/(1+w**2)
        p=np.sign(y)*c
        t=w*c
        
        # 1st branch
        mask = w*y_abs < ksi
        p[mask] =  y[mask]
        t[mask] = ksi[mask]
        return np.array([p,t])
    
    # indicator of the constraint set
    def __call__(self, y):
        if np.all( self.w*np.abs(y) <= self.ksi):
            return 0
        return np.inf

    def _check(self, y):
       if np.size(y) != np.size( self.ksi):
           raise Exception(" 'y' must have the same size as 'ksi'")
       
       
    




