#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-16-2022)
 
    DEPENDENCIES:
     - 'newton.py' located in the folder 'utils'
     
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np
from prox.utils.newton import newton_


class EpiDistance:
    """
    This class computes the projection onto the epigrapgh of:
        
            phi(y) = w * ||y - P_C(y)||_2^q       with q=1 or q=2
    
     When the input 'y' is an array, the computation can vary as follows:
    
      - dir = None --> 'y' is processed as a single vector [DEFAULT]
                    (in this case, 'xi' must be scalar)
      - dir >= 0 --> 'y' is processed block-wise along the specified direction
                    (in this case, 'xi' must be singleton along 'dir')
      - dir < 0 --> 'y' is processed element-by-element.
                    (in this case, 'xi' must be the same size as 'y')
    
     The caller must ensure that 'project_C' uses the same type of processing.
    
      INPUTS
     ========
      y         - ND array
      ksi        - ND array compatible with the blocks of 'y'
      w         - positive, scalar or ND array with the same size as 'ksi'
      dir       - integer, direction of block-wise processing [OPTIONAL]
      q         - integer, scalar [OPTIONAL, default: 2]
      project_C - function handle with an argument at least [OPTIONAL, default: 0]
      varargin  - additional parameters for the function 'prox_phi' [OPTIONAL]

     """
    
    def __init__(self, ksi, direction=None, project_C=lambda y: 0.0, q=2, w=1 ):

        if np.any( w <= 0 ) :
            raise Exception("'w' (or all of its components if it is an array) must be positive")   
        if np.size(w)>1 and ( np.size(w) != np.size(ksi)):
           raise Exception(" 'w' must be a scalar or must have the same size as 'ksi'")
        
        if np.size(ksi)<=1:
            ksi=np.array([ksi])
        
        self.w=w
        self.ksi=ksi
        self.direction=direction
        self.q=q
        self.project_C=project_C
        
    # proximal operator (i.e the projetion on the constraint set)   
    def prox(self, y):
        self._check(y)
        ksi=self.ksi
        w=self.w
        direction=self.direction
        project_C=self.project_C
        q=self.q
        eps=1e-16
        sz=np.shape(y)
        
        if np.size(y)<=1:
            y=np.array([y])
        if self.direction==None:
            y=np.reshape(y, (-1))
            
        # preliminaries
        py = project_C(y)
        dy = np.linalg.norm(y-py, ord=2, axis=direction) # distance separating y from C
        
        if q==1:
            t=np.maximum(0, dy+w*ksi) / (1+w**2)
        else:
            # Finding the root of with the Newton's method
            def polynom_phi(t):
                return q*w**2*t**(2*q-1) - q*w*ksi*t**(q-1) +t- np.abs(y)

            def derive_polynom_phi(t):
                return (2*q-1)*q*w**2*t**(2*q-2) - (q-1)*q*w*ksi*t**(q-2) + 1
    
            # starting point
            low = (np.maximum( ksi, 0)/w)**(1/q)
            root_init= 3*low
            
            t = newton_(polynom_phi, fp=derive_polynom_phi, x0=root_init, low=low, high= np.inf)
    
        # 3rd branch
        p = py + t/(dy+eps)*(y-py) 
        
        t = w *np.linalg.norm(p-project_C(p), ord=2, axis=direction)**q
        
        
        # default cas: direction==None => y is processed as single vector
        if direction==None :
            if dy==0 and ksi<0:
                return [y, np.zeros_like(ksi)]
            elif w*dy**q <= ksi:
                return [y, ksi]
            return [p, t]
        
        
        # 2nd branch
        mask_p=np.ones(np.shape(y), dtype=bool)
        mask_t= np.logical_and(dy == 0 , ksi< 0)
        mask_p= mask_t*mask_p
        p[mask_p] = y[mask_p]
        t[mask_t] = 0
        
        # 1st branch
        
        mask_t = w * dy**q <= ksi
        mask_t= np.logical_and(dy == 0 , ksi< 0)
        mask_p= mask_t*mask_p
        p[mask_p] =  y[mask_p]
        t[mask_t] = ksi[mask_t]
        
        p=np.reshape(p, sz)
        return [p,t]
    
    # indicator of the constraint set
    def __call__(self, y):
        self._check(y)
        TOL=1e-10
        if self.direction==None:
            y=np.reshape(y, (-1))
        py = self.project_C(y)
        dy = np.linalg.norm(y-py, ord=2, axis=self.direction)

        if np.all( self.w*dy**self.q <= self.ksi+TOL):
            return 0
        return np.inf

    def _check(self, y):
        sz_=1
        if self.direction !=None:
           sz= np. array( np.shape(y) , dtype=int)
           sz_= np.delete(sz, self.direction)
        
        if (self.direction !=None and  np.any( sz_!= np.shape(self.ksi) ) ) or ( self.direction==None and np.any(np.size(self.ksi) != sz_)):
           try:
               self.ksi= np.reshape( self.ksi, sz_)
           except:
               raise Exception("'ksi' is not compatible with the size of 'y' ")
           
       
    





