#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-16-2022)
 
    DEPENDENCIES:
     - 'newton.py' in the folder 'utils'
     
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np
from prox.utils.newton import newton_


class EpiExp:
    """
    This class computes the projection onto the epigrapgh of:
        
                phi(y)= w * max( y, 0)
        
    When the inputs are arrays, the outputs are computed element-wise
     INPUTS
    ========
     y    - ND array
     kxi  - ND array with the same size as 'y'
     w    - positive, scalar or ND array with the same size as x
     """
    
    def __init__(self, ksi, w=1):

        if np.any( w <= 0 ) :
            raise Exception("'w' (or all of its components if it is an array) must be positive")   
        if np.size(w)>1 and ( np.size(w) != np.size(ksi)):
           raise Exception(" 'w' must be a scalar or must have the same size as 'ksi'")
        if np.size(ksi)<=1:
            ksi=np.array([ksi])
        self.w=w
        self.ksi=ksi
        
    # proximal operator (i.e the projetion on the constraint set)   
    def prox(self, y):
        self._check(y)
        ksi=self.ksi
        w=self.w
        
        if np.size(y)<=1:
            y=np.array([y])
        # 1st branch
        p=y[:]
        t=np.maximum( ksi, w*np.exp(p))
        
        eps=1e-16
        mask= np.logical_and( w*np.exp(y)> ksi, y>np.log(eps)+1)
        yy=y[mask]
        xx=ksi[mask]
        ww=w
        if np.size(ww)>1:
            ww=w[mask]
            
        # Newton's method
        def polynom_phi(t):
            return t**2 -xx*t +np.log(t)*yy

        def derive_polynom_phi(t):
            return 2*t-xx +1/(t+eps)
        
        # starting point
        low = np.maximum(eps, xx)
        root_init= np.abs(xx)
        
        # Finding the root of the polynom with the Newton method
        root = newton_(polynom_phi, fp=derive_polynom_phi, x0=root_init, low=low, high= np.inf, max_iter=1000)
        
        # 2nd branch
        p[mask]=np.log(root/ww)
        t[mask]=root
 
        return np.array([p,t])
    
    # indicator of the constraint set
    def __call__(self, y):
        if np.all( self.w*np.maximum(y, 0) <= self.ksi):
            return 0
        return np.inf

    def _check(self, y):
       if np.size(y) != np.size( self.ksi):
           raise Exception(" 'y' must have the same size as 'ksi'")
       
       
    





