#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-16-2022)

 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np

class EpiL2Norm:
    """
    This class computes the projection onto the epigrapgh of:
        
                phi(y)= w * ||y||_2 
        
     When the input 'y' is an array, the computation can vary as follows:
    
      - dir = None --> 'y' is processed as a single vector [DEFAULT]
                           (in this case, 'xi' must be scalar)
      - dir >= 0 --> 'y' is processed block-wise along the specified direction
                    (in this case, 'xi' must be singleton along 'dir')
                    
     INPUTS
    ========
     y    - ND array
     ksi  - ND array compatible with the blocks of 'y' (see above)
     w    - positive, scalar or ND array with the same size as ksi
     dir  - integer, direction of block-wise processing
     """
    
    def __init__(self, ksi, direction=None, w=1):

        if np.any( w <= 0 ) :
            raise Exception("'w' (or all of its components if it is an array) must be positive")   
        if np.size(w)>1 and ( np.size(w) != np.size(ksi)):
           raise Exception(" 'w' must be a scalar or must have the same size as 'ksi'")
        if np.size(ksi)<=1:
            ksi=np.array([ksi])
        
        self.w=w
        self.direction=direction
        self.ksi=ksi
        
    # proximal operator (i.e the projetion on the constraint set)   
    def prox(self, y):
        self._check(y)
        ksi=self.ksi
        w=self.w
        direction=self.direction 
        eps=1e-16
        sz=np.shape(y)
        
        if self.direction==None:
            y=np.reshape(y, (-1))
        
        yy=np.linalg.norm( y, ord=2, axis=direction)
        # branch 3
        a= np.maximum( 0, 1+w*ksi/(yy+eps))/( 1+w**2) 
        # branch 1
        a[w*yy<=ksi]=1                  
        # branch 2            
        a[ np.logical_and(yy==0, ksi<0)] = 0
        
        # compute the projection
        
        p=a*y
        t=np.maximum(ksi, w*a*yy)
        
        #reshape
        p=np.reshape(p, sz)
        return [p, t]
    
    # indicator of the constraint set
    def __call__(self, y):
        self._check(y)
        TOL=1e-10
        if self.direction==None:
            y=np.reshape(y, (-1))
        mask= self.w*np.linalg.norm(y, ord=2, axis=self.direction) <= self.ksi+TOL
        if np.all( mask ) :
            return 0
        return np.inf

    def _check(self, y):
        sz_=1
        if self.direction !=None:
           sz= np. array( np.shape(y) , dtype=int)
           sz_= np.delete(sz, self.direction)

        if (self.direction !=None and  np.any( sz_!= np.shape(self.ksi) ) ) or ( self.direction==None and np.any(np.size(self.ksi) != sz_)):
           try:
               self.ksi= np.reshape( self.ksi, sz_)
           except:
              raise Exception("'ksi' is not compatible with the size of 'y' ")
       
       





