#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-16-2022)

 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np

class EpiMax:
    """
    This class computes the projection onto the epigrapgh of:
        
                phi(y)= w * max( x1, ..., x_n)
        
     When the input 'y' is an array, the computation can vary as follows:
    
      - dir = None --> 'y' is processed as a single vector [DEFAULT]
                           (in this case, 'xi' must be scalar)
      - dir >= 0 --> 'y' is processed block-wise along the specified direction
                    (in this case, 'xi' must be singleton along 'dir')
                    
     INPUTS
    ========
     y    - ND array
     ksi  - ND array compatible with the blocks of 'y' (see above)
     w    - positive, scalar or ND array with the same size as 'ksi'
     dir  - integer, direction of block-wise processing
     """
    
    def __init__(self, ksi, direction=None, w=1):

        if np.any( w <= 0 ) :
            raise Exception("'w' (or all of its components if it is an array) must be positive")   
        if np.size(w)>1 and ( np.size(w) != np.size(ksi)):
           raise Exception(" 'w' must be a scalar or must have the same size as 'ksi'")
        if np.size(ksi)<=1:
            ksi=np.array([ksi])
    
        self.w=w
        self.direction=direction
        self.ksi=ksi
        
    # proximal operator (i.e the projetion on the constraint set)   
    def prox(self, y):
        self._check(y)
        ksi=self.ksi
        w=self.w
        direction=self.direction 
        if len(np.shape(y))==1 and np.size(ksi)<=1:
            direction=None
        
        if direction != None:
            if direction==0:
                ksi=np.reshape( ksi, (2*direction+1)*(1,-1))
            else:
                ksi=np.reshape( ksi, (-1,1))
        # sort inputs
        v= y*w
        v= np.sort(v, axis=direction)
        
        # compute the cumulative sum
        s=np.concatenate( (w**2*ksi, np.flip(v, axis=direction) ), axis=direction)
        s=np.cumsum(s, axis=direction)
        s=np.flip(s, axis=direction)
        
        dir_=direction
        if direction==None:
            dir_=0
        # Normalize...
        s=s/ ( w**2 + np.shape(v)[dir_]+1 - np.cumsum( np.ones_like(s), axis=direction))
        
        infty = np.inf*np.ones(np.shape(ksi))
        v_low= np.concatenate( (-infty, v), axis=direction)
        v_high=np.concatenate( (v, infty), axis=direction)
        
        mask = np.logical_and( v_low < s, s <= v_high)

        sz_=np.shape(mask)[dir_]
        mat=(np.arange( np.size(mask))%sz_).reshape( np.shape(mask))
        mask= mask*mat
       
        n=np.max(mask, axis=direction)

        if direction==None:
            t= np.reshape( s[n], np.shape(ksi))
        else:
            t= np.reshape( np.diag(np.take(s, n, axis=direction)), np.shape(ksi))
    
        p=np.minimum(y, t/w)
        
        return [p, t]
    
    # indicator of the constraint set
    def __call__(self, y):
        self._check(y)
        TOL=1e-10
        if self.direction==None:
            y=np.reshape(y, (-1))
        mask= self.w*np.max(y, axis=self.direction) <= self.ksi+TOL
        
        if np.all( mask ) :
            return 0
        return np.inf

    def _check(self, y):
        sz_=1
        if self.direction !=None:
           sz= np. array( np.shape(y) , dtype=int)
           sz_= np.delete(sz, self.direction)
           
        if (self.direction !=None and  np.any( sz_!= np.shape(self.ksi) ) ) or ( self.direction==None and np.any(np.size(self.ksi) != sz_)):
           try:
               self.ksi= np.reshape( self.ksi, sz_)
           except:
              raise Exception("'ksi' is not compatible with the size of 'y' ")
       
       





