#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-15-2022)
 
 DEPENDENCIES:
     - 'newton.py' in the folder 'utils'
     
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np
from prox.utils.newton import newton_


class EpiPower:
    """
    This class computes the projection onto the epigrapgh of:
        
                phi(y)= w * |y|^ q
        
    When the inputs are arrays, the outputs are computed element-wise
     INPUTS
    ========
     y    - ND array
     kxi  - ND array with the same size as 'y'
     w    - positive, scalar or ND array with the same size as ksi
     q    - positive scalar
     """
    
    def __init__(self, ksi, q=2, w=1):

        if np.any( w <= 0 ) :
            raise Exception("'w' (or all of its components if it is an array) must be positive")  
        if np.size(q)>1 or q <= 0 :
            raise Exception("'q' (or must be a positive scalar") 
        if np.size(w)>1 and ( np.size(w) != np.size(ksi)):
           raise Exception(" 'w' must be a scalar or must have the same size as 'ksi'")
        if np.size(ksi)<=1:
            ksi=np.array([ksi])
            
        self.q=q
        self.w=w
        self.ksi=ksi
        
    # proximal operator (i.e the projetion on the constraint set)   
    def prox(self, y):
        self._check(y)
        ksi=self.ksi
        w=self.w
        q=self.q
        if np.size(y)<=1:
            y=np.array([y])
        
        # 2nd branch
        def polynom_phi(t):
            return q*w**2*t**(2*q-1) - q*w*ksi*t**(q-1) +t- np.abs(y)

        def derive_polynom_phi(t):
            return (2*q-1)*q*w**2*t**(2*q-2) - (q-1)*q*w*ksi*t**(q-2) + 1

        # starting point
        low = (np.maximum( ksi, 0)/w)**(1/q)
        root_init= 3*low
        
        # Finding the root of the polynom with the Newton method
        root = newton_(polynom_phi, fp=derive_polynom_phi, x0=root_init, low=low, high= np.inf)
        t = w* (root)**q
        p= np.sign(y)*root
        
        # 1st branch
        mask = w*np.abs(y)**q <= ksi
        p[mask] =  y[mask]
        t[mask] = ksi[mask]
        return np.array([p,t])
    
    # indicator of the constraint set
    def __call__(self, y):
        TOL=1e-10
        if np.all( self.w*np.abs(y) <= self.ksi)+TOL:
            return 0
        return np.inf

    def _check(self, y):
       if np.size(y) != np.size( self.ksi):
           raise Exception(" 'y' must have the same size as 'ksi'")
       
       
    




