#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-15-2022)
 
  DEPENDENCIES:
     - 'newton.py' in the folder 'utils'
     
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np
from prox.utils.newton import newton_


class EpiSquaredHingeLoss:
    """
    This class computes the projection onto the epigrapgh of:
        
                phi(y)= w * max( y, 0) ^ 2
        
    When the inputs are arrays, the outputs are computed element-wise
     INPUTS
    ========
     y    - ND array
     kxi  - ND array with the same size as 'y'
     w    - positive, scalar or ND array with the same size as 'ksi'
     """
    
    def __init__(self, ksi, w=1):
        if np.any( w <= 0 ) :
            raise Exception("'w' (or all of its components if it is an array) must be positive")   
        if np.size(w)>1 and ( np.size(w) != np.size(ksi)):
           raise Exception(" 'w' must be a scalar or must have the same size as 'ksi'")
        if np.size(ksi)<=1:
            ksi=np.array([ksi])
        self.w=w
        self.ksi=ksi
        
    # proximal operator (i.e the projetion on the constraint set)   
    def prox(self, y):
        self._check(y)
        ksi=self.ksi
        w=self.w
        
        if np.size(y)<=1:
            y=np.array([y])
        
        #Note: same computation as in 'EpiPower.py' for q=2
        def polynom_phi(t):
            return 2*w**2*t**3 +(1- 2*w*ksi)*t - np.abs(y)

        def derive_polynom_phi(t):
            return 6*w**2*t**2 - 2*w*ksi +1

        # starting point
        low = (np.maximum( ksi, 0)/w)**(1/2)
        root_init= 3*low
        
        # Finding the root of the polynom with the Newton method
        root = newton_(polynom_phi, fp=derive_polynom_phi, x0=root_init, low=low, high= np.inf)
        t = w* (root)**2
        p= np.sign(y)*root
        
        # 1st and 2nd branches
        mask= np.logical_or( w*np.maximum(y, 0)<=ksi, y<=0)
        p[mask] = y[mask]
        t[mask] = np.maximum( ksi[mask], 0)
        return np.array([p,t])
    
    # indicator of the constraint set
    def __call__(self, y):
        if np.all( self.w*np.maximum(y, 0)**2 <= self.ksi):
            return 0
        return np.inf

    def _check(self, y):
       if np.size(y) != np.size( self.ksi):
           raise Exception(" 'y' must have the same size as 'ksi'")
       
       
    




