#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-15-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np

class EpiSupport:
    """
    This class computes the projection onto the epigrapgh of:
        
      
                               /  * a * x   if x < 0
     phi(y)= sigma_[a,b](y) = |   0         if x = 0         
                              \   b * x     otherwise
 
        
    When the inputs are arrays, the outputs are computed element-wise
     INPUTS
    ========
     y    - ND array
     kxi  - ND array with the same size as 'y'
     a    - negative, scalar or ND array with the same size as x
     b    - positive, scalar or ND array with the same size as x
     """
    
    def __init__(self, ksi, a=-1, b=1):
        if np.any( a > 0 ) :
            raise Exception("'a' (or all of its components if it is an array) must be negative")
        if np.any( b < 0 ) :
            raise Exception("'b' (or all of its components if it is an array) must be positive")   
        if np.size(a)>1 and ( np.size(a) != np.size(ksi)):
           raise Exception(" 'a' must be a scalar or must have the same size as 'ksi'")
        if np.size(b)>1 and ( np.size(b) != np.size(ksi)):
           raise Exception(" 'b' must be a scalar or must have the same size as 'ksi'")
        if np.size(ksi)<=1:
            ksi=np.array([ksi])
        self.a=a
        self.b=b
        self.ksi=ksi
        
    # proximal operator (i.e the projetion on the constraint set)   
    def prox(self, y):
        self._check(y)
        ksi=self.ksi
        a=self.a
        b=self.b
        if np.size(y)<=1:
            y=np.array([y])
        
        # 4th branch
        p = np.zeros(np.size(y))
        t = np.zeros(np.size(y))
        
        # 3rd branch
        mask = np.logical_and( a*y > ksi, -y/a <= ksi)
        pp = (y+a*ksi) / (1+a**2)
        tt = a * pp
        p[mask] = pp[mask]
        t[mask] = tt[mask]
        
        # 2nd branch
        mask = np.logical_and( b*y > ksi ,  -y/b <= ksi)
        pp = (y+b*ksi) / (1+b**2)
        tt = b * pp
        p[mask] = pp[mask]
        t[mask] = tt[mask]
        
        # 1st branch
        mask =np.logical_and(  a*y <= ksi , b*y <= ksi)
        p[mask] =  y[mask]
        t[mask] = ksi[mask]
        return np.array([p,t])
    
    # indicator of the constraint set
    def __call__(self, y):
        if np.all( self.a*y <= self.ksi) and np.all( self.b*y <= self.ksi):
            return 0
        return np.inf

    def _check(self, y):
       if np.size(y) != np.size( self.ksi):
           raise Exception(" 'y' must have the same size as 'ksi'")
       
       
    




