#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-8-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np
from prox.indicator.Simplex import Simplex

class ExtendedSimplex(Simplex):
    """
    This class computes the projection onto the extended simplex:

                               x >= 0 and  (1,..., 1).T * X <= eta
                             
                           
     where (1, ..., 1) is a ND array with all components equal to one, and (1,..., 1).T denotes its transpose
      
     When the input 'x' is an array, the computation can vary as follows:
         - dir = 0 --> 'x' is processed as a single vector [DEFAULT]
         - dir > 0 --> 'x' is processed block-wise along the specified direction
     
     INPUTS
    ========
     x    - ND array
     eta  - positive, scalar or ND array compatible with the blocks of 'x'
     dir  - integer, direction of block-wise processing
     """
    
    def __init__(self, eta, direction=0):
        super().__init__(eta, direction)
        
    # proximal operator (i.e projetion on the extended simplex)   
    def prox(self, x):
        
        # pre-project onto the positive orthant
        p= np.maximum(x, 0)
        scalar_prod = np.sum(p, self.direc)
        
        # find the blocks still outside the simplex
        mask = scalar_prod > self.eta
        
        # project onto the simplex
        p[mask] = Simplex.prox(p[mask])
        
        return p
    
    # indicator of the extended simplex
    def __call__(self, x):
        scalar_prod = np.sum(x)
        if np.all(x) >= 0 and (scalar_prod <= self.eta):
            return 0
        else:
            return np.inf

