#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-10-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np

class Hyperslab :
    """
    This class computes the projection onto the set:

                         low  <= u.T * x <= high
                           
     where u.T denotes the transpose of u
    
     INPUTS
    ========
     x    - ND array
     u    - ND array with the same size as 'x'
     low  - scalar 
     high - scalar 
    """
    
    def __init__(self,u, low, high):
        if np.size(low)>1 or np.size(high) > 1:
            raise Exception("'low' and 'high' must be scalar or an array of one element")
        if low > high:
            raise Exception("'low' must be lower than 'high'")
        self.low=low
        self.high=high
        self.u = u
        
        
    # proximal operator (i.e projetion on the hyperslab)   
    def prox(self, x):
        self._check(x)
        scalar_prod = np.sum(x * self.u)
        u=self.u
        if scalar_prod < self.low:
            return x + ((self.low - np.sum(u * x)) / (np.linalg.norm(u)) ** 2) * u
        elif scalar_prod > self.high:
            return x + ((self.high - np.sum(u * x)) /(np.linalg.norm(u)) ** 2) *u
        return x
    
    # indicator of the hyperslab
    def __call__(self, x):
        self._check(x)
        scalar_prod = np.sum(x * self.u)
        
        if scalar_prod >= self.low and scalar_prod <= self.high:
            return 0
        else:
            return np.inf
    
    def _check(self, x):
        if  (np.shape(self.u)) != np.shape(x):
            raise Exception("'u' must have the same size as 'x'")
