#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-14-2022)
 
 DEPENDENCIES:

     - project_L1.m - located in the folder 'indicator'
     - 'Max.py' in the folder 'multi'
     - 'L2Norm.py' in the folder 'multi'
     
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np
from prox.indicator.L1Ball import L1Ball

class L1_2Ball:
    """
    Computes the projection onto the constraint set:
        
         ||x^(1)||_2 + ... + ||x^(B)||_2 <= eta
         

     INPUTS
    ========
     x         - ND array
     eta       - positive scalar
     direction - integer, direction of block-wise processing
    """
    
    def __init__(self, eta, direction=None):
        if np.size(eta)>1 or eta<=0:
            raise Exception("'eta' must be a positive scalar")
        self.eta=eta
        self.gamma=1
        self.direction=direction

    def prox(self, x):
        # compute the block norms
        xa  = np.linalg.norm(x, ord=2, axis=self.direction, keepdims=True)
        TOL=1e-10
        # compute the projection
        if np.sum( xa ) <= self.eta + TOL: # the vector if already in the set => it is equal to its'projection
            return x
        
        # project the block norms
        beta = L1Ball(self.eta)(xa)
        
        # compute the scale factors
        ya = beta / xa
        ya[xa < 1e-16] = 0
        
        # rescale the blocks
        p =  ya*x  
    
        return  p
    
    def __call__(self, x):
        # compute the block norms
        xa  = np.linalg.norm(x, ord=2, axis=self.direction, keepdims=True)
 
        if np.sum( xa ) <= self.eta: 
            return 0
        return np.inf
    