#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-13-2022)

 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np 

class L2PositiveNorm:
    """
    Computes the proximity operator of the function :
    
                        f(x) = gamma * tau*( ||x||_2 + ind( x>0 )
    
    ind( . > 0 ) is the indicator function of the set [0, +inf)^2 ( +inf if x<0 and 0 if x>0 or x==+inf)
    
    When the input 'x' is an array, the computation can vary as follows:
     - direction <  0 --> 'x' is processed as a single vector [DEFAULT]
     - direction => 0 --> 'x' is processed block-wise along the specified direction
                          (0 -> columns, 1 -> rows, etc.)
    
     INPUTS
    ========
     x         - ND array
     gamma     - positive, scalar or ND array compatible with the blocks of 'x'
     direction - integer, direction of block-wise processing
        """
    def __init__(self, gamma=1, direction=None):
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
        self.direction= direction
        
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale= tau*self.gamma
        xx= np.maximum(x, 0)
        l2 = np.linalg.norm(xx, 2, axis=self.direction, keepdims=True)
        coef = np.maximum(0, 1 - scale / l2)
        p = coef * xx
        return p
        
    def __call__(self, x):
        l2norm_= np.linalg.norm(x, 2, axis=self.direction, keepdims=True) 
        if np.all( x> 0):
            indi=0
        else:
            indi=np.inf
        
        return l2norm_ + indi
    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")