#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-8-2022)
 
 DEPENDENCIES:
 'Max.py' in the folder 'multi'
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np
from prox.multi.Max import Max

class Simplex:
    """
    This class computes the projection onto the simplex:

                               x >= 0 and  (1,..., 1).T * X = eta
                             \  
                           
     where (1, ..., 1) is a ND array with all components equal to one, and (1,..., 1).T its transpose
      
     When the input 'x' is an array, the computation can vary as follows:
         - dir = 0 --> 'x' is processed as a single vector [DEFAULT]
         - dir > 0 --> 'x' is processed block-wise along the specified direction
     
     INPUTS
    ========
     x    - ND array
     eta  - positive, scalar or ND array compatible with the blocks of 'x'
     dir  - integer, direction of block-wise processing
     """
    
    def __init__(self, eta, direction=0):
        if np.any( eta <= 0 ):
            raise Exception("'eta' (or all of its components if it is an array) must be positive")
        self.eta=eta
        self.direc=direction
        
    # proximal operator (i.e projetion on the simplex)   
    def prox(self, x):
        self._check(x, self.eta, "'eta'")
        return x-Max( self.eta, self.direc).prox(-x)
    
    # indicator of the simplex
    def __call__(self, x):
        scalar_prod = np.sum(x)
        tol = 1e-10
        if np.all(x) >= 0 and np.abs(scalar_prod - self.eta) < tol:
            return 0
        else:
            return np.inf

    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
          