#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-22-2022)
 
 Author  : Mbaye Diongue
        
 Copyright (C) 2022
     
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np


class AbsDiff:
    """
    This class computes the proximity operator of the function (gamma*tau)*D where D is
    defined as:
                                 
  
                  /  |x-y|        if x >= 0 and y >= 0
        D(x,y) = |                                 
                 \  + inf         otherwise
                
            'gamma * tau' is the scale factor

            When the inputs are arrays, the outputs are computed element-wise

    ========
     INPUTS
    ========
    x          - scalar or ND array 
    y          - scalar if 'x' is a scalar , ND array with the same size as 'x' otherwise
    gamma      - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    tau        - positive, scalar or ND array with the same size as 'x' [default: tau=1] [Only for the method 'prox']
    """
    
    def __init__(self, gamma=1):
        if np.any( gamma <= 0 ) :
            raise ValueError("'gamma'  must be strictly positive")
        self.gamma=gamma
      
        
        
    def prox(self, x, y, tau=1):
        if np.size(x) != np.size(y):
            raise ValueError("'x' and 'y' must have the same size")
        scale=tau*self.gamma
        self._check(x,scale)
        
        if np.size(x)<=1:
            x=np.reshape(x, (-1))
            y=np.reshape(y, (-1))
        sz= np.shape(x)
        # 4th branch
        p = np.zeros(sz)
        q = np.zeros(sz)
        
        # 3rd branch
        mask = (y > scale) * (x <= -scale);
        yy = y - scale
        q[mask]= yy[mask]
        
        # 2nd branch
        mask = (x > scale) * (y <= - scale)
        xx = x - scale
        p[mask] = xx[mask]
        
        # 1st branch
        t = np.sign(x-y) * np.maximum(0, np.abs(x-y) - 2*scale)
        mask = np.abs(t) < x+y
        xy = x[mask]+ y[mask]
        tt = t[mask]
        p[mask] = 0.5 * (xy + tt)
        q[mask]= 0.5 * (xy - tt)
        
        return [ p, q]
    
    def __call__(self, x, y):
       if np.any( x<0) or np.any( y<0 ):
            return np.inf
       return np.sum(self.gamma*np.abs(x-y))


    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ) or ( np.size(tau)>1 and np.size(tau) != np.size(x)):
           ValueError("'tau' and 'gamma' must be positive scalars or positive ND arrays with the same size as 'x'")