#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-13-2022)

 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np 

class AffineBariier:
    """
    Computes the proximity operator of (gamma*tau)*f where f is definded as :
    
                               / -log(b- a.T*x)    if u.T*x < b
                        f(x) =|   
                              \   + inf            otherwise
                              
                              
             When the input 'x' is an array, the computation can vary as follows:
                 
    - direction = None --> 'x' is processed as a single vector [DEFAULT]
       In this case 'gamma' must be a scalar
    - direction >=0   --> 'x' is processed block-wise along the specified direction
        (0 -> rows, 1-> columns ect. in this case, 'gamma', 'epsilon' and 'tau' must
         be singletons along 'direction')
                              
    As f is an indicator funtion, its proximity operator is the projection onto the affine barriers
    
     INPUTS
    ========
     x         - ND array
     a         - ND array with the same shape as 'x'
     b         - scalar or ND array compatible with the blocks of 'x'
     gamma     - positive, scalar or ND array compatible with the blocks of 'x' (default: gamma=1)
     tau       - positive, scalar or ND array compatible with the blocks of 'x' (default: tau=1)( only for the method 'prox')
     direction - integer, direction of block-wise processing [default: direction=None] ( 0 -> rows, 1->columns etc.)
     
     =========
     Examples
     ==========
     
    
     >>> import numpy as np
     >>> from proximityOperator import AffineBariier
     >>>
     >>> x=np.array([1,2,3])
     >>> a= np.array([-1, 5, 3])
     >>> b= 3.5
     >>> AffineBariier(x,b)(x)
     array(inf)      

    As the result below is infinite, that means 'x' does not belong to the affine set.
    
    Projection of x onto the affine set:

     >>> px= AffineBariier(x,b).prox(x) 
     >>> px
     >>> AffineBariier(x,b)(px)
     array(-0.17973361)    #

    As we could have suspected, the projection of 'x' belongs to the affine set.
     
     >>> x=np.arange(6)
     >>> x=x.reshape((2,3))
     >>> x
     array([[0, 1, 2],
           [3, 4, 5]])
     >>> a =np.ones
     >>> a[0,:]=2
     >>> a
     array([[2., 2., 2.],
            [1., 1., 1.]])
     >>> b=np.array([-1, 2, 4])
     
     Set 'direction=0' to process along the rows of the matrix 'x':
         
     >>> AffineBariier(x,b, direction=0)(x)
     array([inf, inf, inf])
     
     Projection of x onton the affine set:
         
     >>> px = AffineBariier(x,b, direction=0).prox(x)  
     >>> px
     array([[0.        , 0.05537521, 0.19926499],
            [0.        , 0.22150085, 0.49816248]])
     >>> AffineBariier(x,b, direction=0)(px)
     array([-0.        , -0.05696748, -0.10495226])

     """
    def __init__(self, a, b, gamma=1, direction=None):
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")

        if (direction is not None) and (direction < 0):
            direction = None
        
        self.direction=direction
        self.gamma = gamma
        self.a =a
        self.b =b
        
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale= tau*self.gamma
        a=self.a
        b=self.b
        sz=np.shape(x)
        sz=np. array( sz , dtype=int)
        sz[self.direction]=1
        if np.size(scale)>1:
            scale=np.reshape(scale, sz)
        if np.size(b)>1:
            b=np.reshape(b, sz) 
                             
        scalar_prod = np.sum( self.a*x, axis=self.direction).reshape(sz)
        l2_a= np.sum(self.a**2, axis=self.direction).reshape(sz)
        return x+ ( b-scalar_prod- np.sqrt( (b- scalar_prod)**2 + 4*scale*l2_a))/(2*l2_a)*a
        
    def __call__(self, x):
        self._check(x, self.gamma, "gamma")
        scalar_prod = np.sum( self.a*x, self.direction)
        b=self.b
        if np.size(b) != np.size(scalar_prod):
            raise ValueError("'b' must be compatible with the blocks of 'x'")
        b=np.reshape(b, np.shape(scalar_prod))
        res = np.zeros_like(1.0*scalar_prod)
        mask=scalar_prod< self.b
        res[mask]=-np.log( b[mask]- scalar_prod[mask])
        res[ np.logical_not(mask)]=np.inf
        return res
    
    def _multiply_tuple(self, tup):
        res=1
        for i in tup:
            res=res*i
        return res
    
    def _check(self, x, scale, name_name="'tau'"):
       if np.any( scale <= 0 ):
           raise ValueError(name_name+" (or all of its components if it is an array) must be strictly positive")
           
       if (self.direction==None) and np.size(scale)>1:
            raise ValueError("'gamma' and 'tau' must be scalars when the argument 'direction' is equal to 'None'")
            
       sz=np.shape(x)
       if len(sz)<=1:
           self.direction=None
       if np.size(scale)<=1:
           return 
       
       if len(sz)<=1 and np.size(scale)>1 :
           raise ValueError("'tau' and 'gamma' must be scalar when 'x' is one dimensional")
           
       if len(sz)> 1 and self.direction != None:
           sz=np. array( sz , dtype=int)
           sz[self.direction]=1
           if np.size(scale)>1 and ( self._multiply_tuple(sz) != np.size(scale)):
               raise ValueError("The dimension of 'tau' or 'gamma' is not compatible with the blocks of 'x'")
           
          