#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-22-2022)
 
 DEPENDENCIES:
     -'Ialpha.py' - located in the folder 'multi'
     
 Author  : Mbaye Diongue
        
 Copyright (C) 2022
     
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np
from prox.multi.Ialpha import Ialpha

class Hellinger:
    """
    This class computes the proximity operator of the function (gamma*tau)*D where D is
    defined as:
                                 
  
                  /  ( sqrt(x) - sqrt(y) )^2   if x >= 0 and y >= 0
        D(x,y) = |                                 
                 \  + inf                    otherwise
                
            'gamma * tau' is the scale factor

            When the inputs are arrays, the outputs are computed element-wise
            
     ========
     INPUTS
    ========
    x          - scalar or ND array 
    y          - scalar if 'x' is a scalar , ND array with the same size as 'x' otherwise
    gamma      - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    tau        - positive, scalar or ND array with the same size as 'x' [default: tau=1] [Only for the method 'prox']
    """
    
    def __init__(self, gamma=1):
        if np.any( gamma <= 0 ) :
            raise Exception("'gamma'  must be strictly positive")
        self.gamma=gamma
      
        
        
    def prox(self, x, y, tau=1):
        scale=tau*self.gamma
        return Ialpha(alpha=2, gamma=2*scale).prox(x,y)
    
    def __call__(self, x, y):
       if np.any( x<0) or np.any( y<0 ):
            return np.inf
       return np.sum(self.gamma*( np.sqrt(x) - np.sqrt(y))**2 )
    
    def _check(self, y, tau, name_name="'tau'"):
       if np.any( tau <= 0 ) or ( np.size(tau)>1 and np.size(tau) != np.size(y)):
           raise Exception(name_name+" must be positive scalar or an positive ND array with the same size as 'y'")