#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-13-2022)

 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np 
from prox.utils.newton import newton_


class HyperslabBariier:
    """
    Computes the proximity operator of (gamma*tau)*f where f is definded as :
    
                               / -log( high - a.T*x)-log( a.T - low)    if low < u.T*x < high
                        f(x) =|   
                              \   + inf                                otherwise
                              
                              
                   'gamma*tau' is the scale factor
                   
    When the input 'x' is an array, the computation can vary as follows:
     - direction <  0 --> 'x' is processed as a single vector [DEFAULT]
     
     - direction => 0 --> 'x' is processed block-wise along the specified direction
        (0 -> columns, 1 -> rows, etc. in this case, 'gamma', 'high', 'low' and 'tau' must be singletons along 'direction')
    
                              
     INPUTS
    ========
     x         - ND array
     a         - ND array with the same size as 'x', all of its components can not be equal to zero
     low       - scalar or ND array compatible with the blocks of 'x'    [default: low  = -1]
     high      - scalar or ND array compatible with the blocks of 'x'    [default: high =  1]
     gamma     - positive, scalar or ND array compatible with the blocks of 'x' [default: gamma=1]
     tau       - positive, scalar or ND array compatible with the blocks of 'x' [default: tau=1] [only for the method 'prox'] 
    """
    def __init__(self, a, low=-1, high=1, direction=None, gamma=1):
        if np.any( gamma <= 0 ):
            raise ValueError("'gamma' (or all of its components if it is an array) must be strictly positive")
        if np.any(low > high):
            raise ValueError(" 'high' must be greater than 'low'")
        if np.size(low)<=1:
            low=np.reshape(low, (-1))
            high=np.reshape(high, (-1))
        self.gamma = gamma
        self.a = a
        self.low = low
        self.high =high
        self.direction=direction
        
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale= tau*self.gamma
        self._check(x, scale)
        low=self.low
        high= self.high
        sz0=np.shape(x)
        sz0=np. array( sz0 , dtype=int)
        sz0[self.direction]=1
        if np.size(scale)>1:
            scale=np.reshape(scale, sz0)
        
        sp = np.sum( self.a*x, self.direction) # scalar product
        l2_a= np.sum(self.a**2, axis=self.direction)
    
        def polynom_phi(t):
            return t ** 3 - (low+high+sp)*t**2+( low*high+ sp*(low+high)-2*scale*l2_a)*t -low*high*sp+ scale*(low+high)*l2_a

        def derive_polynom_phi(t):
            return 3*t**2 - 2*(low+high+sp)*t+( low*high+ sp*(low+high)-2*scale*l2_a)

        # starting point
        root_init = (low+high)/2

        # Find the root of the polynom with the Newton method
        root = newton_(polynom_phi, fp=derive_polynom_phi, x0=root_init, low=low, high= high)
        
        coef= np.reshape(  (root - sp)/l2_a , sz0)
        return x+ coef*self.a
    
    def __call__(self, x):
        self._check(x, self.gamma)
        scalar_prod = np.sum( self.a*x)
        if np.size(scalar_prod)<=1:
            scalar_prod= np.reshape(scalar_prod, (-1))
        res=np.inf*np.ones_like(scalar_prod)
        mask= (scalar_prod< self.high) *(scalar_prod > self.low)
        res[mask]= -np.log( self.high[mask]- scalar_prod[mask])- np.log( scalar_prod [mask]-self.low[mask])
        
        return self.gamma*res
    
          
    def _multiply_tuple(self, tup):
        res=1
        for i in tup:
            res=res*i
        return res
    def _check(self, x, scale, name_name="'tau'"):
       if np.any( scale <= 0 ):
           raise ValueError(name_name+" (or all of its components if it is an array) must be strictly positive")
           
       if (self.direction==None) and np.size(scale)>1:
            raise ValueError("'gamma' and 'tau' must be scalars when the argument 'direction' is equal to 'None'")
            
       sz=np.shape(x)
       if len(sz)<=1:
           self.direction=None
       if np.size(scale)<=1:
           return 
       
       if len(sz)<=1 and np.size(scale)>1 :
           raise ValueError("'tau' and 'gamma' must be scalar when 'x' is one dimensional")
           
       if len(sz)> 1 and self.direction != None:
           sz=np. array( sz , dtype=int)
           sz[self.direction]=1
           if np.size(scale)>1 and ( self._multiply_tuple(sz) != np.size(scale)):
               raise ValueError("The dimension of 'tau' or 'gamma' is not compatible with the blocks of 'x'")
           
           
    