#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-22-2022)

 DEPENDENCIES:
     -'newton.py' in the folder 'utils'
     
 Author  : Mbaye Diongue
        
 Copyright (C) 2022
     
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np
from prox.utils.newton import newton_


class Ialpha:
    """
    This class computes the proximity operator of the function (gamma*tau)*D where D is
    defined as:
                                 
  
                  /  - ( x* y^{alpha-1})^{1/alpha}   if x >= 0 and y >= 0
        D(x,y) = |                                 
                 \   + inf                             otherwise
                
            'gamma * tau' is the scale factor

            When the inputs are arrays, the outputs are computed element-wise
    
    ========
     INPUTS
    ========
    x          - scalar or ND array 
    y          - scalar if 'x' is a scalar , ND array with the same size as 'x' otherwise
    alpha      - >1, scalar or ND array [ default: alpha=2]
    gamma      - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    tau        - positive, scalar or ND array with the same size as 'x' [default: tau=1] [Only for the method 'prox']
    """
    
    def __init__(self, alpha=2, gamma=1):
        if np.any( gamma <= 0 ) :
            raise Exception("'gamma'  must be strictly positive")
        if np.any( alpha <= 1 ) or np.size(alpha)>1:
            raise ValueError("'alpha'  must be a scalar greter than 1")
        self.gamma=gamma
        self.alpha=alpha
        
    def prox(self, x, y, tau=1):
        scale=tau*self.gamma
        self._check(y, scale)
        alpha=self.alpha
        if np.size(x) != np.size(y):
            raise Exception("'x' and 'y' must have the same size")
            
        # scalar-like inputs handling 
        if np.size(x) <=1:
            x= np.reshape( x, (-1))
            y= np.reshape( y, (-1)) 
            
        # 2nd branch
        sz=np.shape(x)
        p = np.zeros( sz )
        q = np.zeros( sz ) 
        
        # branch selection
        mask = np.logical_or( alpha*x >= scale,   1-alpha*y/(scale*(alpha-1)) < scale/(scale-alpha*x)**(1/(alpha-1)))
        xx = x[mask]
        yy = y[mask]
        gg= scale
        if np.size(scale) > 1:
            gg = scale[mask]
        # newton's method
        a = alpha*xx/gg - 1
        b = alpha-1-alpha*yy/gg
        def fun_phi(t):
            return t**(2*alpha) +   a * t**(alpha+1) + b * t - alpha +1
        def der_fun_phi(t):
            return 2*alpha*t**(2*alpha-1) +   alpha*a * t**alpha + b 
        
        # root finding
        eps= 1e-10
        low= np.maximum(0, 1-alpha*xx/gg)**(1/(alpha-1)) + eps
        t_init = 50/alpha + low
        t =  newton_(fun_phi, fp=der_fun_phi, x0=t_init, low=low, high= np.inf)
        
        # special cases
        t[ (a==0) *(b==0)] = 1
        
        # 1st branch
        p[mask] = xx + gg/alpha*( t**(alpha-1) - 1)
        q[mask] = yy + gg*(1-1/alpha)*(1/t - 1)
        return [p,q]
    
    def __call__(self, x, y):
       if np.any( x<0) or np.any( y<0 ):
            return np.inf
       alpha=self.alpha
       return np.sum(self.gamma*( -x**alpha * y**(1-1/alpha) ))


    def _check(self, y, tau, name_name="'tau'"):
       if np.any( tau <= 0 ) or ( np.size(tau)>1 and np.size(tau) != np.size(y)):
           raise Exception("'gamma' and 'tau' must be positive scalars or an positive ND arrays with the same size as 'y'")