#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-20-2022)
 
 DEPENDENCIES: 
     - 'L2Norm.py' located in the folder 'multi'
     
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np 
from prox.multi.L2Norm import L2Norm

class L21rows:
    """
    Computes the proximity operator of (gamma*tau)*f where f is the function defined as follows:

                               
                         f(x) = f(x) = \sum_{i=1}^N|\sum_{j=1}^M|X(i,j)|^2|^{\frac{1}{2}}
                             
            where X = U*diag(s)*V.T \in R^{M*N}  (Singular Value decomposition) 
        
     INPUTS
    ========
     x         -  (M,N) -array_like ( representinf a M*N matrix )
     gamma     - positive, scalar or ND array compatible with the size of 'x'[default: gamma=1]
     tau       - positive, scalar or ND array compatible with the size of 'x'[default: tau=1] [only for the method 'prox']
     """
     
    def __init__(self, gamma=1):
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
       
    
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale= tau*self.gamma
        p=L2Norm(gamma=scale, direction=1).prox(x)
        return p
    
    def __call__(self, x):
        p = L2Norm(gamma=self.gamma, direction=1)(x)
        return np.sum(p)
    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       if len( np.shape(x))!=2:
           raise ValueError("'x' must be an (M,N) -array_like ( representinf a M*N matrix )")
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
