#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-14-2022)

 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np 
from prox.utils.newton import newton_


class L2ballBarrier:
    """
    Computes the proximity operator of (gamma*tau)*f where f is definded as :
    
                               / -log( radius2 -  || x - c||^2   if      || x - c||_2)^2 < radius2
                        f(x) =|   
                              \   + inf                        otherwise
                              
    It is a barrier of l2 ball constraint       
    
     INPUTS
    ========
     x         - ND array
     c         - ND array with the same size as 'x' ( center of the ball constraint)
     radius2     - scalar , positive  (square of the radius)
     gamma     - positive, scalar or ND array compatible with the blocks of 'x'
     direction - integer, direction of block-wise processing
     """
     
    def __init__(self, center, radius2, gamma=1):
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        if np.size(radius2) > 1 or radius2 <0 :
            raise Exception( "'radius' must be a positive scalar")
        
        self.gamma = gamma
        self.r= radius2
        self.c= center
        
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale= tau*self.gamma
        
        c=self.c
        r= self.r
        l2_norm=np.linalg.norm( x- c, 2)
        
        def polynom_phi(t):
            return t ** 3 - l2_norm*t**2 -( r+ 2*scale)*t + r*l2_norm

        def derive_polynom_phi(t):
            return 3*t**2 - 2*l2_norm*t -( r+ 2*scale)

        # starting point
        root_init = np.sqrt(r)/2

        # Finding the root of the polynom with the Newton method
        root = newton_(polynom_phi, fp=derive_polynom_phi, x0=root_init, low=0, high= np.sqrt(r))

        return c + ( r- root**2)/(r-root**2+2*scale)*(x-c)
    
    def __call__(self, x):
        l2_norm2=np.linalg.norm( x- self.c, 2)**2
        if l2_norm2 < self.r :
            return -np.log(self.r -l2_norm2)
        return np.inf
    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")