#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-14-2022)

 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 REFERENCE:
     Emmanuel Soubies and Michael Unser. "Computational Super-Sectioning for Single-Slice
     Structured-Illumination Microscopy", 5-2 (2019), pp. 240--250
     
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np 
from numpy import matlib as mb
import numpy.fft as fft

class L2DownConv:
    """
    Computes the proximity operator of (gamma*tau)*f where f is definded as :
    
                        f(x) = || S_d*H*x - y||_2 ^2
                              
    Where S_d is the downsampling operator with factor d>0 and H is a convolution operator that
    is described by the Fourier Kernel Lamb [i.e. Hx = ifftn( Lamb*fftn(x) ) ]     
    
     INPUTS
    ========
     x         - ND array
     Lamb      - ND array with the same size as 'x' ( Fourier kernel of the
     convolution operator H)
     d         - vector of positive integers (dowsampling in each dimension of x)
     gamma     - positive scalar 
     y         - ND array with size : size(x)/d
     """
     
    def __init__(self, Lamb, d, y, gamma=1):
        if np.size(gamma)> 1 or  gamma <= 0 :
            raise Exception("'gamma' must be a strictly positive scalar")
 
        self.gamma = gamma
        self.Lamb= Lamb
        self.d = d
        self.y = y
        
        # Precomputations
        sz= np.shape(Lamb)
        sel=[]
        
        if np.size(d)<=1:
            sel.append( np.arange(1, sz[0], d) )
        else:
            for ii in range( np.size(d)):
                sel.append( np.arange(1, sz[ii], d[ii]) )
        self.sel=sel
        
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale=self.gamma*tau
        y=self.y
        d=self.d
        Lamb=self.Lamb
        yUP= np.zeros( np.shape(x))
        yUP[ tuple(self.sel)] = y
        sz_= np.array([ dimk for dimk in np.shape(x) ])
        szPatch=  np.array(sz_/d, dtype=int)
        sel2=[]
        for n in range(len(np.shape(x))):
            if np.size(d) <=1:
                sel2.append( szPatch[n]*np.ones( (1, d ) ))
            else:
                sel2.append( szPatch[n]*np.ones( (1, d[n]) ))
        
        
        def Pt( x, sel, szPatch):
            y=np.zeros(szPatch)
            sz_=[]
            for k in range(len(szPatch)):
                sz_.append( szPatch[k])
            sz_.append(-1)
            sz_=tuple(sz_)
            tmp=np.reshape(x, sz_)
            for n in range( np.shape(tmp)[-1]):
                y=y+tmp[...,n]
            return y
                
        H2H2t = Pt( np.abs(Lamb)**2, sel2, szPatch )
        
        def P(xx):
            d_=d
            if np.size(d)<=1:
                d_=[d,1]
            
            resP=mb.repmat(x,d_[0], d_[1])
            return resP
        
        # Prox computation
        fftr = np.conj(Lamb)*np.fft.fftn(yUP)+np.fft.fftn(x/scale)
        p=fft.ifftn(scale*(fftr - np.conj(Lamb)*P(Pt(Lamb*fftr,sel2,szPatch)/(np.size(d)/scale+H2H2t))))
        
        return p.real[0]
                           
                           
    def __call__(self, x, tau=1):
        self._check( x, tau)
        Hx=fft.fftn( self.Lamb*np.fft.ifftn(x))
        Hx=Hx.real  # the real part of the Fourier transform
        res= self.gamma/2*np.linalg.norm( Hx[ tuple(self.sel)]-self.y, 2)**2
        return res
    
    def _check(self, x, tau):
       if np.size(tau)>1 or tau <= 0 :
           raise Exception("'tau' must be a strictly positive scalar")
       if np.size(self.y) != np.size(x)/self.d :
           raise Exception("'y' must be an ND array with size: np.size(x)/d")
       if np.size( self.Lamb) != np.size(x):
           raise Exception("'Lamb' must be an ND array with size as 'x'")