#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-13-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np 

class L2Norm :
    """
    Computes the proximity operator of the function
    
                        f(x) = (gamma*tau)* ||x||_2
    
    'gamma*tau' is the scale factor
    
    When the input 'x' is an array, the computation can vary as follows:
     - direction <  0 --> 'x' is processed as a single vector [DEFAULT]
     
     - direction => 0 --> 'x' is processed block-wise along the specified direction
        (0 -> columns, 1 -> rows, etc. in this case, 'gamma', 'epsilon' and 'tau' must be singletons along 'direction')
    
     INPUTS
    ========
     x         - ND array
     direction - integer, direction of block-wise processing
     gamma     - positive, scalar or ND array compatible with the blocks of 'x'  [default: gamma=1]
     tau      - positive, scalar or ND array compatible with the blocks of 'x' [default: tau=1] [only for the method 'prox'] 
      
     """
    def __init__(self, gamma=1, direction=None):
        if np.any( gamma <= 0 ):
            raise ValueError("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
        self.direction= direction
        
    def prox(self, x, tau=1):
        
        scale= tau*self.gamma
        self._check( x, scale)
        
        xx = np.sqrt( np.sum(x**2, axis=self.direction, keepdims=True) )
        xx = np.maximum(0, 1 - scale / xx)
        p = x * xx
        return p
        
    def __call__(self, x):
        self._check(x, self.gamma)
        l2=np.sqrt( np.sum(x**2, axis=self.direction))
        g=self.gamma
        if np.size(g)<=1:
            return g*l2
        g=np.reshape(g, l2.shape)
        return g*l2
    
    def _multiply_tuple(self, tup):
        res=1
        for i in tup:
            res=res*i
        return res
    
    def _check(self, x, scale, name_name="'tau'"):
       if np.any( scale <= 0 ):
           raise ValueError(name_name+" (or all of its components if it is an array) must be strictly positive")
           
       if (self.direction==None) and np.size(scale)>1:
            raise ValueError("'gamma' and 'tau' must be scalars when the argument 'direction' is equal to 'None'")
            
       sz=np.shape(x)
       if len(sz)<=1:
           self.direction=None
       if np.size(scale)<=1:
           return 
       
       if len(sz)<=1 and np.size(scale)>1 :
           raise ValueError("'tau' and 'gamma' must be scalar when 'x' is one dimensional")
           
       if len(sz)> 1 and self.direction != None:
           sz=np. array( sz , dtype=int)
           sz[self.direction]=1
           if np.size(scale)>1 and ( self._multiply_tuple(sz) != np.size(scale)):
               raise ValueError("The dimension of 'tau' or 'gamma' is not compatible with the blocks of 'x'")
           
          