#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-13-2022)
 
 DEPENDENCIES:
     - 'Max.py' in the folder 'multi'
     
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""
import numpy as np
from prox.multi.Max import Max


class Linf(Max):
    """
    Computes the proximity operator of the function
    
                        f(x) = gamma *tau* max( |x1|, ..., |xn|)

        'gamma * tau' is the scale factor
            
    When the input 'x' is an array, the computation can vary as follows:
    - direction = None --> 'x' is processed as a single vector [DEFAULT]
       In this case 'gamma' must be a scalar
    - direction >=0   --> 'x' is processed block-wise along the specified direction
    
    INPUTS
    ========
     x         - ND array
     gamma     - positive, scalar or ND array compatible with the blocks of 'x' (default value : gamma=1)
     tau       - positive, scalar or ND array compatible with the blocks of 'x' (default value : tau=1)
     direction - integer, direction of block-wise processing
    """
    
    def __init__(self, gamma=1, direction=None):
        super().__init__(gamma=gamma, direction=direction)

    def prox(self, x, tau=1):
        scale= tau*self.gamma
        Max._check(self, x, scale)
        
        abs_x = np.abs(x)
        p = np.sign(x) *  Max(gamma=scale, direction=self.direction).prox(abs_x)
        return p
    
    def __call__(self, x):
        Max._check(self, x, self.gamma)
        l1=np.max( np.abs(x), self.direction, keepdims=True)
        g=self.gamma
        if np.size(g)<=1:
            return g*l1
        g=np.reshape(g, l1.shape)
        return g*l1
    