#!#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-21-2022)
 
 DEPENDENCIES: 
     - 'prox_svd.py' located in the folder 'utils'
     - 'prox_svd.py' located in the folder 'utils'
     
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """


from prox.utils.prox_svd import prox_svd
from prox.utils.fun_svd import fun_svd
import numpy as np 

class LogDet:
    """
    Computes the proximity operator of (gamma*tau)*f where f is the function defined as follows:

               /-log( det(X) ) = -log( prod(s))  if X is a symmetric positive definite matrix
         f(x)=|
              \ + inf                            otherwise
                         
    where
           * det(X) is the determinant of the matrix X 
                             
           * X = U*diag(s)*V.T \in R^{M*N}  the Singular Value decomposition of X
            
          * 'gamma*tau' is the scale factor 
        
     INPUTS
    ========
     x               -  (M,N) -array_like ( representing a M*N matrix
     gamma           - positive scalar  [default: gamma=1]
     tau             - positive scalar [default: tau=1] [only for the method 'prox']
     
     """
     
    def __init__(self, gamma=1):
        if np.any( gamma <=0 ) or np.size(gamma)>1:
            raise Exception("'gamma'  must be a strictly positive scalar")
        self.gamma = gamma
        
    
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale= tau*self.gamma
        
        def prox_phi(s, gam):
            return 0.5*(s+ np.sqrt(s**2+4*gam))
        
        return prox_svd(x, scale, prox_phi, hermitian=True)
    
    def __call__(self, x):
        self._check(x, self.gamma)
        TOL=1e-12
        # Check if the matrix is symmetric
        if not np.allclose(x, np.transpose(x)):
            return np.inf
        def fun_phi(s):
            if np.any(s<=TOL):
                return np.inf
            return -np.log(np.prod(s))
        
        return self.gamma*fun_svd(x, 1, fun_phi)
    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ) or np.size(tau)>1 :
           raise Exception("'mu'  must be a strictly positive scalar")
       if len( np.shape(x))!=2:
           raise ValueError("'x' must be an (M,N) -array_like ( representinf a M*N matrix )")
