#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-13-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""
import numpy as np

class Max :
    """
    Computes the proximity operator of the function
    
                        f(x) = (gamma tau)* max( x1, ..., xn)
                        
            'gamma * tau' is the scale factor
            
    When the input 'x' is an array, the computation can vary as follows:
    - direction = None --> 'x' is processed as a single vector [DEFAULT]
       In this case 'gamma' must be a scalar
    - direction >=0   --> 'x' is processed block-wise along the specified direction
        (in this case, 'gamma', 'epsilon' and 'tau' must be singletons along 'direction')
        
     INPUTS
    ========
     x         - scalar or ND array
     gamma     - positive, scalar or ND array compatible with the blocks of 'x'[default: gamma=1]
     tau       - positive, scalar or ND array compatible with the blocks of 'x'[default: tau=1] [only for the method 'prox']
     direction - integer, direction of block-wise processing
    """
    
    def __init__(self, gamma=1, direction=None):
        if np.any( gamma <= 0 ):
            raise ValueError("'gamma' (or all of its components if it is an array) must be strictly positive")

        if (direction is not None) and (direction < 0):
            direction = None
        
        self.direction=direction
        self.gamma=gamma

        
    def prox(self, x, tau=1):
        scale= tau*self.gamma
        self._check( x, scale)
        direction= self.direction
        sz = np.shape(x)
        
        if np.size(x)<=1:
            x=np.reshape(x, (-1))
        
        sz0=np.shape(x)
        direction=self.direction
        sz0=np. array( sz0 , dtype=int)
        sz0[direction]=1
        if np.size(scale)>1:
            scale=np.reshape(scale, sz0)
            
        # 0. Linearize
        if direction==None:
            x=np.reshape(x,(-1))
        
        # 1. Order the column elements (sort in decreasing order)
        s = -np.sort( -x, axis=direction)
         
        # 2. Compute the partial sums: c(j) = ( sum(s(1:j)) - gamma ) / j
        ones_=np.ones(np.shape(s))

        c = ( np.cumsum(s,axis=direction) - scale ) / np.cumsum( ones_, axis= direction )
    
        # 3. Find the index: n = max{ j \in {1,...,B} : s(j) > c(j) }
        mask=s>c
        mat=(np.arange( np.size(mask))).reshape( np.shape(mask))
        mask = mask*mat
        n = np.argmax(mask, axis=direction)

        if np.size(n)<=1:
            t=np.reshape( c[n], np.shape(scale))
        else:
            t=np.minimum( np.diag(np.take(c, n, axis=direction)).reshape(sz0) , x)
            t=np.reshape(t, np.shape(x) )
        
        
        # 4. Compute the prox
        p= np.minimum( x, t)
        
        # 5. Output is zero if 's(j) > c(j)' for all j
        p =  p*( (1 - np.all(mask, direction)).reshape(sz0))
        
        # 6. revert back
        p = np.reshape(p, sz)
        
        return p
    
    
    def __call__(self, x):
        self._check(x, self.gamma)
        Maxi_=np.max( x, axis=self.direction, keepdims=True)
        g=self.gamma
        if np.size(g)<=1:
            return g*Maxi_
        g=np.reshape(g, Maxi_.shape)
        return g*Maxi_
        
    def _multiply_tuple(self, tup):
        res=1
        for i in tup:
            res=res*i
        return res
    def _check(self, x, scale, name_name="'tau'"):
       if np.any( scale <= 0 ):
           raise ValueError(name_name+" (or all of its components if it is an array) must be strictly positive")
           
       if (self.direction==None) and np.size(scale)>1:
            raise ValueError("'gamma' and 'tau' must be scalars when the argument 'direction' is equal to 'None'")
            
       sz=np.shape(x)
       if len(sz)<=1:
           self.direction=None
       if np.size(scale)<=1:
           return 
       
       if len(sz)<=1 and np.size(scale)>1 :
           raise ValueError("'tau' and 'gamma' must be scalar when 'x' is one dimensional")
           
       if len(sz)> 1 and self.direction != None:
           sz=np. array( sz , dtype=int)
           sz[self.direction]=1
           if np.size(scale)>1 and ( self._multiply_tuple(sz) != np.size(scale)):
               raise ValueError("The dimension of 'tau' or 'gamma' is not compatible with the blocks of 'x'")
           
           
    