#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-14-2022)
 
 DEPENDENCIES: 
     - 'AbsValue.py' located in the folder 'scalar'
     - 'prox_svd.py' located in the folder 'utils'
     - 'NuclearNorm.py' located in the folder 'multi'
     - 'Linf.py' located in the folder 'multi'
     - 'Max.py' in the folder 'multi'
     - 'L1Ball.py' in the folder 'indicator'
     
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np 
from prox.indicator.L1Ball import L1Ball
from prox.utils.prox_svd import prox_svd
from prox.multi.NuclearNorm import NuclearNorm

class NuclearBall:
    """
    Computes the proximity operator of (gamma*tau)*f where f is the indicator of
    the constraint set :
                               
                         ||X||_N =||s||_1 <= eta   (nuclear norm ball constraint)
                             
            where X = U*diag(s)*V.T \in R^{M*N} 
     INPUTS
    ========
     x         -  (M,N) -array_like ( representinf a M*N matrix )
     gamma     - positive, scalar or ND array compatible with the size of 'x'[default: gamma=1]
     tau       - positive, scalar or ND array compatible with the size of 'x'[default: tau=1]
     eta       - positive, scalar or ND array compatible with the size of 'x'
     """
     
    def __init__(self, eta, gamma=1):
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
        self.eta=eta
    
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale= tau*self.gamma
        p=prox_svd(x, 1, L1Ball(gamma=scale).prox )
        return p
    
    def __call__(self, x):
        nuclear_norm= NuclearNorm(gamma=self.gamma)(x)
        if nuclear_norm <= self.eta:
            return 0
        return np.inf
    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
