#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-20-2022)
 
 DEPENDENCIES: 
      - 'AbsValue.py' located in the folder 'scalar'
     - 'prox_svd.py' located in the folder 'utils'
     - 'NuclearNorm.py' located in the folder 'multi'
     -
     
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 
"""

import numpy as np 
from prox.scalar.AbsValue import AbsValue
from prox.multi.NuclearNorm import NuclearNorm
from prox.utils.prox_svd import prox_svd


class NuclearBlocks:
    """
    Computes the proximity operator of (gamma*tau)*f where f is the indicator of
    the function :
                               
                f(x) = gamma * \sum_{i= 1}^N \sum_{j=1}^M w_{i,j} \|X_{i,j}\|_N
                             
            where X = U*diag(s)*V.T \in R^{M*N} 
     INPUTS
    ========
     x         -  (M,N) -array_like ( representinf a M*N matrix )
     
     ind_r     - Vector partitioning the rows of X in groups EXAMPLE: ind_r=[1 2 2 3 3 1] 
     means the are max(ind_r)=3 blocks:  the first block contains the first and last rows of x,
     the 2nd block contains the 2nd and the third row, the third block contains the 4th and the 5th row.
     
     ind_c     - Vector partitioning the columns of X in groups 
     W         - positive ND array of size (max(ind_r) , max(ind_c) ) 
     gamma     - positive, scalar or ND array compatible with the size of 'x' [default: gamma=1]
     tau       - positive, scalar or ND array compatible with the size of 'x'[default: tau=1]
     """
     
    def __init__(self, W, ind_r, ind_c, gamma=1):
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        if np.any( W <= 0) or np.any( np.shape(W)!=(np.max(ind_r), np.max(ind_c))):
            raise Exception("'W' must be positive and of shape '( max(ind_r), max(ind_c))'")
        self.gamma = gamma
        self.W=W
        self.ind_r=np.array(ind_r)
        self.ind_c=np.array(ind_c)
        
    
        
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale= tau*self.gamma
        ind_r=self.ind_r
        ind_c=self.ind_c
        W=self.W
        
        m,n=np.shape(x)
        p=np.zeros((m,n))
        for i in range( max(ind_r)):
            for j in range( max(ind_c)):
              I = ind_r == i
              J = ind_c == j
              M = np.dot(I.T, J)
              A = x*M     #The elements that don't belong to the block are replaced by zeros
              p = p+ NuclearNorm(gamma=W[i,j]*scale).prox(A) #Update of p
          
        if np.size(ind_r)!=m or np.size(ind_c)!=n:
            raise ValueError("'ind_r' or 'ind_c' has not the right length")
        p=prox_svd(x, scale, AbsValue().prox )
        
        return p
    
    def __call__(self, x):
        ind_r=self.ind_r
        ind_c=self.ind_c
        W=self.W
        p=0
        for i in range( max(ind_r)):
          for j in range(max(ind_c)):
                I = ind_r == i
                J = ind_c == j
                M = M = np.dot(I.T, J)
                A = x*M       #The elements that don't belong to the block are replaced by zeros
                p = p + NuclearNorm(gamma=W[i,j]*self.gamma)(A) #Update of p  
        return p
        
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
