#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-20-2022)
 
 DEPENDENCIES: 
     - 'Max.py' located in the folder 'multi'
     - 'Linf.py' located in the folder 'multi'
     - 'prox_svd.py' located in the folder 'utils'
     - 'prox_svd.py' located in the folder 'utils'
     - 'L0Norm.py' located in the folder 'nonconvex'
     
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np 
from prox.utils.prox_svd import prox_svd
from prox.utils.fun_svd import fun_svd


class NuclearNormRidge:
    """
    Computes the proximity operator of (gamma*tau)*f where f is the function defined as follows:

                               
                         f(x) = (1/2)*||X||_F^2 + mu*R1(C) =(1/2)||S||_2^2 + mu*||s||_1
                             
            where X = U*diag(s)*V.T \in R^{M*N}  (Singular Value decomposition) 
            
            'gamma*tau' is the scale factor 
        
     INPUTS
    ========
     x         -  (M,N) -array_like ( representing a M*N matrix )
     mu        - positive scalar [default: mu=1]
     gamma     - positive scalar  [default: gamma=1]
     tau       - positive scalar [default: tau=1] [only for the method 'prox']
     """
     
    def __init__(self, mu=1, gamma=1):
        if np.any( gamma <=0 ) or np.size(gamma)>1:
            raise Exception("'gamma'  must be a strictly positive scalar")
        if np.any( mu <=0 ) or np.size(mu)>1:
            raise Exception("'mu'  must be a strictly positive scalar")
        self.gamma = gamma
        self.mu=mu
    
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale= tau*self.gamma
        mu=self.mu
        
        def prox_phi(s, gam_):
            a=-mu/(1+gam_)
            b=mu/(1+gam_)
            ss=s/(1+gam_)
            p=np.minimum(0, ss- a*gam_) + np.maximum(0, ss - b * gam_)
            return p
        return  prox_svd(x, scale, prox_phi)
    
    def __call__(self, x):
        def fun_phi(s):
            return 0.5*np.sum(s**2) + self.mu*np.sum( np.abs(s))
        return self.gamma*fun_svd(x, 1, fun_phi)
    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ) or np.size(tau)>1 :
           raise Exception("'mu'  must be a strictly positive scalar")
       if len( np.shape(x))!=2:
           raise ValueError("'x' must be an (M,N) -array_like ( representinf a M*N matrix )")
