#!#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-21-2022)
 
 DEPENDENCIES: 
     - 'Max.py' located in the folder 'multi'
     - 'Linf.py' located in the folder 'multi'
     - 'prox_svd.py' located in the folder 'utils'
     - 'prox_svd.py' located in the folder 'utils'
     - 'L0Norm.py' located in the folder 'nonconvex'
     
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

from prox.utils.prox_svd import prox_svd
from prox.utils.fun_svd import fun_svd
import numpy as np 

class PermutationInvariant:
    """
    Computes the proximity operator of (gamma*tau)*f where f is the function defined as follows:

                               
                            f(X) = phi(s)
                         
    where :
          * X = U*diag(s)*V.T \in R^{M*N}  the Singular Value decomposition of X
            
          * 'gamma*tau' is the scale factor 
        
     INPUTS
    ========
     x               -  (M,N) -array_like ( representing a M*N matrix
     gamma           - positive scalar  [default: gamma=1]
     tau             - positive scalar [default: tau=1] [only for the method 'prox']
     ClassPhi       - A class that has at least the two methods: 'ClassPhi.__call___(x)' and 
     'phi.prox(x, tau)'. >>> ClassPhi.__call__(x) must return phi(x) (i.e the value of the function phi evaluated
     at  x) and '>>> ClassPhi.prox(x, tau)' the proximity operator of tau*phi evaluated at x
     """
     
    def __init__(self, class_phi, gamma=1):
        if np.any( gamma <=0 ) or np.size(gamma)>1:
            raise Exception("'gamma'  must be a strictly positive scalar")
        self.gamma = gamma
        self.class_phi=class_phi
    
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale= tau*self.gamma
        class_phi = self.class_phi
        
        return prox_svd(x, scale, class_phi().prox )
    
    def __call__(self, x):
        return self.gamma*fun_svd(x, 1, self.class_phi().prox )
    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ) or np.size(tau)>1 :
           raise Exception("'mu'  must be a strictly positive scalar")
       if len( np.shape(x))!=2:
           raise ValueError("'x' must be an (M,N) -array_like ( representinf a M*N matrix )")
