#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-22-2022)
 
 DEPENDENCIES:
     -'newton.py' in the folder 'utils'
     
 Author  : Mbaye Diongue
        
 Copyright (C) 2022
     
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np
from prox.utils.newton import newton_

class PersSqrt:
    """
    This class computes the proximity operator of the function (gamma*tau)*f where f is
    defined as:
                                 
                
                /  -sqrt(xi^2 - ||y||_2^2)       if xi > 0 and ||y||_2^2 <= xi
      f(y,xi) = |  0                             if y = 0 and xi = 0
                \  +inf                          otherwise

        
            'gamma * tau' is the scale factor

            
     When the input 'y' is an array, the computation can vary as follows:
    - direction= None --> 'x' is processed as a single vector [DEFAULT]
             ( In this case 'gamma' must be a scalar )
    - direction >=0   --> 'y' is processed block-wise along the specified direction
             (in this case, 'xi' must be singleton along 'direction')
    INPUTS
    ========
    y          - scalar or ND array
    xi         - scalar or ND array compatible with the blocks of 'y'
    direction  - direction of block-wise processing [default: direction=None]
    gamma      - positive, scalar or ND array compatible with the blocks of 'y' [default: gamma=1]
    tau        - positive, scalar or ND array compatible with the blocks of 'y' [default: tau=1] [Only for the method 'prox']
    """
    
    def __init__(self, xi, gamma=1, direction=None):
        if np.any( gamma <= 0 ) :
            raise Exception("'gamma'  must be strictly positive")
        
        if direction==None and  np.size(gamma)>1 :
           raise Exception("a 'direction' must be specified when 'gamma' is not a scalar")
        if (direction is not None) and (direction < 0):
            direction = None
            
        if np.size(gamma)<=1:
            gamma=np.array([gamma])
        if np.size(xi)<=1:
            xi=np.array([xi])
            
        self.direction=direction
        self.gamma=gamma
        self.xi=xi
        
        
        
    def prox(self, y, tau=1):
        self._check( y, tau)
        scale=tau*self.gamma
     
        direction=self.direction
        xi=self.xi
        sz0=np.shape(y)
        if np.size(y)<=1:
            y=np.array([y])
            
        sz=np.shape(y)
        sz=np. array( sz , dtype=int)
        sz[direction]=1
 
        if np.size(scale)>1:
            scale=np.reshape(scale, sz)
        if np.size(xi)>1:
            xi=np.reshape(xi, sz)
            
        # linearize
        if direction==None:
            y=np.reshape(y, (-1))

        # init
        p = np.zeros(sz0)
        q = np.zeros(sz)
        
        # condition
        yy   = np.linalg.norm(y, ord=2, axis=direction).reshape(sz)
        mask = xi + np.sqrt(scale**2 + yy) > 0
        if np.size(scale) >1:
            scale = scale[mask]
        if np.size(xi) >1:
            xi = xi[mask]
            
        yy_mask = yy[mask]
        
        # compute t
        def fun_phi(t):
            return 2*scale*t + xi*t/np.sqrt(1+t**2) - yy_mask
        def der_fun_phi(t):
            return 2*scale+ (xi*np.sqrt(1+t**2) - xi*t**2/np.sqrt(1+t**2))/(1+t**2)
        
        # Finding a root with the Newton method
        t = newton_(fun_phi, fp=der_fun_phi, x0=yy_mask, low=0, high= np.inf)
        
        coef       = np.ones_like(yy)
    
        coef[mask] = scale*t/ yy_mask
        p          = y - y*coef
        
        qq = xi - scale*np.sqrt(1+t**2)

        q[mask] = qq
        
        # revert back
        p = np.reshape(p, sz0)
    
        return [p,q]
    
    def __call__(self, y):
        if np.size(y)<=1:
            y=np.array([y])
        xi=self.xi
        #evaluate the function
        yy = np.sum(y**2, self.direction)
        if np.any( xi<0) or np.any(  yy > xi**2 ):
            return np.inf
    
        p =-np.sqrt(xi**2 - yy);
        return np.sum(self.gamma*p)

    def _multiply_tuple(self, tup):
        res=1
        for i in tup:
            res=res*i
        return res
    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ) :
           raise Exception(name_name+" must be strictly positive")
       sz=np.shape(x)
       if len(sz)<=1:
           self.direction=None
       if len(sz)<=1 and ( np.size(tau)>1 or np.size(self.gamma)>1 or np.size(self.xi)>1):
           raise Exception("'tau', 'gamma' and 'xi' must be scalar when 'x' is one dimensional")
           
       if len(sz)> 1 and self.direction != None:
           sz=np. array( sz , dtype=int)
           sz[self.direction]=1
           if np.size(tau)>1 and ( self._multiply_tuple(sz) != np.size(tau)):
               raise Exception("The dimension of"+name_name+" is not compatible with the blocks of 'x'")
           if np.size(self.gamma)>1 and (self._multiply_tuple(sz) != np.size(self.gamma)):
               raise Exception("The dimension of 'gamma' is not compatible with the blocks of 'x'")
           if  np.size(self.xi)>1  and (self._multiply_tuple(sz) != np.size(self.xi)):
               raise Exception("The dimension of 'xi is not compatible with the blocks of 'x'")