#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-21-2022)
 
 Author  : Mbaye Diongue
        -'solver_cubic.py' in the folder 'utils'
        
 Copyright (C) 2022
     

 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np
from prox.utils.solver_cubic import solver_cubic

class PersSquare:
    """
    This class computes the proximity operator of the function (gamma*tau)*f where f is
    defined as:
                                 
                
                /   ||y||_2^2 / xi          if xi > 0
      f(y,xi) = |  0                        if y = 0 and xi = 0
                \  +inf                     otherwise

        
            'gamma * tau' is the scale factor

    When the input 'y' is an array, the computation can vary as follows:
    - direction= None --> 'x' is processed as a single vector [DEFAULT]
             ( In this case 'gamma' must be a scalar )
    - direction >=0   --> 'y' is processed block-wise along the specified direction
             (in this case, 'xi' must be singleton along 'direction')
    INPUTS
    ========
    y          - scalar or ND array
    xi         - scalar or ND array compatible with the blocks of 'y'
    direction  - direction of block-wise processing [default: direction=None]
    gamma      - positive, scalar or ND array compatible with the blocks of 'y' [default: gamma=1]
    tau        - positive, scalar or ND array compatible with the blocks of 'y' [default: tau=1] [Only for the method 'prox']
    """
    
    def __init__(self, xi, gamma=1, direction=None):
        if np.any( gamma <= 0 ) :
            raise Exception("'gamma'  must be strictly positive")
        
        if direction==None and  np.size(gamma)>1 :
           raise Exception("A 'direction' must be specified when 'gamma' is not a scalar")
        if (direction is not None) and (direction < 0):
            direction = None
            
        if np.size(gamma)<=1 :
            gamma=np.reshape( gamma, (-1))
        if np.size(xi)<=1:
            xi=np.reshape( xi, (-1))
            
        self.direction=direction
        self.gamma=gamma
        self.xi=xi
        self._check_done=False
        
        
        
    def prox(self, y, tau=1):
        self._check( y, tau)
        scale=tau*self.gamma
    
        direction=self.direction
        xi=self.xi
        sz0=np.shape(y)
        if np.size(y)<=1 or direction==None:
            y=np.reshape(y, (-1))
            
        sz=np.shape(y)
        sz=np. array( sz , dtype=int)
        sz[direction]=1
 
        if np.size(scale)>1:
            scale=np.reshape(scale, sz)
        if np.size(xi)>1:
            xi=np.reshape(xi, sz)
            
        # 3rd branch
        l2_x= np.sqrt( np.sum(y**2, axis=direction )).reshape(sz)
        t = solver_cubic(scale, 0, 4*xi+8*scale, - 8*l2_x)[0]
        #t= np.max( np.real(t), axis=direction).reshape(sz)
        pp = scale* t  / l2_x
        qq = scale* t**2 / 4

        # 1st branch
        mask = l2_x**2 <= -4*scale*xi
    
        pp[mask] = 1
        if np.size(xi)>1:
            qq[mask]= -xi[mask]
        else:
            qq[mask]=-xi
        
        # 2nd branch
        mask = np.logical_and(l2_x==0,  xi > 0).reshape(sz)
        pp[mask]= 1
        qq[mask] = 0

        # compute the prox
        p =  y - y*pp
        q = xi + qq

        # revert back
        p = np.reshape(p, sz0)
 
        return [p,q]
    
    def __call__(self, y):
        if np.size(y)<=1:
            y=np.array([y])
        
        if np.any( self.xi<0 ):
            return  np.inf
        l2_y2 = np.sum(y**2, axis=self.direction)
        p = l2_y2/ self.xi
        p[self.xi == 0] = 0
        
        return np.sum(self.gamma*p)

    def _multiply_tuple(self, tup):
        res=1
        for i in tup:
            res=res*i
        return res
    
    def _check(self, x, tau, name_name="'tau'"):
       if self._check_done: # the check is done onnly one time
           return 
       
       if np.any( tau <= 0 ) :
           raise Exception(name_name+" must be strictly positive")
       sz=np.shape(x)
       if len(sz)<=1:
           self.direction=None
       if len(sz)<=1 and ( np.size(tau)>1 or np.size(self.gamma)>1 or np.size(self.xi)>1):
           raise Exception("'tau', 'gamma' and 'xi' must be scalar when 'x' is one dimensional")
           
       if len(sz)> 1 and self.direction != None:
           sz=np. array( sz , dtype=int)
           sz[self.direction]=1
           if np.size(tau)>1 and ( self._multiply_tuple(sz) != np.size(tau)):
               raise Exception("The dimension of"+name_name+" is not compatible with the blocks of 'x'")
           if np.size(self.gamma)>1 and (self._multiply_tuple(sz) != np.size(self.gamma)):
               raise Exception("The dimension of 'gamma' is not compatible with the blocks of 'x'")
           if  np.size(self.xi)>1  and (self._multiply_tuple(sz) != np.size(self.xi)):
               raise Exception("The dimension of 'xi is not compatible with the blocks of 'x'")
               
       self._check_done=True