#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-22-2022)
 
 Author  : Mbaye Diongue
        
 Copyright (C) 2022
     
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np
from scipy.optimize import minimize

class PersVapnik:
    """
    This class computes the proximity operator of the function (gamma*tau)*f where f is
    defined as:
                                 
                
                 /   inf_{y \in R}( |y| + 1_{[-eps*xi, eps*ksi]}    if xi >= 0
      f(y,xi) = |    
                \    + inf                                         otherwise
                
                
        
            'gamma * tau' is the scale factor

            If the inputs are arrays, the result is computed element-wise for the proximity operator of f
            and element-wise sum for the function f
            
    INPUTS
    ========
    y          -  scalar or ND array
    xi         -  scalar or ND array with the same size as 'y'
    eps        - scalar or ND array with the same size as 'y'
    gamma      - positive, scalar or ND array compatible with the blocks of 'y' [default: gamma=1]
    tau        - positive, scalar or ND array compatible with the blocks of 'y' [default: tau=1] [Only for the method 'prox']
    """
    
    def __init__(self, eps, xi, gamma=1):
        if np.any( gamma <= 0 ) :
            raise Exception("'gamma'  must be strictly positive")
            
        if np.any( eps <= 0 ) :
            raise Exception("'eps'  must be strictly positive")
            
        if np.size(eps) !=np.size(xi) and  np.size(eps)>1 and np.size(xi) >1 :
            raise Exception("'eps'  and 'xi' must have the same size when they are scalars")
            
        if np.size(gamma)<=1:
            gamma=np.reshape( gamma, (-1))
        if np.size(xi)<=1:
            xi=np.reshape( xi, (-1))
        if np.size(eps)<=1:
            eps=np.reshape( eps, (-1))
            
        self.gamma=gamma
        self.eps=eps
        self.xi=xi
        
        
        
    def prox(self, y, tau=1):
        self._check( y, tau)
        scale=tau*self.gamma
        eps= self.eps
        xi=self.xi
        
        if np.size(y)<= 1 :
            y=np.reshape(y, (-1))
        sz= np.shape(y)
        if np.size(y)>1:
            if np.size(xi)<=1:
                xi= xi*np.ones(sz)
            if np.size(eps)<=1:
                eps= eps*np.ones(sz)
            if np.size(scale)<=1:
                scale= xi*np.ones(sz)
            
        abs_y=np.abs(y)
        sign_y=np.sign(y)

        # 1st branch
        p= np.zeros_like(1.0*y)
        q=  np.zeros_like(1.0*xi)
        
        # 2nd branch
        mask=( xi <= -scale*eps )*( abs_y > scale)
        p[mask]=y[mask]- scale[mask]*sign_y[mask]
        
        # 3rd branch
        mask= ( xi > -scale*eps )*( abs_y > eps*xi+scale*(1+eps**2))
        p[mask]= y[mask]- scale[mask]*sign_y[mask]
        q[mask]= eps[mask]+scale[mask]*eps[mask]
        
        # 4th branch
        mask=  ( abs_y > -xi/eps )*( abs_y > eps*xi+scale*(1+eps**2))*( abs_y <= eps*xi)
        q[mask]= ( xi[mask]+ eps[mask]*sign_y[mask])/(1+eps[mask]**2)
        p[mask]= eps[mask]*sign_y[mask]*q[mask]
        
        return [ p, q]
    
    def __call__(self, y):
        eps=self.eps
        xi=self.xi
        xi=np.reshape( xi, (-1))
        eps=np.reshape( eps, (-1))
        y=np.reshape( y, (-1))
        g= np.reshape( self.gamma, (-1))
        def fun_phi(t):
            mask= (y-t>= -eps*xi )*(y-t <= eps*xi )
            return np.sum( np.abs(y)+ mask)
        
        res= minimize(fun_phi, y)

        return np.sum( fun_phi(g*res.x))

    def _multiply_tuple(self, tup):
        res=1
        for i in tup:
            res=res*i
        return res
    
    def _check(self, y, tau, name_name="'tau'"):
       if np.any( tau <= 0 ) or ( np.size(tau)>1 and np.size(tau) != np.size(y)):
           raise Exception(name_name+" must be positive scalar or an positive ND array with the same size as 'y'")
       if ( np.size(self.xi)>1 and np.size(self.xi) != np.size(y)):
           raise Exception("'xi' must be scalar or an ND array with the same size as 'y'")
       if ( np.size(self.eps)>1 and np.size(self.eps) != np.size(y)):
           raise Exception("'eps' must be scalar or an ND array with the same size as 'y'")