#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-13-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""
            
import numpy as np

class Vapnik :
    """
    Computes the proximity operator of the function
    
                        f(x) = gamma *tau* max( ||x||_2 -epsilon, 0)
                        
                        
                        'gamma * tau' is the scale factor
            
    When the input 'x' is an array, the computation can vary as follows:
    - direction = None --> 'x' is processed as a single vector [DEFAULT]
       In this case 'gamma', 'epsilon' and 'tau' must be a scalars
    - direction >=0   --> 'x' is processed block-wise along the specified direction
       (in this case, 'gamma', 'epsilon' and 'tau' must be singletons along 'direction')
    
    INPUTS
    ========
     x         - ND array
     epsilon   - positive, scalar or ND array compatible with the blocks of 'x' 
     gamma     - positive, scalar or ND array compatible with the blocks of 'x' (default value : gamma=1)
     tau       - positive, scalar or ND array compatible with the blocks of 'x' (default value : tau=1)
     direction - integer, direction of block-wise processing
    """
    def __init__(self, epsilon, gamma=1, direction=None):
        if np.any( gamma <= 0 ):
            raise ValueError("'gamma' (or all of its components if it is an array) must be strictly positive")
        if np.any( epsilon <= 0 ):
            raise ValueError("'epsilon' (or all of its components if it is an array) must be strictly positive")
        if (direction is not None) and (direction < 0):
            direction = None
        self.gamma = gamma
        self.epsilon= epsilon
        self.direction=direction
        
    def prox(self, x, tau=1):
        scale= tau*self.gamma
        self._check( x, scale)
        direction=self.direction
        epsilon=self.epsilon
        
        if np.size(x)<=1:
            x=np.reshape(x, (-1))
            
        sz=np.shape(x)
        sz=np. array( sz , dtype=int)
        sz[direction]=1
        if np.size(scale)>1:
            scale=np.reshape(scale, sz)
        if np.size(epsilon)>1:
            epsilon=np.reshape(epsilon, sz) 
            
        # preliminaries
        l2_x = np.linalg.norm(x, ord=2, axis=direction ).reshape(sz)
        
        # First branch
        p= 1.0*x
        
        # 2nd branch
        mask_ones=np.ones(np.shape(x), dtype=bool)
        mask= mask_ones*( l2_x-epsilon > 0 )*( l2_x-epsilon <= scale )
        p[mask]= (epsilon/l2_x*x)[mask]
        
        # 3rd branch
        mask= mask_ones*( l2_x-epsilon > scale)
        p[mask]=((1 - scale/ l2_x) * x)[mask]
        
        return p
        
    def __call__(self, x):
        if np.size(x)<=1:
            x=np.reshape(x, (-1))
        l2_x = np.linalg.norm(x, ord=2, axis=self.direction)
        return self.gamma*np.maximum(l2_x - self.epsilon, 0)
     
    def _multiply_tuple(self, tup):
        res=1
        for i in tup:
            res=res*i
        return res
    
    def _check(self, x, scale, name_name="'tau'"):
       if np.any( scale <= 0 ):
           raise ValueError(name_name+" (or all of its components if it is an array) must be strictly positive")
           
       if (self.direction==None) and np.size(scale)>1:
            raise ValueError("'gamma' and 'tau' must be scalars when the argument 'direction' is equal to 'None'")
            
       sz=np.shape(x)
       if len(sz)<=1:
           self.direction=None
       if np.size(scale)<=1:
           return 
       
       if len(sz)<=1 and np.size(scale)>1 :
           raise ValueError("'tau' and 'gamma' must be scalar when 'x' is one dimensional")
           
       if len(sz)> 1 and self.direction != None:
           sz=np. array( sz , dtype=int)
           sz[self.direction]=1
           if np.size(scale)>1 and ( self._multiply_tuple(sz) != np.size(scale)):
               raise ValueError("The dimension of 'tau' or 'gamma' is not compatible with the blocks of 'x'")
           
           
    

        