#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-18-2022)
 
  DEPENDENCIES:
     'newton.py' in the folder 'utils'
     
 Author  : Mbaye Diongue

 Copyright (C) 2022

 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np
from prox.utils.newton import newton_

class BurgCauchy:
    """
    This class computes the proximity operator of the function (gamma*tau)*f where f is defined as:

                       /   -log(x)+ w*log( delta + x^2 )   if x>0
               f(x) = |
                      \   +inf                      otherwise
               
                with w>0 and delta>0
    'gamma * tau' is the scale factor
    
    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
    x     - ND array
    gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    tau   - positive, scalar or ND array with the same size as 'x' [default: tau=1]
    delta   - positive, scalar or ND array with the same size as 'x' [default: delata=1e-10]
    w     - positive, scalar or ND array with the same size as 'x' [default: w=1]
    
    Note: When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>BurgCauchy(delta=0.01, w=2)(x) will
    return a scalar even if x is a vector:
        
    >>> BurgCauchy(delta=0.01, w=2)(np.array([10, 20, 30]))
    26.098816
    
    But as expected, >>>BurgCauchy(delta=0.01, w=2).prox(x) 
    will return a vector with the same size as x:
        
    >>> BurgCauchy(delta=0.01, w=2).prox(np.array([10, 20, 30]))
    array([ 9.69046116, 19.84886296, 29.89966593])
    """
    
    def __init__(self, w=1, delta=1e-10, gamma=1):
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        if np.any( delta <= 0 ):
            raise Exception("'delta' (or all of its components if it is an array) must be strictly positive")
        if np.any( w <= 0 ):
            raise Exception("'w' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
        self.delta= delta
        self.w=w
        
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale=tau*self.gamma
        delta=self.delta
        w=self.w
        
        def f(t):
            return t**4 - x*t**3 + (delta+2*scale*w-scale)*t**2 -delta*x*t -delta*scale
        def df(t):
            return 4*t**3 -3*x*t**2+2*(delta+2*scale*x-scale)*t -delta*x
         
        def f_prox(t):
            res=np.inf*np.ones_like(t)
            mask_= t>0
            w_=w
            delta_=delta
            if np.size(delta)>1:
                delta_=delta[mask_[0]]
            if np.size(w)>1:
                w_=w[mask_[0]] 
            res[mask_]=-np.log(t[mask_])+w_*np.log( delta_+ t[mask_]**2)
            return 0.5*np.abs(x-t)**2  + scale*res
        
        # use the Newton's method (starting 3 times to try to find different roots)
        abs_x=np.abs(x)
        p_init = np.maximum(1,abs_x)
        low=1*1e-10
        p= newton_(f, df, p_init, low=low, high= 10*abs_x)
        
        p_init = np.maximum(10*abs_x,abs_x)
        p2= newton_(f, df, p_init, low=1, high= np.inf)
        
        p_init = np.maximum(1+20*abs_x,abs_x)
        p3= newton_(f, df, p_init, low=10*abs_x, high= np.inf)
        
        mask = f(p2)<f(p) 
        
        p[mask]=p2[mask]
        
        mask =f(p3)<f(p)
        p[mask]=p3[mask]
        
        return p
    
    def __call__(self, x):
        if np.any(x<=0):
            return np.inf
        return np.sum( -np.log(x)+self.w*np.log( self.delta+ x**2) )
    
    def fun(self, x):
        res=np.inf*np.ones_like(1.0*x)
        mask= x>0
        w=self.w
        delta=self.delta
        if np.size(delta)>1:
            delta=delta[mask]
        if np.size(w)>1:
            w=w[mask]    
        
        res[mask] = -np.log(x[mask])+w*np.log( delta+ x[mask]**2)
        return self.gamma*res
    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
    
        
    