#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-18-2022)
 
 Author  : Mbaye Diongue

 Copyright (C) 2022
     

 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np

class CEL0:
    """
    This class computes the proximity operator of the function (gamma*tau)*f where f is the 
    Continuous Exact L0 (CEL0) penalty  defined as:
        
        f(x)= lamb - 0.5*ai^2*(|x| - sqrt(2*lamb)/ai)^2*1_{|x| <= sqrt(2*lamb)/ai}  
        
        
            'gamma * tau' is the scale factor
            
    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
    x     - ND array
    gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    a     - positive, scalar or ND array with the same size as 'x' 
    lamb  - positive, scalar or ND array with the same size as 'x' 
   
    
     Note: When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>CEL0(lamb=2, a=3)(x) will
    return a scalar even if x is a vector:
        
    >>> CEL0(lamb=2, a=3)(np.array([-1, 2, 3]))
    6.0
    
    But as expected, >>>CEL0(lamb=2, a=3).prox(x)
    will return a vector with the same size as x:
        
    >>> CEL0(lamb=2, a=3).prox(np.array([-1, 2, 3]))
    array([0, 0, 3])
    """
    
    def __init__(self, lamb, a, gamma=1):
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        if np.any( lamb <= 0 ):
            raise Exception("'lamb' (or all of its components if it is an array) must be strictly positive")
        if np.any( a <= 0 ):
            raise Exception("'a' (or all of its components if it is an array) must be strictly positive ")
        self.lamb = lamb
        self.gamma=gamma
        self.a=a
        
    
        
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale=tau*self.gamma
        a=self.a
        lamb=self.lamb
        abs_x=np.abs(x)
        sign_x=np.sign(x)
        
        # 1st case:  a**2*scale >=1
        mask1=abs_x > np.sqrt(2*scale*lamb)
        p = 1.0*x*mask1
        
        if np.size(a**2*scale)<=1 and (a**2*scale >=1):
            return np.reshape(p, np.shape(x))
        
        # 2nd case: a**2*scale < 1
        mask= a**2*scale < 1
        if np.size(lamb)>1:
            lamb=lamb[mask]
        if np.size(a)>1:
            a=a[mask]
        if np.size(scale)>1:
            scale=scale[mask]

        p[mask]=sign_x[mask]*np.minimum( abs_x[mask], np.maximum( abs_x[mask]-np.sqrt(2*lamb)*scale*a, 0)/(1-a**2*scale))
    
        return np.reshape(p, np.shape(x))
    
    def __call__(self, x):
        self._check_size(x)
        a=self.a
        lamb=self.lamb
        if np.size(x)<=1:
            x=np.array([x])
            
        xx= np.abs(x)- np.sqrt(2*lamb)/a
        mask= xx<=0
        p= lamb- 0.5*(a*xx)**2*mask 
    
        return np.sum(self.gamma*p)

    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.size(tau) != np.size(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
           
    def _check_size(self, x):
        sz_x=np.size(x)
        if ( np.size(self.lamb) > 1)  and (np.size(self.lamb) != sz_x):
           raise Exception("'lamb' must be either scalar or the same size as 'x'")
        if ( np.size(self.a) > 1)  and (np.size(self.a) != sz_x ):
           raise Exception("'a' must be either scalar or the same size as 'x'")
        if ( np.size(self.gamma) > 1)  and (np.size(self.gamma) != sz_x ):
           raise Exception("'gamma' must be either scalar or the same size as 'x'")
    
        
    