#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-18-2022)
 
 Author  : Mbaye Diongue

 Copyright (C) 2022
     

 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np

class CapL1:
    """
    This class computes the proximity operator of the function (gamma*tau)*f where f is the 
    capped-L1 (or truncated L1) penalty defined as:

               f(x) = min( 1, theta*|x| )
                      
    'gamma * tau' is the scale factor
    
    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
    x     - ND array
    gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    tau   - positive, scalar or ND array with the same size as 'x' [default: tau=1]
    theta  - positive, scalar or ND array with the same size as 'x' 
   
    
     Note: When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>CapL1(theta=0.1)(x) will
    return a scalar even if x is a vector:
        
    >>>  CapL1(theta=0.1)(np.array([0, 2, -3]))
    0.5
    
    But as expected, >>>CapL1(theta=0.1).prox(x) 
    will return a vector with the same size as x:
        
    >>> CapL1(theta=0.1).prox(np.array([0, 2, -3]))
    array([ 0. ,  1.9, -2.9])
    """
    
    def __init__(self, w=1, theta=1, gamma=1):
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        if np.any( theta <= 0 ):
            raise Exception("'theta' (or all of its components if it is an array) must be strictly positive")
        
        self.gamma = gamma
        self.theta = theta
    
        
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale=tau*self.gamma
        theta=self.theta
        
        p1= np.sign(x)*(np.abs(x)-scale*theta)*(np.abs(x) > scale*theta) * (np.abs(x)<(1/theta +scale*theta/2))* (scale*theta**2 <2)
        p2= x * (scale*theta**2 <2)* (np.abs(x)>=(1/theta +scale*theta/2)) 
        p3= x * (np.abs(x)>np.sqrt(2*scale)) * (scale*theta**2 >= 2)       
    
        p=p1+p2+p3
        return p
    
    def __call__(self, x):
        res= np.minimum(1, self.theta*np.abs(x))
        return np.sum(self.gamma*res)

    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.size(tau) != np.size(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
           
    def _check_size(self, x):
        sz_x=np.size(x)
        if ( np.size(self.w) > 1)  and (np.size(self.w) != sz_x):
           raise Exception("'w' must be either scalar or the same size as 'x'")
        if ( np.size(self.delta) > 1)  and (np.size(self.delta) != sz_x ):
           raise Exception("'delta' must be either scalar or the same size as 'x'")
        if ( np.size(self.gamma) > 1)  and (np.size(self.gamma) != sz_x ):
           raise Exception("'gamma' must be either scalar or the same size as 'x'")
    
        
    