#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 07-18-2022
 
  DEPENDENCIES:
     'solver_cubic.py' in the folder 'utils'
     
 Author  : Mbaye Diongue

 Copyright (C) 2022

 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np
from prox.utils.solver_cubic import solver_cubic

class Cauchy:
    """
    This class computes the proximity operator of the function (gamma*tau)*f where f is defined as:

        
               f(x) = log( delta + x^2) 
               

    'gamma * tau' is the scale factor
    
    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
    x     - ND array
    gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    tau   - positive, scalar or ND array with the same size as 'x' [default: tau=1]
    delta   - positive, scalar or ND array with the same size as 'x' [default: delata=1e-10]
    
    Note: When calling the function (and not the proximity operator) the result 
    is computed element-wise SUM. So the command >>>Cauchy(delta=0.01, gamma=1)(x) will return
    a scalar even if x is a vector:
        
    >>> Cauchy(0.01, 1)(np.array([-1., 2., 3.]))
    3.597076643791892
    
    But as expected, >>>Cauchy(delta=0.01, gamma=1).prox(x) will
    return a vector with the same size as x:
        
    >>> Cauchy(0.01, 1).prox(np.array([-1., 2., 3.]))
    array([-0.00498744,  0.01005025,  0.0152717 ])
    """
    
    def __init__(self, delta=1e-10, gamma=1):
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        if np.any( delta <= 0 ):
            raise Exception("'delta' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
        self.delta= delta
        
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale=tau*self.gamma
        delta=self.delta
        
        # Solve the cubic equation
        p, p2, p3= solver_cubic(1, -x, delta+2*scale, -x*delta)
        
        def f(t):
            return 0.5*np.abs(x-t)**2 + scale*np.log(delta+t**2)
        
        mask =np.logical_and( np.isreal(p2), f(np.real(p2))<f(p) )
        p[mask]=np.real(p2[mask])
        
        mask =np.logical_and( np.isreal(p2), f(np.real(p3))<f(p) )
        p[mask]=np.real(p3[mask])
        
        return p.reshape(np.shape(x))
    
    def __call__(self, x):
        return np.sum(self.gamma*np.log( self.delta+ x**2))
    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
    
        
    