#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-18-2022)
 
 Author  : Mbaye Diongue

 Copyright (C) 2022
     

 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np

class ConicL0:
    """
    This class computes the proximity operator of the function:
       
                   f(x,d) = gamma  \ell_0(x) + \iota_{S}(x,d) 
       
    with S = {x \in C^N, d \in C^N, s.t. (\forall n \in {1,...,N}) 
              \exist delta_n \in [-Delta_n,Delta_n] s.t. d_n = \delta_n x_n 
              }
    where \Delta \in [0,+\infty)^N
       
    When the inputs '(x,d)' are arrays, the outputs '(p,q)' are computed element-wise.
       
     INPUTS
    ========
     x     - ND array complex valued
     d     - ND array complex valued with the same size as 'x'
     gamma - positive, scalar 
     Delta - real positive,  scalar or ND array with the same size as 'x'
    
    Note: When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>ConicL0(Delta=2, d=2*x)(x) will
    return a scalar even if x is a vector:
    >>> x=np.array([ 1+1j, 4+2J])
    >>> x
    array([1.+1.j, 4.+2.j])
    >>> ConicL0(Delta=3, d=2*x)(x)
    2
    
    But as expected, >>> ConicL0(Delta=3, d=2*x).prox(x)
    will return two vectors with the same size as x:
        
    >>> px=ConicL0(Delta=3, d=2*x).prox(x)
    [array([1.+1.j, 4.+2.j]), array([2.+2.j, 8.+4.j])]
    """
    
    def __init__(self, Delta, d, gamma=1):
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        if np.any( Delta < 0 ):
            raise Exception("'lamb' (or all of its components if it is an array) must be positive or null")
        
        self.Delta=Delta
        self.gamma=gamma
        self.d=d
        
        
    def prox(self, x, tau=1):
        self._check( x, tau)
        self._check_size(x)
        
        scale=tau*self.gamma
        Delta= self.Delta
        d=self.d
        
        p = np.zeros_like( 1.0*x)
        q = np.zeros_like( 1.0*x)
        
        z = np.real(x*np.conj(d))
        abs_d2=np.abs(d)**2
        abs_x2=np.abs(x)**2
        
        eta = np.sqrt(((abs_d2 - abs_x2))**2 + 4*z**2)
        
        deltahat = Delta*np.ones(np.shape(x))
        mask= z != 0
        deltahat[mask] = (np.minimum((eta + abs_d2 - abs_x2)/(2*np.abs(z)), Delta)*np.sign(z) )[mask]
        deltahat[(z == 0 )*( abs_x2 >= abs_d2) ]= 0
         
        seuilmod = ( np.abs(deltahat*x - d)**2)/(1 + deltahat**2) + 2*scale
        indseuil = abs_x2 + abs_d2 >= seuilmod
         
        upmod = (x + deltahat*d)/(1 + deltahat**2)
        
        p[indseuil] = upmod[indseuil]
        q[indseuil] = upmod[indseuil]*deltahat[indseuil]
        return [p,q]
    
    def __call__(self, x):
        self._check_size(x)
        d=self.d
        Delta=self.Delta
        if np.size(x)<=1:
            x=np.array([x])
        if np.size(d)<=1:
            d=np.array([d])
        TOL=1e-10
        # check the constraints
        if abs(np.sum( np.imag( x*np.conj(d) )))>TOL:  # x and d does not have the same arg
            return np.inf
        
        # x and d have the same arg
        mask_x= x==0
        mask_d= d==0
        if np.any( mask_x != mask_d):
            return np.inf
        # evaluate the l0 function
        p = np.sum( self.gamma* (x != 0) )
        
        if np.size(mask_x)==np.size(x):
            return p
        
        mask= x!=0
        if np.any( np.abs(x[mask])/np.abs(d[mask]) > Delta): #delta not in [-Delta;Delta]
            return np.inf
        
        return p

    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.size(tau) != np.size(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
           
    def _check_size(self, x):
        sz_x=np.size(x)
        if np.size(self.d)>1 and (np.size(self.d) != sz_x):
           raise Exception("'d' must have the same size as 'x'")
        if ( np.size(self.Delta) > 1)  and (np.size(self.Delta) != sz_x ):
           raise Exception("'Delta' must be either scalar or the same size as 'x'")
        if ( np.size(self.gamma) > 1)  and (np.size(self.gamma) != sz_x ):
           raise Exception("'gamma' must be either scalar or the same size as 'x'")
    
        
    