#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-18-2022)
 
  DEPENDENCIES:
     'newton.py' in the folder 'utils'
     
 Author  : Mbaye Diongue

 Copyright (C) 2022
     

 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np
from prox.utils.newton import newton_

class EntropyCauchy:
    """
    This class computes the proximity operator of the function (gamma*tau)*f where f is defined as:

                       /   xlog(x)+ w*log( delta + x^2 )    if x>0
               f(x) = |    w*log(delta)                      if x=0
                      \   +inf                               otherwise
               
                with w>0 and delta>0
    'gamma * tau' is the scale factor
    
    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
    x     - ND array
    gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    tau   - positive, scalar or ND array with the same size as 'x' [default: tau=1]
    delta   - positive, scalar or ND array with the same size as 'x' [default: delata=1e-10]
    w     - positive, scalar or ND array with the same size as 'x' [default: w=1]
    
     Note: When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>EntropyCauchy(delta=0.01, w=2)(x) will
    return a scalar even if x is a vector:
        
    >>> EntropyCauchy(delta=0.01, w=2)(np.array([10, 20, 30]))
    219.77474905520018
    
    But as expected, >>>EntropyCauchy(delta=0.01, w=2).prox(x) 
    will return a vector with the same size as x:
        
    >>> EntropyCauchy(delta=0.01, w=2).prox(np.array([10, 20, 30]))
    array([ 7.44509635, 16.92485178, 26.55998001])
    """
    
    def __init__(self, w=1, delta=1e-10, gamma=1):
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        if np.any( delta <= 0 ):
            raise Exception("'delta' (or all of its components if it is an array) must be strictly positive")
        if np.any( w <= 0 ):
            raise Exception("'w' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
        self.delta= delta
        self.w=w
        
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale=tau*self.gamma
        delta=self.delta
        w=self.w
        
    
        def f(t):
            return t**3   +   (delta-x)*t**2  + (delta+2*scale*w)*t +  scale*delta*np.log(t) + scale*t**2*np.log(t)+ delta*(scale-x)
        def df(t):
            return 3*t**2 + 2*(scale-x) * t   + (delta+2*scale*w)   +   scale*delta/t        + scale*t*(2*np.log(t)+1)
         
        def f_prox(t):
            res=np.inf*np.ones_like(t)
            mask_= t>0
            w_=w
            delta_=delta
            if np.size(delta)>1:
                delta_=delta[mask_[0]]
            if np.size(w)>1:
                w_=w[mask_[0]] 
            res[mask_]=t[mask_]*np.log(t[mask_])+w_*np.log( delta_+ t[mask_]**2)
            
            return 0.5*np.abs(x-t)**2  + scale*res
        
        abs_x=np.abs(x)
        p_init = np.maximum(1,abs_x)
        low=1*1e-10
        p= newton_(f, df, p_init, low=low, high= 10*abs_x)


        return np.reshape( p, np.shape(x))
    
    def __call__(self, x):
        self._check_size(x)
        if np.any( x<0):
            return np.inf
        
        res=np.zeros_like(1.0*x)
        mask= x>0
        w=self.w
        delta=self.delta
        if np.size(delta)>1:
            delta=delta[mask]
        if np.size(w)>1:
            w=w[mask]    
        
        res= x[mask]*np.log(x[mask])+w*np.log( delta+ x[mask]**2)
        
        mask= x==0
        if np.size(self.delta)>1:
            delta=self.delta[mask]
        if np.size(self.w)>1:
            w=self.w[mask] 
        res[mask]= w*np.log(delta)
        return np.sum(self.gamma*res)
        
    def fun(self, x):
        res=np.inf*np.ones_like(1.0*x)
        mask= x>0
        w=self.w
        delta=self.delta
        if np.size(delta)>1:
            delta=delta[mask]
        if np.size(w)>1:
            w=w[mask]    
        
        res[mask] = x[mask]*np.log(x[mask])+w*np.log( delta+ x[mask]**2)
        
        mask= x==0
        if np.size(self.delta)>1:
            delta=self.delta[mask]
        if np.size(self.w)>1:
            w=self.w[mask] 
        res[mask]= w*np.log(delta)
        return self.gamma*res
    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.size(tau) != np.size(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
           
    def _check_size(self, x):
        sz_x=np.size(x)
        if ( np.size(self.w) > 1)  and (np.size(self.w) != sz_x):
           raise Exception("'w' must be either scalar or the same size as 'x'")
        if ( np.size(self.delta) > 1)  and (np.size(self.delta) != sz_x ):
           raise Exception("'delta' must be either scalar or the same size as 'x'")
        if ( np.size(self.gamma) > 1)  and (np.size(self.gamma) != sz_x ):
           raise Exception("'gamma' must be either scalar or the same size as 'x'")
    
        
    