#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-18-2022)
      DEPENDENCIES:
     'Lambert_W.py' in the folder 'utils'
     
     
 Author  : Mbaye Diongue

 Copyright (C) 2022
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np
from prox.utils.lambert_W import lambert_W

class EntropyL0:
    """
    This class computes the proximity operator of the function (gamma*tau)*f where f is defined as:

                                       /   x*log(x) + w      if x>0
               f(x) = min{ x^2, w}  = |    0                 if x=0
                                      \    +inf             otherwise

    'gamma * tau' is the scale factor
    
    When the input 'x' is an array, the output is computed element-wise SUM.

     INPUTS
    ========
    x     - ND array
    w     - positive, scalar or ND array with the same size as 'x' 
    gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    tau   - positive, scalar or ND array with the same size as 'x' [default: tau=1] [Only for the method 'prox']
   
    
    Note: When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>EntropyL0(w=2)(x) will
    return a scalar even if x is a vector:
        
    >>> EntropyL0(w=2)(np.array([1, 2, 3]))
    10.68213122712422
    
    But as expected, >>>EntropyL0(w=2).prox(x) 
    will return a vector with the same size as x:
        
    >>> EntropyL0(w=2).prox(np.array([1, 2, 3]))
    array([0.       , 0.       , 1.5571456])
    """
    
    def __init__(self, w, gamma=1):
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        if np.any( w <= 0 ):
            raise Exception("'w' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
        self.w=w
        
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale=tau*self.gamma
        if np.size(x)<=1:
            x=np.array([x])
            
        p= np.zeros_like(1.0*x)
        # we use a taylor approximation to avoid divergence when the input is too big
        u= np.log(1/scale)+x/scale-1
        mask = u>100
        
        p[mask]= u[mask]- u[mask]/(1+u[mask])*np.log(u[mask])
        
        # we use the Lambert_W function otherwise
        mask=np.logical_not(mask)
        if np.size(scale)>1:
            scale=scale[mask]
        p[mask]=scale*lambert_W( 1/scale*np.exp(x[mask]/scale-1))
        
        # 3rd branch
        scale=tau*self.gamma
        mask= p**2+2*scale*p < 2*self.w*scale
        # 1st and 2nd branch
        p[mask]=0
        return np.reshape( p, np.shape(x))
         
    def __call__(self, x):
        if np.size(x)<=1:
            x=np.array([x])
        if np.any(x<0):
            return np.inf
        mask= x>0
        w=self.w
        if np.size(w)>1:
            w=w[mask]
        g=self.gamma
        if np.size(g)>1:
            g=g[mask]
        p=g*(x[mask]*np.log(x[mask]) + w)
        return np.sum(p)
    
    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
    
        
    