#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-18-2022)
 
 Author  : Mbaye Diongue

 Copyright (C) 2022
     

 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np

class L0L2Norm:
    """
    This class computes the proximity operator of the function (gamma*tau)*f where f is
    defined as:
                                   / 0      if ||x||_2=0
               f(x)= ||x||_2^0 =  |
                                  \  1      otherwise
        
        
            'gamma * tau' is the scale factor
            
    When the input 'x' is an array, the computation can vary as follows:
    - direction = None --> 'x' is processed as a single vector [DEFAULT]
       In this case 'gamma' must be a scalar
    - direction >=0   --> 'x' is processed block-wise along the specified direction

    INPUTS
    ========
    x     - ND array
    gamma - positive, scalar or ND array compatible with the blocks of 'x' [default: gamma=1]
    tau   - positive, scalar or ND array compatible with the blocks of 'x' [default: tau=1]
    direction   - direction of block-wise processing [default: direction=None]
    """
    
    def __init__(self, gamma=1, direction=None):
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        if direction==None and np.size(gamma)>1 :
           raise Exception("A 'direction' must be specified when 'gamma' is not a scalar")
        if (direction is not None) and (direction < 0):
            direction = None
        self.direction=direction
        self.gamma=gamma
        
        
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale=tau*self.gamma
        direction=self.direction
        sz0=np.shape(x)
        if direction==None:
            x=np.reshape(x,(-1))
            
        sz=np.shape(x)
        direction=self.direction
        sz=np. array( sz , dtype=int)
        sz[direction]=1
        if np.size(scale)>1:
            scale=np.reshape(scale, sz)
            
        p=np.sum(x**2, axis=direction).reshape(sz)
    
        mask= p>=2*scale
        p=x*mask

        return np.reshape(p,sz0 )
    
    def __call__(self, x):
        direction=self.direction
        if np.size(x)<=1:
            x=np.array([x])
        if direction==None:
            x=np.reshape(x,(-1))
        
        p= np.linalg.norm(x, ord=2, axis=direction)
        
        return np.sum(self.gamma*np.array(p!=0))
  
    def _multiply_tuple(self, tup):
        res=1
        for i in tup:
            res=res*i
        return res
    
    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       sz=np.shape(x)
       if len(sz)<=1:
           self.direction=None
       if len(sz)<=1 and ( np.size(tau)>1 or np.size(self.gamma)>1):
           raise Exception("'tau' and 'gamma' must be scalar when 'x' is one dimensional")
           
       if len(sz)> 1 and self.direction != None:
           sz=np. array( sz , dtype=int)
           sz[self.direction]=1
           if np.size(tau)>1 and ( self._multiply_tuple(sz) != np.size(tau)):
               raise Exception("The dimension of"+name_name+" is not compatible with the blocks of 'x'")
           if np.size(self.gamma)>1 and (self._multiply_tuple(sz) != np.size(self.gamma)):
               raise Exception("The dimension of 'gamma' is not compatible with the blocks of 'x'")

        
    