#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-18-2022)
     
 Author  : Mbaye Diongue

 Copyright (C) 2022
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np

class L0Norm:
    """
    This class computes the proximity operator of the function (gamma*tau)*f where f is defined as:

                                 / 0   if x=0
               f(x) = ||x||_0 = |
                                \  1   otherwise

    'gamma * tau' is the scale factor
    
    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
    x     - ND array
    gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    tau   - positive, scalar or ND array with the same size as 'x' [default: tau=1]
    
    Note: When calling the function (and not the proximity operator) the result 
    is computed element-wise SUM. So the command >>>L0Norm(gamma=1)(x) will return
    a scalar even if x is a vector:
    >>> L0Norm(gamma=1)(np.array([-1.,2.,3]))
    3
    
    But as expected, >>>L0Norm(gamma=1).prox(x) will
    return a vector with the same size as x:
        
    >>> L0Norm(gamma=1).prox(np.array([-1.,2.,3]))    
    array([-0.,  2.,  3.])

    """
    
    def __init__(self, gamma=1):
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
        
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale=tau*self.gamma
        mask= x**2>=2*scale
        return x*mask
         
    def __call__(self, x):
        TOL=1e-12
        return np.sum( self.gamma*(1*(np.abs(x)>TOL) ))
    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
    
        
    