#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-18-2022)
 
 Author  : Mbaye Diongue

 Copyright (C) 2022
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np

class LogSum:
    """
    This class computes the proximity operator of the function (gamma*tau)*f where f is defined as:

        
               f(x) = log( delta + |x|) 
               

    'gamma * tau' is the scale factor
    
    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
    x     - ND array
    gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    tau   - positive, scalar or ND array with the same size as 'x' [default: tau=1]
    delta - positive, scalar or ND array with the same size as 'x' [default: delata=1e-10]
    
    Note: When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>LogSum(delta=0.01, gamma=1)(x) will
    return a scalar even if x is a vector:
    >>> LogSum(delta=0.01, gamma=1)(np.array([-1, 2, 3]))
    1.8100251316849367
    
    But as expected, >>>LogSum(delta=2, gamma=1).prox(x) 
    will return a vector with the same size as x:
    >>> LogSum(delta=0.01, gamma=1).prox(np.array([-1, 2, 3]))
    array([0.        , 1.09512492, 2.6197333 ])
    """
    
    def __init__(self, delta=1e-10, gamma=1):
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        if np.any( delta <= 0 ):
            raise Exception("'delta' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
        self.delta= delta
        
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale=tau*self.gamma
        delta=self.delta
        
        # 1st branch
        p=np.zeros_like(1.0*x)
        
        # 2ND branch
        mask= np.abs(x) > np.sqrt(4*scale)-delta
        abs_x=np.abs(x[mask])
        if np.size(delta)>1:
            delta=delta[mask]
        if np.size(scale)>1:
            scale= scale[mask]
    
        p[mask]=np.sign(x[mask])*0.5*( abs_x-delta +np.sqrt( (abs_x+delta)**2-4*scale))
        return np.reshape(p, np.shape(x))
    
    def __call__(self, x):
        return np.sum( self.gamma*np.log( self.delta+ np.abs(x)))
    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
    
        
    