#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-18-2022)
 
 Author  : Mbaye Diongue

 Copyright (C) 2022
     

 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np

class MCP:
    """
    This class computes the proximity operator of the function (gamma*tau)*f where f is the 
    Minimax Concave Penalty (MCP) defined as:

        
                  |  lamb^2*beta/2                    if |x| > beta*lamb
        f(x) =   |  
                 |  lamb*|x| - x^2/(2*beta)            if |x| <= a*lamb
        
            'gamma * tau' is the scale factor
            
    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
    x     - ND array
    gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    beta  - positive, scalar or ND array with the same size as 'x' 
    lamb  - positive, scalar or ND array with the same size as 'x' 
   
    
     Note: When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>MCP(lamb=2)(x) will
    return a scalar even if x is a vector:
        
    >>> MCP(lamb=2)(np.array([-1, 2, 3]))
    5.5
    
    But as expected, >>>MCP(lamb=2).prox(x)
    will return a vector with the same size as x:
        
    >>> MCP(lamb=2).prox(np.array([-1, 2, 3]))
    array([0, 0, 3]
    """
    
    def __init__(self, lamb, beta=1, gamma=1):
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        if np.any( lamb <= 0 ):
            raise Exception("'lamb' (or all of its components if it is an array) must be strictly positive")
            
        if np.any( beta <= 0 ):
            raise Exception("'beta' (or all of its components if it is an array) must be strictly positive")
        
        self.lamb = lamb
        self.gamma=gamma
        self.beta=beta
        
    
        
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale=tau*self.gamma
        beta=self.beta
        lamb=self.lamb
        abs_x=np.abs(x)
        if np.size(x)<=1:
            x=np.array([x])

        mask1=np.abs(x) > np.sqrt(scale*beta)*lamb
        p =   x * mask1
        if np.size( beta-scale)<=1 and (beta<=scale):
            return np.reshape(p, np.shape(x))
        
        mask= beta>scale
        if np.size(lamb)>1:
            lamb=lamb[mask]
        if np.size(beta)>1:
            beta=beta[mask]
        if np.size(scale)>1:
            scale=scale[mask]
        p[mask]=np.sign(x[mask])* np.minimum(beta/(beta-scale)* np.maximum(abs_x[mask]-lamb*scale,0)  , abs_x[mask])
        
        return np.reshape(p, np.shape(x))
    
    def __call__(self, x):
        self._check_size(x)
        beta=self.beta
        lamb=self.lamb
        if np.size(x)<=1:
            x=np.array([x])
            
        # 1st branch
        abs_x=np.abs(1.0*x)
        p= (lamb*abs_x-0.5*x**2/beta)
        
        # 2nd branch
        mask=abs_x > beta*lamb
        if np.size(lamb)>1:
            lamb=lamb[mask]
        if np.size(beta)>1:
            beta=beta[mask]
        p[mask]=0.5*beta*lamb**2
        
        return np.sum(self.gamma*p)

    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.size(tau) != np.size(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
           
    def _check_size(self, x):
        sz_x=np.size(x)
        if ( np.size(self.lamb) > 1)  and (np.size(self.lamb) != sz_x):
           raise Exception("'lamb' must be either scalar or the same size as 'x'")
        if ( np.size(self.beta) > 1)  and (np.size(self.beta) != sz_x ):
           raise Exception("'beta' must be either scalar or the same size as 'x'")
        if ( np.size(self.gamma) > 1)  and (np.size(self.gamma) != sz_x ):
           raise Exception("'gamma' must be either scalar or the same size as 'x'")
    
        
    