#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-19-2022)
 
 Author  : Mbaye Diongue

 Copyright (C) 2022
     

 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np

class TruncatedNorm:
    """
    This class computes the proximity operator of the function (gamma*tau)*f where f is
    defined as:
                                 
                    f(x)= min{ ||x||_2^2, w}
             
        
            'gamma * tau' is the scale factor

            
    When the input 'x' is an array, the computation can vary as follows:
    - dir = None --> 'x' is processed as a single vector [DEFAULT]
       In this case 'gamma' must be a scalar
    - direction >=0   --> 'x' is processed block-wise along the specified direction
    
    INPUTS
    ========
    x     - one dimentional ND array
    gamma - positive, scalar [default: gamma=1]
    tau   - positive, scalar[default: tau=1]
    w     - positive, scalar or ND array compatible with the blocks of 'x' 
   direction   - direction of block-wise processing [default: direction=None]
    """
    
    def __init__(self, w, gamma=1, direction=None):
        if np.any( gamma <= 0 ) :
            raise Exception("'gamma'  must be strictly positive")
        if np.any(w <= 0 ) :
            raise Exception("'w'  must be strictly positive")
        if direction==None and ( np.size(gamma)>1 or np.size(w)>1):
           raise Exception("A 'direction' must be specified when 'w' or 'gamma' is not a scalar")
        if (direction is not None) and (direction < 0):
            direction = None
        self.direction=direction
        self.gamma=gamma
        self.w=w
        
        
    def prox(self, x, tau=1):
        self._check( x, tau)
        scale=tau*self.gamma
        w=self.w
        sz0=np.shape(x)
        if np.size(x)<=1:
            x=np.array([x])
            
        sz=np.shape(x)
        direction=self.direction
        sz=np. array( sz , dtype=int)
        sz[direction]=1
        
        if np.size(w)>1:
            w=np.reshape(w, sz)
        if np.size(scale)>1:
            scale=np.reshape(scale, sz)
            
        mask=np.sum(x**2, axis=direction).reshape(sz) < w*(1+2*scale)
    
        p=x/(1+2*scale*mask)
 
        return np.reshape(p, sz0)
    
    def __call__(self, x):
        p= np.minimum( np.sum(x**2), self.w)
        
        return np.sum(self.gamma*p)

    def _multiply_tuple(self, tup):
        res=1
        for i in tup:
            res=res*i
        return res
    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ) or np.size(tau)>1:
           raise Exception(name_name+" must be strictly positive scalar")
       sz=np.shape(x)
       if len(sz)<=1:
           self.direction=None
       if len(sz)<=1 and ( np.size(tau)>1 or np.size(self.gamma)>1 or np.size(self.w)>1):
           raise Exception("'tau', 'gamma' and 'w' must be scalar when 'x' is one dimensional")
           
       if len(sz)> 1 and self.direction != None:
           sz=np. array( sz , dtype=int)
           sz[self.direction]=1
           if np.size(tau)>1 and ( self._multiply_tuple(sz) != np.size(tau)):
               raise Exception("The dimension of"+name_name+" is not compatible with the blocks of 'x'")
           if np.size(self.gamma)>1 and (self._multiply_tuple(sz) != np.size(self.gamma)):
               raise Exception("The dimension of 'gamma' is not compatible with the blocks of 'x'")
           if  np.size(self.w)>1  and (self._multiply_tuple(sz) != np.size(self.w)):
               raise Exception("The dimension of 'w' is not compatible with the blocks of 'x'")