#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Original Version : 1.0 (01-02-2019), updated  (06-09-2022)
 
 Author  : MOHAMED KERROUMI

 Copyright (C) 2019
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).

"""
  
import numpy as np


class ArctanActi:
    """
    This method computes the proximity operator of the arc tangent activation function

                 / -(2/pi)*log(cos(pi*x/2))-(x^2)/2   if |x| < 1
         f(x) = |
                 \ +Inf                 otherwise

        
    When the input 'x' is an array, the output is computed element-wise.

    
     INPUTS
    ========
     x     - scalar or ND array
     
    ========
    Examples
    ========
    
    Evaluate the 'direct' function :
    >>> ArctanActi()(-0.5)
    array([0.0956356])
    >>>ArctanActi()([-2., -0.5, 0, 0.25, 3])
    array([   inf,  0.0956356 , -0. ,  0.01915347, inf])
    
    Compute the proximity operator at a given point:
    >>> ArctanActi().prox( 3)
    0.7951672353008666
    >>> ArctanActi().prox([ -3., 0.5, 6.])
    array([-0.79516724,  0.29516724,  0.89486309])
    """

    def __init__(self):
        self.gamma = 1

    def prox(self, x):
        return 2 /np.pi* np.arctan(x)

    def __call__(self, x):
        if not isinstance(x, np.ndarray):
            x= np.array(x)

        if np.size(x)<=1:
            x=np.reshape(x, (-1))
           
        p = np.zeros(np.size(x))
        mask=abs(x) < 1
        p[mask] = (
            -(2 / np.pi) * np.log(np.cos(np.pi * x[mask] / 2)) - x[mask]** 2 / 2
        )
        p[abs(x) >= 1] = np.inf
        return p 

      
        