#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (27-04-2017), updated (08-06-2022)
 
 Author  : Giovanni Chierchia
 Updated by Mbaye Diongue

 Copyright (C) 2017
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """
 
import numpy as np


class Bathtub :
    """
    This class computes the proximity operator of the function

            f(x) = gamma *tau* max{|x|-w,0}     with w > 0

    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
     x     - scalar or ND array
     w     - positive, scalar or ND array with the same size as 'x'
     gamma - positive, scalar or ND array with the same size as 'x' [default value: gamma=1]
     tau   - positive, scalar or ND array with the same size as 'x' [default value : tau=1] [only for the method 'prox']
     
     ========
     Examples
     ========
     
     Evaluate the 'direct' function :
     >>> Bathtub()(5)
     4
     >>> Bathtub(2, gamma=3)([-4., 2., 3., 5])
     array([6., 0., 3., 9.])
     
     Compute the proximity operator :
         
     >>> Bathtub(2).prox( 3)
     array([2])
     >>> Bathtub(3).prox([ -3., 1., 6.])
     array([-3.,  1.,  5.])
     
     Use a scale factor 'gamma'>0 to commute the proximity operator of gamma*f
     
     >>> Bathtub(1, gamma=2).prox([ -3., 1., 6.])
     array([-1.,  1.,  4.])
      
     Or equivalently with the parameter 'tau':
         
     >>> Bathtub(1).prox([ -3., 1., 6.], tau=2)
     array([-1.,  1.,  4.])
     
     When 'gamma' has the same size as 'x' the computation is done element-wise:
     >>> Bathtub(1, gamma=[1, 2, 3]).prox([ -3., 1., 6.])
     array([-2.,  1.,  3.])
    """
        
    def __init__(self, w, gamma=1):
        if np.size(w)>1 and (not isinstance(w, np.ndarray)):
            w=np.array(w)
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
        if np.any( w <= 0 ):
            raise Exception("'w' must be positive")
        if np.size(w)>1 and (not isinstance(w, np.ndarray)):
            w=np.array(w)
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        self.w = w
        self.gamma=gamma
    
    def prox(self, x, tau=1):

        if not isinstance(x, np.ndarray):
            x= np.array(x)
            tau=np.array(tau)
        if np.size(x)<=1:
            x=np.reshape(x, (-1))
            
        self.__check(x)
        self._check_gamma( x, tau)
        scale=self.gamma*tau
        
        # preliminaries
        abs_x = np.abs(x)
        sign_x = np.sign(x)

        # preliminaries
        abs_x  = np.abs(x)
        sign_x = np.sign(x)

        # 2nd branch
        p = self.w * sign_x

        # 1st branch
        mask = abs_x < self.w
        p[mask] = x[mask]

        # 3rd branch
        mask = abs_x > self.w + scale
        p3 = x - scale * sign_x
        p[mask] = p3[mask]
        return p

    def __call__(self, x):
        self.__check(x)
        return self.gamma*np.maximum(0,  np.abs(x) - self.w) 
        
        
    def __check(self, x):
        if (not np.isscalar(self.w)) and (np.shape(self.w) != np.shape(x) and (len(self.w)>1)):
            raise Exception("'w' (or all its components if it is an array) must be either scalar or the same size as 'x'")
    
        
    def _check_gamma(self, x, tau, name_name="'tau'"):
          
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
           
       if (np.size(self.gamma) > 1)  and (np.shape(self.gamma) != np.shape(x) ):
           raise Exception("gamma' must be either scalar or the same size as 'x'")  
       