#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-07-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""
import numpy as np


class BentIdentity :
    """
    This method computes the proximity operator of the function

                 /   x/2 - ln( x+ 1/2)/4     if  |x| <= 1
          f(x)= |                                   
                \  +INF                      otherwise

    When the input 'x' is an array, the output is computed element-wise.

    
     INPUTS
    ========
     x     - scalr or ND array
     
     
     ========
     Examples
     ========
     
    Evaluate the 'direct' function :
        
    >>> BentIdentity()(-0.5)
    array([inf])
    >>> BentIdentity()([-2., -0.5, 0, 0.25, 3])
    array([   inf,  inf, 0.1732868 , 0.19692052, 1.18680926])
    
    Compute the proximity operator at a given point :
        
    >>> BentIdentity().prox(3)
    2.58113883008419
    >>> BentIdentity().prox([ -3., 0.5, 6.])
    array([-0.41886117,  0.30901699,  5.54138127])
    """

    def __init__(self):
        self.gamma = 1

    def prox(self, x):
        if not isinstance(x, np.ndarray):
            x= np.array(x)
        p=(x + np.sqrt(x ** 2 + 1) - 1) / 2
        return p

    def __call__(self, x):
        if not isinstance(x, np.ndarray):
            x= np.array(x)
        if np.size(x)<=1:
            x=np.reshape(x, (-1))
        
        result=np.inf*np.ones(np.shape(x))
        mask= x > -0.5
        result[mask]=  x[mask]/2-np.log( x[mask] + 1/2)/4
    
        return result
      
        