#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-09-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).

"""

import numpy as np

class Berhu:
    """
    Computes the proximity operator of the function (gamma*tau)*f where f is
    defined as:

                  / ( x^2 + w^2 ) / ( 2* W)     if |x| > w
           f(x)= | 
                 \  |w|                         otherwise
                 
                 with w>0
            
    'gamma*tau' is the scale factor
    
    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
     x     - ND array
     w     - positive, scalar or ND array with the same size as 'x'

     ========
     Examples
     ========
     
     Evaluate the 'direct' function:
         
     >>> Bh=Berhu(w=2)
     >>> Bh(3)
      array([3.25])
     >> Bh([-1, 2, 3])
     array([2.  , 2.  , 3.25])
     
     Compute the proximity operator at a given point :
         
     >>> Bh.prox( 3)
     array([2])
     >>> Bh.prox([ -3., 1., 6.])
     array([-2.,  0.,  4.])
    """
    
    def __init__(self, w):
        if np.size(w)>1 and (not isinstance(w, np.ndarray)):
            w=np.array(w)
        if np.any( w <= 0 ):
            raise Exception("'w' (or all of its components if it is an array) must be strictly positive")
        self.w=w
        self.gamma =1
        

    def prox(self, x):
        if not isinstance(x, np.ndarray):
            x= np.array(x)
        if np.size(x)<=1:
            x= np.reshape(x, (-1))
            
        self._check(x, self.w)
        # preliminaries
        abs_x = np.abs(x)
        sign_x = np.sign(x)
        w=self.w

        # 2nd brabch
        p = x - sign_x

        # 1st branch
        mask = abs_x > w + 1
        
        if ( np.size(w) > 1) :
            p[mask]= w[mask]*x[mask]/( 1+ w[mask])
        else:
            p[mask] = w * x[mask]/ (w + 1)
        
        # third branch
        mask2 = abs_x <= 1
        p[mask2] = 0

        return p

        
    def __call__(self, x):
        if not isinstance(x, np.ndarray):
            x= np.array(x)
        if np.size(x)<=1:
            x= np.reshape(x, (-1))
        self._check(x, self.w)
        abs_x = np.abs(x)
        w=self.w
        
        result= np.abs(w)*np.ones(np.shape(x))
        mask = abs_x > w 

        if ( np.size(w) > 1) :
            result[mask]= (w[mask]**2+x[mask]**2)/(2*w[mask])
            
        else:
            result[mask]= (w**2+x[mask]**2)/(2*w)
        
        return self.gamma*result
    
    def _check(self, x,w):
       if ( np.size(w) > 1)  and (np.size(w) != np.size(x) ):
           raise Exception("'w' must be either scalar or the same size as 'x'")