#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-23-2022)
     
 Author  : Mbaye Diongue
        
 Copyright (C) 2022
     
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np

class BregAbsEntropy:
    r"""
    This class computes the proximity operator prox_{scale*f}^{phi} where f and phi are
    defined as follows:
                                 
              f(x)=  | x - delta |
             
                      /  u*log(u)     if u>0
             phi(u)= |   0            if u=0
                     \   +inf         otherwise
                     
                     
            scale='gamma * tau' is the scale factor

            When the inputs are arrays, the outputs are computed element-wise
            
    INPUTS
    ========
    x          - scalar or ND array 
    delta      - scalar or ND array with the same size as 'x' 
    gamma      - scalar or ND array compatible with the blocks of 'y' [default: gamma=1]
    tau        - scalar or ND array compatible with the blocks of 'y' [default: tau=1] [Only for the method 'prox']
    
    ========
     Examples
     ========
     
     Evaluate the  function f :
         
     >>> BregAbsEntropy(delta=0.001)(4)
     3.999
     >>> BregAbsEntropy(delta=0.001)( [-2, 3, 4] )
     array([2.001, 2.999, 3.999])
     
     Evaluate The function  phi
     
     >> BregAbsEntropy(delta=0.001).phi([-2, 3, 4])
     array([       inf, 3.29583687, 5.54517744])
     
     Compute the proximity operator at a given point :
         
     >>> BregAbsEntropy(delta=0.1).prox( [-1, 2, 4] )
     array([-2.71828183,  0.73575888,  1.47151776])
     >>> BregAbsEntropy(delta=0.1, gamma=2).prox( [-1, 2, 4] )
     array([-7.3890561 ,  0.27067057,  0.54134113]
    """
    
    def __init__(self, delta, gamma=1):
        if np.size(delta)>1 and (not isinstance(delta, np.ndarray)):
            delta=np.array(delta)
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.any( gamma <= 0 ) :
            raise ValueError("'gamma'  must be strictly positive")
        if np.any( delta <= 0 ) :
            raise ValueError("'delta'  must be strictly positive")
        self.gamma=gamma
        self.delta=delta
    
        
    def prox(self, x, tau=1):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)): 
            tau=np.array(tau)
       
        scale=tau*self.gamma
        delta=self.delta
        self._check(x, scale)
        
        # scalar-like inputs handling 
        if np.size(x) <=1:
            x= np.reshape( x, (-1))
       
        # 3rd branch
        p = x*np.exp(-scale)
        
        # 2nd branch
        mask1= x>=delta*np.exp(-scale)
        mask = mask1* (x <= delta*np.exp(scale))
        dd=delta
        if np.size(dd)>1:
            dd=dd[mask]
        p[mask]= dd
        
        # 1st branch
        mask=np.logical_not( mask1)
        gg=scale
        if np.size(gg)>1:
            gg=gg[mask]
        p[mask] = x[mask]*np.exp(gg)
        
        return p

    def __call__(self, x):
       if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
       return self.gamma*np.abs(x-self.delta)
   
    def phi(self, u):
        if np.size(u)>1 and (not isinstance(u, np.ndarray)):
            u=np.array(u)
        if np.size(u)<=1:
            u=np.reshape(u, (-1))
        res=np.zeros(np.shape(u))
        mask=u>0
        res[mask]= u[mask]*np.log(u[mask])
        res[u<0]=np.inf
        
        return res
   
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ) or ( np.size(tau)>1 and np.size(tau) != np.size(x)):
           raise ValueError("'tau' and 'gamma' must be positive scalars or positive ND arrays with the same size as 'x'")