#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-23-2022)
 
 Author  : Mbaye Diongue
        
 Copyright (C) 2022
     
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np

class BregBoltzShannon2:
    """
    This class computes the proximity operator prox_{scale*f}^{phi} where f and phi are
    defined as follows:
        
        
                        /  x*log(x) - w*x      if x>0     
               f(x)=   |
                       \  +inf                 otherwise
               
               
             
                         /  u*log(u) + (1-u)*log( 1-u)    if u \in ]0, 1[
              phi(u) =  |   0                             if u \in {0,1}
                        \   +inf                          otherwise
                     
            scale='gamma * tau' is the scale factor

            When the inputs are arrays, the outputs are computed element-wise
            
    INPUTS
    ========
    x          - scalar or ND array 
    w          - scalar or ND array with the same size as 'x'  [default: w=1]
    gamma      - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    tau        - positive, scalar or ND array with the same size as 'x' [default: tau=1] [Only for the method 'prox']
    
    
    =======
    Examples
    ========
    
    Evaluate the function f:
        
    >>> BregBoltzShannon2()(1)
     array([-1.])
     >>> BregBoltzShannon2(w=0)( [-2, 3, 4, np.e] )
     array([     inf, 3.29583687, 5.54517744, 2.71828183])
     
     Evaluate the function  phi
     
     >> BregBoltzShannon2(w=0).phi( [-2., 0.3, 0.125, 1.] )
      array([        inf, -0.6108643 , -0.37677016,  0.        ])
      
     Compute the proximity operator at a given point :
         
     >>> BregBoltzShannon2(w=2).prox(  [-2, 3, 4, np.e] )
     array([0.44986882, 0.98232606, 0.99335135, 0.97683598])
     >>> BregBoltzShannon(2, gamma=2).prox( [-2, 3, 4] )
     array([  1.        , 148.4131591 , 403.42879349])
    """
    
    def __init__(self, w=1, gamma=1):
        if np.size(w)>1 and (not isinstance(w, np.ndarray)):
            w=np.array(w)
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.any( gamma <= 0 ) :
            raise ValueError("'gamma'  must be strictly positive")
        self.gamma=gamma
        self.w=w
        
    def prox(self, x, tau=1):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)): 
            tau=np.array(tau)
        scale=tau*self.gamma
        self._check(x, scale)
        w=self.w
        xx= np.exp(x+w-1)
        p=-0.5*xx + np.sqrt( 0.25*xx**2 + xx)
        return p

    def __call__(self, x):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        self._check(x, self.gamma)
        if np.size(x)<=1:
            x=np.reshape(x, (-1))
        res=np.zeros(np.shape(x))
        mask=x>0
        w=self.w
        if np.size(w) >1:
            w=w[mask]
        res[mask]= x[mask]*np.log(x[mask])-w*x[mask]
        res[x<=0]=np.inf
        return self.gamma*res
   
    def phi(self, u):
        if np.size(u)>1 and (not isinstance(u , np.ndarray)):
            u=np.array(u)
        if np.size(u)<=1:
            u=np.reshape(u, (-1))
        res=np.zeros(np.shape(u))
        mask=(u>0) *(u<1)
        res[mask]= u[mask]*np.log(u[mask]) + (1-u[mask])*np.log(1-u[mask])
        res[ np.logical_or(u>1, u<0) ]=np.inf
        return res
   
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ) or ( np.size(tau)>1 and np.size(tau) != np.size(x)):
           raise ValueError("'tau' and 'gamma' must be positive scalars or positive ND arrays with the same size as 'x'")
    
       if ( np.size(self.w)>1 and np.size(self.w) != np.size(x)):
           raise ValueError("'w' must be a scalar or an ND arrays with the same size as 'x'")

