#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-23-2022)
     
 Author  : Mbaye Diongue
        
 Copyright (C) 2022
     
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np

class BregExp:
    """
    This class computes the proximity operator prox_{scale*f}^{phi} where f and phi are
    defined as follows:
                                 
              f(x)=  exp(x)
             
              phi(u)= exp(u)
                     
            scale='gamma * tau' is the scale factor

            When the inputs are arrays, the outputs are computed element-wise
            
    INPUTS
    ========
    x          - scalar or ND array 
    gamma      - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    tau        - positive, scalar or ND array with the same size as 'x' [default: tau=1] [Only for the method 'prox']
    
    =======
    Examples
    ========
    
    Evaluate the function f:
        
    >>> BregExp()( np.log(2) )
     array([2.])
     >>> BregExp()( [-2, 3, 4, np.log(3)] )
     array([ 0.13533528, 20.08553692, 54.59815003,  3.        ])
     
     Evaluate The function  phi
     
     >> BregExp().phi( [-2, 3, 4, np.log(3)] )
     array([ 0.13533528, 20.08553692, 54.59815003,  3.        ])
      
     Compute the proximity operator at a given point :
         
     >>> BregExp().prox(  [-2, 3, 4, np.log(3)] )
      array([-2.69314718,  2.30685282,  3.30685282,  0.40546511])
     >>> BregExp(gamma=2).prox(  [-2, 3, 4, np.log(3)] )
     array([-3.09861229,  1.90138771,  2.90138771,  0.        ])
    """
    
    def __init__(self, gamma=1):
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.any( gamma <= 0 ) :
            raise ValueError("'gamma'  must be strictly positive")
        self.gamma=gamma

    def prox(self, x, tau=1):
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)): 
            tau=np.array(tau)
        scale=tau*self.gamma
        self._check(x, scale)

        return x- np.log(1+scale)

    def __call__(self, x):
       return self.gamma*np.exp(x)
   
    def phi(self, u):
        return np.exp(u)
   
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ) or ( np.size(tau)>1 and np.size(tau) != np.size(x)):
           raise ValueError("'tau' and 'gamma' must be positive scalars or positive ND arrays with the same size as 'x'")