#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-23-2022)
 
 DEPENDENCIES:
     
     -'lambert_W.py' in the folder 'utils'
     
 Author  : Mbaye Diongue
        
 Copyright (C) 2022
     
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np
from prox.utils.lambert_W import lambert_W

class BregLpNorm:
    """
    This class computes the proximity operator prox_{scale*f}^{phi} where f and phi are
    defined as follows:
                                 
               f(x)= (1/p)*|x|**p
               
             
                         /  u*log(u)-u     if u>0
              phi(u) =  |
                        \   +inf           otherwise
                     
            scale='gamma * tau' is the scale factor

            When the inputs are arrays, the outputs are computed element-wise
            
    INPUTS
    ========
    x          - scalar or ND array 
    p          - scalar >=1 [default: p=1]
    gamma      - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    tau        - positive, scalar or ND array with the same size as 'x' [default: tau=1] [Only for the method 'prox']
    
    
    =======
    Examples
    ========
    
    Evaluate the function f:
        
     >>> BregLpNorm(p=2)( [1, 2, 3] )
     array([0.5, 2. , 4.5])
     >>> BregLpNorm()( [-2, 3, 4]) # default p=1
     array([2., 3., 4.])
     
     Evaluate The function  phi
     
     >> BregLpNorm().phi( [-2, 3, 4] )
     array([       inf, 0.29583687, 1.54517744])
      
     Compute the proximity operator at a given point :
         
     >>> BregLpNorm().prox(  [-2, 3, 4, np.e] )
     array([ 0.04978707,  7.3890561 , 20.08553692,  5.57494152])
     >>> BregLpNorm(2).prox(  [-2, 3, 4, np.e] )
     array([0.12002824, 2.20794003, 2.92627106, 2.01677976])
     """
    
    def __init__(self, p=1, gamma=1):
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.any( gamma <= 0 ) :
            raise ValueError("'gamma'  must be strictly positive")
        if np.any( p < 1 ) or np.size(p)>1:
            raise ValueError("'p'  must be a scalar grater or equal to 1") 
        self.gamma=gamma
        self.p=p

    
        
    def prox(self, x, tau=1):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)): 
            tau=np.array(tau)
        scale=tau*self.gamma
        self._check(x, scale)
        p=self.p
        
        if p==1:
            return np.exp(x-scale)
    
        t=np.zeros(np.shape(x))

        u= np.log(scale*(p-1))+ (p-1)*x
        # approximation of the lambert_W function when the inputs is to big
        mask = u>100 
        t[mask]= u[mask]- u[mask]/(1+u[mask])*np.log(u[mask])
        # use directly the Lambert_W function otherwise
        mask= np.logical_not(mask)
        t[mask]=lambert_W( np.exp(u[mask]))
        
        return ( t/(scale*(p-1)) )**(1/(p-1))

    def __call__(self, x):
       return self.gamma*(1/self.p)*(np.abs(x))**self.p
   
    def phi(self, u):
        if np.size(u)>1 and (not isinstance(u , np.ndarray)):
            u=np.array(u)
        if np.size(u)<=1:
            u=np.reshape(u, (-1))
        res=np.zeros(np.shape(u))
        mask=u>0
        res[mask]= u[mask]*np.log(u[mask])-u[mask]
        res[u<=0]=np.inf
        return res
   
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ) or ( np.size(tau)>1 and np.size(tau) != np.size(x)):
           raise ValueError("'tau' and 'gamma' must be positive scalars or positive ND arrays with the same size as 'x'")
