#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 ( 06-23-2022)
     
 Author  : Mbaye Diongue
        
 Copyright (C) 2022
     
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np

class BregSquareLog:
    """
    This class computes the proximity operator prox_{scale*f}^{phi} where f and phi are
    defined as follows:
                                 
               f(x)= (1/2)*x^2
             
                         /  - log(u)      if u>0
              phi(u) =  |
                        \   +inf           otherwise
                     
            scale='gamma * tau' is the scale factor

            When the inputs are arrays, the outputs are computed element-wise
            
    INPUTS
    ========
    x          - scalar or ND array 
    gamma      - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    tau        - positive, scalar or ND array with the same size as 'x' [default: tau=1] [Only for the method 'prox']
    
    
    =======
    Examples
    ========
    
    Evaluate the function f:
    
     >>> BregSquareLog()( -3 )
     4.5
     >>> BregSquareLog()( [1, 2, 3] )
     array([0.5, 2. , 4.5])
     
     Evaluate The function  phi
     
     >> BregSquareLog().phi( [-2, 3, 4] )
     array([        inf, -1.09861229, -1.38629436])
      
     Compute the proximity operator at a given point :
         
     >>> BregSquareLog().prox(  [-2, 3, 4, np.e] )
     array([-0.78077641,  0.84712709,  0.88278222,  0.83283647])
     >>> BregSquareLog().prox(  [-2, 3, 4, np.log(2)] )
     array([-0.78077641,  0.84712709,  0.88278222,  0.51167407])
     """
    
    def __init__(self, gamma=1):
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.any( gamma <= 0 ) :
            raise Exception("'gamma'  must be strictly positive")
        self.gamma=gamma

    
        
    def prox(self, x, tau=1):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)): 
            tau=np.array(tau)
        scale=tau*self.gamma
        self._check(x, scale)
        x[x==0]= 1e-16
        return (np.sqrt( 1 + 4*scale*x**2)-1)/(2*scale*x)

    def __call__(self, x):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        return self.gamma*0.5*x**2
   
    def phi(self, u):
        if np.size(u)>1 and (not isinstance(u , np.ndarray)):
            u=np.array(u)
        if np.size(u)<=1:
            u=np.reshape(u, (-1))
        res=np.zeros(np.shape(u))
        mask=u>0
        res[mask]= -np.log(u[mask])
        res[u<=0]=np.inf
        return res
   
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ) or ( np.size(tau)>1 and np.size(tau) != np.size(x)):
           raise Exception("'tau' and 'gamma' must be positive scalars or positive ND arrays with the same size as 'x'")
