#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-08-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np

class ELUacti:
    """
    Computes the proximity operator of the function

          
                          /  0                              if  x>=0
                   f(x)= |  (x+w)*ln( (x+w)/w) - x -x^2/2   Iif -w < x < 0
                         \  w - w^2 /2                      if  x=-w  
                         \  +INF                            otherwise
        
    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
     x     - sclar ND array
     w     - w>=1, scalar or ND array with the same size as 'x'
     
    =======
    Examples
    ========
    
    Evaluate the function f:
    
     >>> ELUacti(w=4)( -3 )
     -2.886294361119891
     >>> ELUacti(2)( [1, -1.5, 3] )
      array([ 0.        , -0.31814718,  0.        ])
   
     Compute the proximity operator at a given point :
    
     >>> ELUacti(2).prox( 3)
     array([3.])
     >>> ELUacti(2).prox(  [-2, 3, 4, np.e] )
     array([-1.72932943,  3.        ,  4.        ,  2.71828183]
    """
    
    def __init__(self, w):
        if np.size(w)>1 and (not isinstance(w, np.ndarray)):
            w=np.array(w)
        self.gamma = 1
        if np.any( w < 1 ):
            raise Exception("'w' (or all of its components if it is an array) must be greater or equal than 1")
        self.w=w
        

    def prox(self, x):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(x)<=1:
            x=np.reshape(x, (-1))
        w=self.w
        p = w * (np.exp(x) - 1)
        mask = x >= 0
        p[mask] = x[mask]

        return p

        
    def __call__(self, x):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        w=self.w
        if np.size(x) <= 1:
            x=np.reshape(x, (-1))
            
        result= np.zeros(np.shape(x))
        mask = np.logical_and( x>-w, x<0)
        ww=w
        if ( np.size(w) > 1) :
            ww= w[mask]
            
        result[mask]= (x[mask]+ww)*np.log( (x[mask]+ww)/ww ) - x[mask] -x[mask]**2/2
        
        mask2= x==-w
        if ( np.size(w) > 1) :
            ww= w[mask2]
        result[ mask2]= ww-ww**2/2
        result[ x< -w]=np.inf
        return self.gamma*result
        