#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-09-2022)
 
 DEPENDENCIES :
     
     lambert_W.py - located in the folder 'utils'
     
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).

"""

import numpy as np
from prox.utils.lambert_W import lambert_W


class Entropy:
    """
    This method computes the proximity operator of the 'Entropy' function

                      /  gamma*tau* xlog(x)            if  x> 0
               f(x)= |  0                              if x =0
                     \  +INF                           otherwise

        
    When the input 'x' is an array, the output is computed element-wise.

     'tau*gamma' is the scale factor
     
      INPUTS
     ========
      x     - scalar or ND array-like
      gamma - positive, scalar or ND array with the same size as 'x' [default value: gamma=1]
      tau   - positive, scalar or ND array with the same size as 'x' [default: tau=1] [Only for the method 'prox']
      
    =======
    Examples
    ========
 
     Evaluate the 'direct' function :
    
     >>> Entropy()( np.e )
     2.718281828459045
     >>> Entropy()( [-1, 0, 3, np.e] )  # compute the result element-wise for vector inputs
     array([       inf, 0.    , 3.29583687, 2.71828183])
     
     Compute the proximity operator at a given point :
         
     >>> Entropy().prox(  [-2, 3, 4, np.e] )
     array([0.04747849, 1.5571456 , 2.20794003, 1.38940572])
     
     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function 'gamma*f' :
    
     >>> Entropy(gamma=2).prox( [-2, 3, 4, np.e ] )
     array([0.12700814, 1.        , 1.37015388, 0.90903194])
     
     This is equivalent to use the parameter 'tau' of the method 'prox' to compute the 
      proximity operator of  the function 'tau*f' :
     
     >>> Entropy().prox( [-2, 3, 4, np.e ] , tau=2)
     array([0.12700814, 1.        , 1.37015388, 0.90903194])
    """

    def __init__(self, gamma=1):
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
        
    def prox(self, x, tau=1):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)):
            tau=np.array(tau)
        if np.size(x) <= 1:
            x=np.reshape(x, (-1))
        self._check_gamma( x, tau)
        scale= tau*self.gamma
        
        inv_scale = 1 / scale
        return scale* lambert_W(inv_scale* np.exp(inv_scale * x -1))

        
    def __call__(self, x):
        if np.size(x) <= 1:
            x=np.reshape(x, (-1))
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        result= np.zeros(np.shape(x))
        mask= x>0
        result[mask] = x[mask]*np.log(x[mask])
        mask2= x<0
        result[mask2]= np.inf  
        
        return self.gamma*result
       
    def _check_gamma(self, x, tau, name_name="'tau'"):
          
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
           
       if (np.size(self.gamma) > 1)  and (np.shape(self.gamma) != np.shape(x) ):
           raise Exception("gamma' must be either scalar or the same size as 'x'")  


