#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-08-2022)
 
 DEPENDENCIES:
     'lambert_W.py' in the folder 'utils'
     
     
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).

"""

import numpy as np
from prox.utils.lambert_W import lambert_W

class Exp:
    """
    This method computes the proximity operator of the function

        f(x)= gamma*tau*exp(x)
        
    When the input 'x' is an array, the output is computed element-wise.
    
    'tau*gamma' is the scale factor
    
     INPUTS
    ========
    x      -  array_like
    gamma  - positive, scalar or ND array with the same size as 'x' [default value: gamma=1]
    tau    - positive, scalar or ND array with the same size as 'x' [default: tau=1] [Only for the method 'prox']
    
    =======
    Examples
    ========
 
     Evaluate the 'direct' function :
    
     >>> Exp()( np.log(2) )
     2.0
     >>> Exp(gamma=2)( [-1, 0, 3, np.log(2)] )  # compute the result element-wise for vector inputs
     array([ 0.73575888,  2.        , 40.17107385,  4.        ])
     
     Compute the proximity operator at a given point :
         
     >>> Exp().prox(  [-2, 3, 4, np.log(2)] )
     array([-2.12002824,  0.79205997,  1.07372894, -0.15945832])
     
     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function 'gamma*f'
    
     >>> Exp(gamma=2.5).prox( [-2, 3, 4, np.log(3) ] )
      array([-2.26069497,  0.13594702,  0.37230635, -0.46761867])
     
      This is equivalent to use the parameter 'tau' of the method 'prox' to compute the 
      proximity operator of  the function 'tau*f'
      
     >>> Exp().prox( [-2, 3, 4, np.log(3) ] , tau=2.5 )
      array([-2.26069497,  0.13594702,  0.37230635, -0.46761867])
     """

    def __init__(self, gamma=1):
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
        
    
    def prox(self, x, tau=1):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)):
            tau=np.array(tau)
            
        self._check( x, tau)
        scale = tau*self.gamma
        
        if np.size(x) <= 1:
            x=np.reshape(x, (-1))
            
        p= np.zeros( np.shape(x) )
        # we use a taylor approximation to avoid divergence when the input is too big
        u=np.log(scale) + x
        mask = u > 100
        u=u[mask]
        p[mask]= u - u/(1+u)*np.log(u)
        not_mask = np.logical_not(mask)
        if np.size(scale)>1:
            scale=scale[not_mask]
        p[ not_mask ] = lambert_W( scale* np.exp(x[not_mask]))

        return x - p
        
    def __call__(self, x):
        return self.gamma*np.exp(x)
    
    def _check(self, x, tau, name_name="'tau'"):
          
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
           
       if (np.size(self.gamma) > 1)  and (np.shape(self.gamma) != np.shape(x) ):
           raise Exception("gamma' must be either scalar or the same size as 'x'")  

        