#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-09-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).

"""

import numpy as np

class FairPotential:
    """
    This method computes the proximity operator of the 'Fair Potential' function

           f(x)=- gamma*( w*|x| - log(1+w|x|) )  with w > 0
        
    When the input 'x' is an array, the output is computed element-wise.
    
    'tau*gamma' is the scale factor
    
     INPUTS
    ========
     x     - scalar or array
     w     - positive , scalar or array with the same size as 'x'
     gamma  - positive, scalar or array with the same size as 'x' [default: gamma=1]
     tau    - positive, scalar or array with the same size as 'x' [default: tau=1] [Only for the method 'prox']
     
     =======
    Examples
    ========
 
     Evaluate the 'direct' function :
    
     >>> FairPotential()( 2 )
     0.9013877113318902
     
      Compute the result element-wise for vector inputs :
     >>> FairPotential(gamma=2)( [-1, 0, 3, np.e] )  
      array([0.61370564, 0.        , 3.22741128, 2.81004028])
     
     Compute the proximity operator at a given point :
         
     >>> FairPotential().prox(  [-2, 3, 4, np.e] )
     array([-1.5       ,  2.5       ,  3.5       ,  2.21828183])
     
     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function 'gamma*f'
    
     >>> FairPotential(gamma=2.5).prox( [-2, 3, 4, np.e ] )
     array([-0.68614066,  1.5       ,  2.38600094,  1.26147065])
     
      This is equivalent to use the parameter 'tau' of the method 'prox' to compute the 
      proximity operator of  the function 'tau*f'
      
     >>> FairPotential().prox( [-2, 3, 4, np.e ] , tau=2.5 )
     array([-0.68614066,  1.5       ,  2.38600094,  1.26147065])
    """

    def __init__(self, w=1, gamma=1):
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.size(w)>1 and (not isinstance(w, np.ndarray)):
            w=np.array(w)
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        if np.any( w <= 0 ):
            raise Exception("'w' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
        self.w = w
        

    def prox(self, x, tau=1):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)):
            tau=np.array(tau)
        if np.size(x)<=1:
            x=np.reshape(x, (-1))
            
        self._check_gamma( x, tau)
        scale= tau*self.gamma
        w=self.w
        
        sign_x = np.sign(x)
        abs_x = np.abs(x)
        p = (
            w * abs_x
            - w ** 2 * scale
            - 1
            + np.sqrt((w * abs_x - w ** 2 * scale) ** 2 + 4 * w * abs_x)
        )

        return sign_x * p / (2 * w)

        
    def __call__(self, x):
        w=self.w
        return self.gamma*( w*np.abs(x) - np.log(1+ w*np.abs(x)) )      
    
    def _check_gamma(self, x, tau, name_name="'tau'"):
          
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
           
       if (np.size(self.gamma) > 1)  and (np.shape(self.gamma) != np.shape(x) ):
           raise Exception("gamma' must be either scalar or the same size as 'x'")  
        