#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-09-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).

"""

import numpy as np

class FermiDIracEntropy:
    """
    This method computes the proximity operator of the function

                      /  gamma*tau*( x*log(x)  + (1-x)log(x) )    if  x> 0
               f(x)= |  0                                       if x in {0, 1}
                     \  +INF                                    otherwise
    
        
    When the input 'x' is an array, the output is computed element-wise.
    
    'tau*gamma' is the scale factor
    
    
    INPUTS
    ========
     x     - scalar or array
     gamma  - positive, scalar or array with the same size as 'x' [default value: gamma=1]
     tau    - positive, scalar or array with the same size as 'x' [default: tau=1] [Only for the method 'prox']
     
     =======
    Examples
    ========
 
     Evaluate the 'direct' function :
    
     >>> FermiDIracEntropy()( 0.5 )
     array([-0.69314718])
     
      Compute the result element-wise for vector inputs :
          
     >>> FermiDIracEntropy(gamma=2)( [-1, 0, 0.25, np.e] )  
     array([        inf,  0.        , -1.12467029,         inf])
     
     Compute the proximity operator at a given point :
         
     >>> FermiDIracEntropy().prox(  [-2, 3, 4, np.e] )
     array([0.10829336, 0.89170664, 0.9545842 , 0.86456313])
     
     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function 'gamma*f' :
    
     >>> FermiDIracEntropy(gamma=2.5).prox( [-2, 3, 4, np.e ] )
     array([0.28609273, 0.71390727, 0.78356663, 0.69220157])
     
      This is equivalent to use the parameter 'tau' of the method 'prox' to compute the 
      proximity operator of  the function 'tau*f' :
      
     >>> FermiDIracEntropy().prox( [-2, 3, 4, np.e ] , tau=2.5 )
      FermiDIracEntropy().prox( [-2, 3, 4, np.e ] , tau=2.5 )
    """
    
    def __init__(self, gamma=1):
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
        
        

    def prox(self, x, tau=1):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)):
            tau=np.array(tau)
        if np.size(x)<=1:
            x=np.reshape(x, (-1))
        
        gamma= tau*self.gamma
        self._check(x, gamma)
        
        limit = 8  
        # RUN HALEY'S METHOD FOR SOLVING w = W_{exp(x/gamma)}(exp(x/gamma)/gamma)
        # where W_r(x) is the generalized Lambert function
        
        w = np.zeros_like(x)
        igamma = 1/gamma
        c = x*igamma - np.log(gamma)
        z = np.exp(c)
        r = np.exp(x*igamma)
         
        # ASYMPTOTIC APPROX
        approx = 1 - np.exp((1-x)*igamma)
        
        #INITIALIZATION
        # Case 1: gamma <= 1/30
        mask= np.logical_and(z>1, gamma<=1/30)
        w[mask] = c[mask]- np.log( c[mask])
        # Case 2: gamma > 1/30
        mask= np.logical_and(z>1, gamma>1/30)
        if np.size(igamma)>1:
            igamma=igamma[mask]
        w[mask]= igamma*approx[mask]
         
        #RUN
        maxiter = 20
        testend = np.zeros_like(x)
        tol = 1e-8
        epsilon= 1e-16 # to avoid dividing by zero
        n=np.size(x)
    
        for it in range(maxiter):
         
            e = np.exp(w)
            y = w*e + r*w - z
            v = e*(1 + w) + r
            u = e*(2 + w)
            wnew = w - y/(v - y*u/(2*v))
            
            mask= np.abs(wnew-w)/( np.abs(w) +epsilon) < tol
            testend[ np.logical_and(mask, testend == 0)] = 1
            not_mask= np.logical_not(mask)
            idx_update = np.logical_and( not_mask, testend==0)
            w[idx_update] = wnew[idx_update] # the rest stays constant !
            if np.sum(testend)==n : #stop !
                break
            
         
        p = gamma*w
        if np.size(p)<=1:
            if c > limit and gamma>1:
                p= approx
            return np.minimum(p, 1)

        #ASYMPTOTIC DVP
        mask_approx=  np.logical_and(c>limit, gamma > 1)
        p[mask_approx]= approx[ mask_approx]
        
        #FINAL TRESHOLD TO AVOID NUMERICAL ISSUES FOR SMALL GAMMA
        p = np.minimum(p,1)
        return p

        
    def __call__(self, x):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(x)<=1:
            x=np.reshape(x, (-1))
        result= np.zeros(np.shape(x))
        mask= np.logical_and( x>0, x<1)
        result[mask] = x[mask]*np.log(x[mask])+ (1-x[mask])*np.log(1-x[mask])
        mask2= np.logical_or( x<0, x>1)
        result[mask2]= np.inf  
        
        return self.gamma*result
    
    def _check(self, x, tau, name_name="'tau'"):
          
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception("'tau' and 'gamma' must be either scalar or the same size as 'x'")
           
    