#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-07-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np

class Huber:
    """
    Computes the proximity operator of the Huber function
    
                  / gamma *tau * 1/2 * x^2             if |x| <= w
           f(x)= | 
                 \  gamma *tau* w* |x| - w^2  /2       otherwise 
                 
                 
                 with w>0
            

    When the input 'x' is an array, the output is computed element-wise.
    
    'gamma*tau' is the scale factor
    
    ========
     x      - ND array
     w      - positive, scalar or ND array with the same size as 'x'
     gamma  - positive, scalar or ND array with the same size as 'x' ( default value: gamma=1)
     tau    - positive, scalar or ND array with the same size as 'x'  ( default value: tau=1)
     
    =======
    Examples
    ========
 
     Evaluate the 'direct' function :
    
     >>> Huber(2)( 0.25 )
     array([0.03125])
     
      Compute the result element-wise for vector inputs :
          
     >>> Huber(2)( [-1, 3, 0.25] )  
     array([0.5    , 4.     , 0.03125])
     
     Compute the proximity operator at a given point :
         
     >>> Huber(3).prox(  [-2, 3, 4 ])
      array([-1,  1,  2])
     
     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function 'gamma*f'
    
     >>> Huber(3, gamma=2.5).prox( [-2, 3, 4, np.e ] )
     array([-0.57142857,  0.85714286,  1.14285714,  0.77665195])
     
      This is equivalent to use the parameter 'tau' of the method 'prox' to compute the 
      proximity operator of  the function 'tau*f' :
      
     >>> Huber(3).prox( [-2, 3, 4, np.e ] , tau=2.5 )
     array([-0.57142857,  0.85714286,  1.14285714,  0.77665195])
    """
    
    def __init__(self, w, gamma=1):
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
        if np.size(w)>1 and (not isinstance(w, np.ndarray)):
            w=np.array(w)
        if np.any( w <= 0 ):
            raise Exception("'w' (or all of its components if it is an array) must be strictly positive")
        self.w= w

    def prox(self, x, tau=1):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)):
            tau=np.array(tau)
        if np.size(x)<=1:
            x=np.reshape(x, (-1))
        self._check_gamma( x, tau)
        
        gamma= self.gamma*tau
        w=self.w
        
        # preliminaries
        abs_x = np.abs(x)
        sign_x = np.sign(x)

        # 2nd branch
        p = x - gamma * w * sign_x

        # 1st branch
        mask = abs_x <= w * (gamma + 1)
        gamma_=gamma
        if np.size(gamma)>1:
            gamma_= gamma[mask]
        p[mask] = x[mask] / (1 + gamma_)

        return p

        
    def __call__(self, x, gamma=1):
        abs_x = np.abs(x)
        w=self.w
        if np.size(x)<=1:
            x=np.reshape(x, (-1))
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(x)<=1:
            if abs_x <= w :
                return self.gamma*(0.5*x**2)
            else:
                return self.gamma*(w*abs_x-w**2/2)
        
        p=0.5*x**2
        mask = abs_x > w 
        
        if np.size(w)>1:
            p[mask]= abs_x[mask]*w[mask]-w[mask]**2/2
            
        else:
            p[mask]= abs_x[mask]*w-w**2/2
            
        return self.gamma*p
    
    def _check_gamma(self, x, tau, name_name="'tau'"):
          
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
           
       if (np.size(self.gamma) > 1)  and (np.shape(self.gamma) != np.shape(x) ):
           raise Exception("gamma' must be either scalar or the same size as 'x'")  

