#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-08-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""
import numpy as np
from prox.utils.newton import newton_

class Hyperbolic:
    """
    Computes the proximity operator of the hyperbolic function 
    
              f(x) = gamma* tau* sqrt ( x^2 + delta^2)
  
    When the input 'x' is an array, the output is computed element-wise.
    
    'gamma*tau' is the the scale factor
    
    INPUTS
    ========
    x      - scalar or array
    delta  - strictly positive, scalar or array with the same size as 'x'
    gamma  - positive, scalar or array with the same size as 'x' [ default value: gamma=1]
    tau    - positive, scalar or array with the same size as 'x' [ default value: tau=1][default value: tau=1]
     
    =======
    Examples
    ========
 
     Evaluate the 'direct' function :
    
     >>> Hyperbolic(delta=3)( 4 )
     5.0
     
      Compute the result element-wise for vector inputs :
          
     >>> Hyperbolic(delta=2, gamma=3)( [-1, 3, 0.25] )  
     array([ 6.70820393, 10.81665383,  6.04669331])
     
     Compute the proximity operator at a given point :
         
     >>> Hyperbolic(3).prox(  [-2, 3, 4 ])
     array([-1.54269095,  2.37870272,  3.26376922])
     
     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function 'gamma*f'
    
     >>> Hyperbolic(3, gamma=2.5).prox( [-2, 3, 4, np.e ] )
     array([-1.12333244,  1.74370688,  2.42744522,  1.5630981 ])
     
      This is equivalent to use the parameter 'tau' of the method 'prox' :
      
     >>> Hyperbolic(3).prox( [-2, 3, 4, np.e ] , tau=2.5 )
     array([-1.12333244,  1.74370688,  2.42744522,  1.5630981 ])
    """
    
    def __init__(self, delta, gamma=1):
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.any(gamma <= 0):
            raise Exception("'gamma' (or all its components if it is an array) must be strictly positive")
        if np.any(delta <= 0):
            raise Exception("'delta' (or all its components if it is an array) must be strictly positive")
        self.delta=delta
        self.gamma=gamma
        
    
    def prox(self, x, tau=1):  
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)):
            tau=np.array(tau)
        scale= self.gamma*tau
        self._check( x, scale)
        abs_x = np.abs(x)

        def fun_phi(p):
            return (
                p ** 4
                + (-2 * abs_x) * p ** 3
                + (x ** 2 - scale** 2 + self.delta ** 2) * p ** 2
                + (-2 * abs_x * self.delta ** 2) * p
                + (self.delta ** 2) * (x ** 2)
            )

        def der_phi(p):
            return (
                4 * p ** 3
                - 6 * abs_x* p ** 2
                + 2 * (x ** 2 - scale ** 2 + self.delta ** 2) * p
                + (-2 * abs_x * self.delta ** 2)
            )
        # starting point
        p_init = abs_x / 2
        p = newton_(fun_phi, fp=der_phi, x0=p_init, low=0, high=abs_x)

        return p * np.sign(x) 

    def __call__(self, x):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        return self.gamma*np.sqrt( x**2+ self.delta**2)
    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(" 'tau' and 'gamma' (or all of their components if they are arrays) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.size(tau) != np.size(x) ):
           raise Exception("'tau' and 'gamma' must be either scalar or ND array with the same size as 'x'")


      
