#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-10-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np

class ISRU:
    """
    This method computes the proximity operator of the 'Inverse Square Root Unit ' function
           
                   / -(x^2)/2-sqrt(1-x^2)   if |x| <= 1
           f(x) = |
                  \ +Inf                     otherwise


    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
     x     - scalar or ND array
     
     =======
     Examples
     ========
     
     >>>ISRU()( -3 )
      array([inf])
     >>> ISRU()( [1, -1.5, -0.5] )
     array([-0.5      ,        inf, -0.9910254])
   
     Compute the proximity operator at a given point :
    
     >>>ISRU().prox(-3)
     -0.9486832980505138
     >>>ISRU().prox(  [-2, 3, -0.5] )
      array([-0.89442719,  0.9486833 , -0.4472136 ])
    """

    def __init__(self):
        self.gamma=1

    def prox(self, x):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        p = x / np.sqrt(1 + x ** 2)
        return p

        
    def __call__(self, x):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(x) <= 1:
            x=np.reshape(x, (-1))

        result= np.zeros(np.shape(x))
        mask = np.abs(x) <= 1

        result[mask] = -x[mask]**2/2 -np.sqrt( 1-x[mask]**2)
        
        result[ np.abs(x) > 1 ]= np.inf  
        
        return result
      
      
        