#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-09-2022)
 
 DEPENDENCIES:
     'newton.py' in the folder 'utils'
     
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np
from prox.utils.newton import newton_


class Inverse :
    """
    Computes the proximity operator of the funcx^2tion
    
                  / gamma*tau* 1/x^q if x>0
           f(x)= | 
                 \  +INF      otherwise with q>=1

    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
     x     - scalar or ND array
     q     - positive >= 1, scalar or ND array with the same size as 'x' [default value: q=1] 
     gamma - positive, scalar or ND array with the same size as 'x' [default value: gamma=1] 
     tau   - positive, scalar or ND array with the same size as 'x' [ default value: tau=1][default value: tau=1]
     
     
    =======
    Examples
    ========
 
     Evaluate the 'direct' function :
    
     >>> Inverse()( 4 )
     0.25
     
      Compute the result element-wise for vector inputs :
          
     >>> Inverse(q=2, gamma=3)( [4., -8., 0.25] )  
     array([ 0.1875,     inf, 48.    ])
     
     Compute the proximity operator at a given point :
         
     >>> Inverse(2).prox(  [-2, 3, 4 ])
     array([8.85033504e-01, 3.60403299e+08, 3.60403300e+08])
     
     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function 'gamma*f'
    
     >>> Inverse(3, gamma=2.5).prox( [-2, 3, 4, np.e ] )
     array([1.23403930e+00, 4.66240534e+11, 2.38715153e+11, 4.66240534e+11])
     
      This is equivalent to use the parameter 'tau' of the method 'prox' :
      
     >>> Inverse(3).prox( [-2, 3, 4, np.e ] , tau=2.5 )
     array([1.23403930e+00, 4.66240534e+11, 2.38715153e+11, 4.66240534e+11])
    """
    
    
    def __init__(self, q=1, gamma=1):
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.size(q)>1 and (not isinstance(q, np.ndarray)):
            q=np.array(q)
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma=gamma
        if np.any(q < 1):
            raise Exception("'q' must be greater than 1")
        self.q=q
        

    def prox(self, x, tau=1):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)):
            tau=np.array(tau)
        self._check_gamma( x, tau)
        q=self.q
        abs_x = np.abs(x)
        scale= self.gamma*tau
        
        def polynom_phi(t):
            return t ** (q + 2) - x * t ** (q + 1) - scale * q

        def derive_polynom_phi(t):
            return (q + 2) * t ** (q + 1) - (q + 1) * x * t * q

        # starting point
        p_init = abs_x / 2

        # Finding the root of the polynom with the Newton method
        p = newton_(polynom_phi, fp=derive_polynom_phi, x0=p_init, low=0, high= np.inf)

        return p

        
    def __call__(self, x):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        self._check_gamma( x, self.gamma, 'gamma')
        if np.size(x)<=1:
            if x > 0:
                return self.gamma*(1/x**self.q)
            return np.inf
        
        xx= np.zeros_like(1.0*x)
        mask= x>0
        gamma_=self.gamma
        if np.size(self.gamma)>1:
            gamma_=self.gamma[mask]
        q=self.q
        q_=q
        if np.size(q)>1:
            q_=q[mask]
            
        xx[mask]=gamma_*(1/x[mask]**q_)
        
        xx[ np.logical_not(mask)]= np.inf
        
        return xx
    
    def _check_gamma(self, x, tau, name_name="'tau'"):
          
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
           
       if (np.size(self.gamma) > 1)  and (np.shape(self.gamma) != np.shape(x) ):
           raise Exception("gamma' must be either scalar or the same size as 'x'") 
            