#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-09-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np

class Log :
    """
    This method computes the proximity operator of the logarithmic function

                      /  - gamma*tau* log(x)          if  x> 0
               f(x)= |                                   
                     \   +INF                         otherwise

        
    When the input 'x' is an array, the output is computed element-wise.
    
    'tau*gamma' is the scale factor
    
     INPUTS
    ========
     x     - scalar or ND array
     gamma - positive, scalar or ND array with the same size as 'x' (default value: gamma=1)
     tau   - positive, scalar or ND array with the same size as 'x' (default value: tau=1)
     
    
    =======
    Examples
    ========
 
     Evaluate the function  f:
    
     >>> Log()( np.e )
      array([-1.])
     
      Compute the result element-wise for vector inputs :
          
     >>> Log(gamma=3)( [-1, 3, np.e] )  
    array([        inf, -3.29583687,   -3.    ])
     
     Compute the proximity operator at a given point :
         
     >>> Log().prox(  [-2, 3, 4 ])
     array([0.41421356, 3.30277564, 4.23606798])
     
     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function 'gamma*f'
    
     >>> Log( gamma=2.5).prox( [-2, 3, 4, np.e ] )
     array([0.87082869, 3.67944947, 4.54950976, 3.44415027])
     
      This is equivalent to use the parameter 'tau' of the method 'prox' :
      
     >>> Log().prox( [-2, 3, 4, np.e ] , tau=2.5 )
     array([0.87082869, 3.67944947, 4.54950976, 3.44415027])
    """
    
    def __init__(self, gamma=1):
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma

        
    def prox(self, x, tau=1):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)):
            tau=np.array(tau)
        self._check_gamma( x, tau)
        scale= tau*self.gamma
        
        return 0.5*(x + np.sqrt(x ** 2 + 4 * scale))

        
    def __call__(self, x):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        self._check_gamma( x, self.gamma, 'gamma')
        if np.size(x) <= 1:
            x=np.reshape( x, (-1))
        result=np.zeros_like(x, dtype=np.float)
        gamma=self.gamma
        mask= x>0
        g_=gamma
        if np.size(gamma)>1:
            g_=gamma[mask]
        result[mask]= -g_*np.log(x[mask])
        result[x<=0] = np.inf

        return result
    
    def _check_gamma(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
           
       if (np.size(self.gamma) > 1)  and (np.shape(self.gamma) != np.shape(x) ):
           raise Exception("gamma' must be either scalar or the same size as 'x'") 
