#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-09-2022)
 
 DEPENDENCIES:
     'newton.py' in the folder 'utils'
     
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""
       
import numpy as np
from prox.utils.newton import newton_

class LogBarriers :
    """
    This method computes the proximity operator of the 'logarithm-Barriers' function

                          /  - gamma*tau* ( log(x-a) + log(b-x) )       if  a < x < b
                   f(x)= |                                   
                         \  +INF                                        otherwise
        
            with a < b
        
    When the input 'x' is an array, the output is computed element-wise.
    
    'tau*gamma' is the scale factor
    
     INPUTS
    ========
     x     - scalar or ND array
     a     - scalar or ND array with the same size as 'x' (default: a=1)
     b     - scalar or ND array with the same size as 'x' (default: b=1)
     gamma - positive, scalar or ND array with the same size as 'x' (default value: gamma=1)
     tau   - positive, scalar or ND array with the same size as 'x'   (default value: tau=1) (only for the method 'prox'')
     
     
    =======
    Examples
    ========
 
     Evaluate the 'direct' function :
    
     >>>  LogBarriers()( 0.5 )
     array([1.38629436])
     
      Compute the result element-wise for vector inputs :
          
     >>> LogBarriers(a=[-2, 0, -1], b=[2, 3, 5])( [-1, 3, 0.25] ) 
     array([-1.09861229,         inf, -1.78128817])
     
     Compute the proximity operator at a given point :
         
     >>>  LogBarriers(a=[-2, 0, -1], b=[2, 3, 5]).prox(  [-2, 3, 4 ])
     array([-1.54269095,  2.37870272,  3.26376922])
     
     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function 'gamma*f'
    
     >>>  LogBarriers(a=[-2, 0, -1], b=[2, 3, 5], gamma=2.5).prox( [-2, 3, 4 ] )
     array([-0.80311322,  1.93746944,  3.20331613])
     
      This is equivalent to use the parameter 'tau' of the method 'prox' :
      
     >>>  LogBarriers(a=[-2, 0, -1], b=[2, 3, 5]).prox( [-2, 3, 4 ] , tau=2.5 )
     array([-0.80311322,  1.93746944,  3.20331613]
    """

    def __init__(self, a=0 ,b=1, gamma=1):
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.size(a)>1 and (not isinstance(a, np.ndarray)):
            a=np.array(a)
        if np.size(b)>1 and (not isinstance(b, np.ndarray)):
            b=np.array(b)
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
        if np.any( b < a):
            raise Exception("'b' (or all its components if it is an array) must be greater than 'a'")
        self.a=a
        self.b=b
        

    def prox(self, x, tau=1):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)):
            tau=np.array(tau)
        
        scale= tau*self.gamma
        self._check( x, scale)
        a=self.a
        b=self.b
        def polynom_phi(t):
            return (
                t ** 3
                - (a + b + x) * t ** 2
                + (a * b + (a + b) * x - 2 * scale) * t
                - a * b * x
                + (a + b) * scale
            )

        def der_polynom_phi(t):
            return (
                3 * t ** 2 - 2 * (a + b + x) * t + a * b + (a + b) * x - 2 * scale
            )

        # starting point
        p_init = (a + b) / 2 
        p = newton_(polynom_phi, fp=der_polynom_phi, x0=p_init, low=a, high=b)

        return p

        
    def __call__(self, x):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        a=self.a
        b=self.b
        self._check( x, self.gamma)
        if np.size(x) <= 1:
            x=np.reshape(x, (-1))
            
        result= np.zeros_like(x)
        mask= np.logical_and( x>a, x<b)
        aa=a
        bb=b
        if np.size(a) >1:
            aa=a[mask]
        if np.size(b) >1:
            bb=b[mask]
            
        result[mask] = -np.log(x[mask]-aa)-np.log(bb-x[mask])
        
        result[ np.logical_not(mask) ]= np.inf  
        
        return self.gamma*result
      
    def _check(self, x, tau, name_name="'tau'"):
          
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception("'tau' and 'gamma' must be either scalar or the same size as 'x'")
           
       if ( np.size(self.a) > 1)  and (np.shape(self.a) != np.shape(x) ):
            raise Exception("'a' must be either scalar or the same size as 'x'")
            
       if ( np.size(self.b) > 1)  and (np.shape(self.b) != np.shape(x) ):
            raise Exception("'b' must be either scalar or the same size as 'x'")
        