#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-09-2022)
 DEPENDENCIES:
     'newton.py' in the folder 'utils'
     
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np
from prox.utils.newton import newton_

class LogInverse:
    """
    This method computes the proximity operator of the 'logarithm+inverse' function

                      / gamma*tau( - log(x) + w * x^(-1) )     if  x> 0
               f(x)= |                                     
                     \  +INF                                  otherwise
    
        with w > 0 and q >= 1
        
    When the input 'x' is an array, the output is computed element-wise.
    
    'tau*gamma' is the scale factor
    
     INPUTS
    ========
     x     - scalar or ND array
     w     - positive , scalar or ND array with the same size as 'x'
     gamma - positive, scalar or ND array with the same size as 'x' (default value: gamma=1)
     tau   - positive, scalar or ND array with the same size as 'x'   (default value: tau=1) (only for the method 'prox')
    
    
    =======
    Examples
    ========
 
     Evaluate the function  f:
    
     >>> LogInverse()( np.e )
     array([-0.63212056])
     
      Compute the result element-wise for vector inputs :
          
     >>> LogInverse(w=[1, 2, 3], gamma=3)( [-1, 3, np.e] )  
     array([        inf, -1.29583687,  0.31091497])
     
     Compute the proximity operator at a given point :
         
     >>> LogInverse(w=[1,2,3]).prox(  [-2, 3, 4 ])
     array([0.41421356, 3.30277564, 4.23606798])
     
     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function 'gamma*f'
    
     >>> LogInverse( [1,2,3,3], gamma=2.5).prox( [-2, 3, 4, np.e ] )
     array([1.3218996 , 3.95255089, 4.83733009, 3.86652778])
     
      This is equivalent to use the parameter 'tau' of the method 'prox' :
      
     >>> LogInverse([1,2,3, 3]).prox( [-2, 3, 4, np.e ] , tau=2.5 )
     array([1.3218996 , 3.95255089, 4.83733009, 3.86652778])
    """

    def __init__(self, w=1, gamma=1):
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.size(w)>1 and (not isinstance(w, np.ndarray)):
            w=np.array(w)
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma

        if np.any( w < 0):
            raise Exception("'w (or all its components if it is an array) must be positive")
        self.w=w

    def prox(self, x, tau=1):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)):
            tau=np.array(tau)
        self._check( x, tau)
        scale= tau*self.gamma
        w=self.w
        def polynom_phi(t):
            return t ** 3 -x* t ** 2 - scale * t - w*scale

        def der_polynom_phi(t):
            return 3*t**2 -2*x*t -scale

        p_init = np.abs(x)

        p = newton_(polynom_phi, fp=der_polynom_phi, x0=p_init, low=1e-16, high=np.inf)

        return p

    def __call__(self, x):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(x) <= 1:
            x=np.reshape(x, (-1))
            
        if ( np.size(self.w) > 1)  and (np.shape(self.w) != np.shape(x) ):
            raise Exception("' w' must be either scalar or the same size as 'x'")
        
        result= np.zeros_like(1.0*x)
        mask=x>0
        ww=self.w
        if np.size(self.w) >1:
            ww=self.w[mask]
        result[mask] = -np.log(x[mask])+ ww*1/x[mask]
        
        result[x<=0]= np.inf  
        
        return self.gamma*result
    
    def _check(self, x, tau, name_name="'tau'"):
          
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.sizee(tau) != np.size(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
           
       if (np.size(self.gamma) > 1)  and (np.size(self.gamma) != np.size(x) ):
           raise Exception("gamma' must be either scalar or the same size as 'x'")  
        