#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-09-2022)
 DEPENDENCIES:
     'newton.py' in the folder 'utils'
     
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np
from prox.utils.newton import newton_
        
class LogPower:
    """
    This method computes the proximity operator of the 'logarithm+power' function

                      / gamma*tau( - log(x) + w * x^ q )     if  x> 0
               f(x)= |                                     
                     \  +INF                                 otherwise
    
        with w > 0 and q >= 1
        
    When the input 'x' is an array, the output is computed element-wise.
    
    
    'tau*gamma' is the scale factor
    
     INPUTS
    ========
     x     - ND array
     q     - positive >=1, scalar or ND array with the same size as 'x'
     w     - positive , scalar or ND array with the same size as 'x'
     gamma - positive, scalar or ND array with the same size as 'x' (default value: gamma=1)
     tau   - positive, scalar or ND array with the same size as 'x'   (default value: tau=1) (only for the method 'prox')
    
    
    =======
    Examples
    ========
 
     Evaluate the function  f:
    
     >>> LogPower(q=1)( np.e )
     array([1.71828183])
     
      Compute the result element-wise for vector inputs :
          
     >>> LogPower(q=2, w=[1, 2, 3], gamma=3)( [-1, 3, np.e] )  
     array([        inf, 50.70416313, 63.50150489])
     
     Compute the proximity operator at a given point :
         
     >>> LogPower(2, [1,2,3]).prox(  [-2, 3, 4 ])
     array([0.33333333, 0.83851648, 0.75951783])
     
     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function 'gamma*f'
    
     >>> LogPower( 2, [1,2,3,3], gamma=2.5).prox( [-2, 3, 4, np.e ] )
     array([0.5       , 0.63221419, 0.5395781 , 0.48925544])
     
      This is equivalent to use the parameter 'tau' of the method 'prox' :
      
     >>> LogPower(2, [1,2,3, 3]).prox( [-2, 3, 4, np.e ] , tau=2.5 )
     array([0.5       , 0.63221419, 0.5395781 , 0.48925544])
    """

    def __init__(self, q, w=1, gamma=1):
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.size(w)>1 and (not isinstance(w, np.ndarray)):
            w=np.array(w)
        if np.size(q)>1 and (not isinstance(q, np.ndarray)):
            q=np.array(q)
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
        if np.any( w < 0):
            raise Exception("'w (or all its components if it is an array) must be strictly positive")
        self.w=w
        
        if np.any( q < 1):
            raise Exception("'q' (or all its components if it is an array) must be greater than 1")
        self.q=q
        

    def prox(self, x, tau=1):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)):
            tau=np.array(tau)
        self._check( x, tau)
        scale= tau*self.gamma
        q=self.q
        w=self.w
        
        def polynom_phi(t):
            return q * w * scale* t ** q + t ** 2 - x * t - scale

        def derive_polynom_phi(t):
            return q * q * w * scale * t ** (q - 1) + 2 * t - x

        abs_x = np.abs(x)
        p_init = abs_x / 2

        p = newton_(polynom_phi, fp=derive_polynom_phi, x0=p_init, low=1e-16, high=np.inf)

        return p

        
    def __call__(self, x):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(x) <= 1:
            x=np.reshape(x, (-1))
    
        if ( np.size(self.w) > 1)  and (np.shape(self.w) != np.shape(x) ):
            raise Exception("' w' must be either scalar or the same size as 'x'")
        
        result= np.zeros_like(x)
        mask=x>0
        ww=self.w
        qq=self.q
        if np.size(self.w) >1:
            ww=self.w[mask]
        if np.size(self.q) >1:
            qq=self.q[mask]
        result[mask] = -np.log(x[mask])+ ww*x[mask]**qq
        
        result[ x<=0]= np.inf  
        
        return self.gamma*result
    
    
    def _check(self, x, tau, name_name="'tau'"):
          
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
           
       if (np.size(self.gamma) > 1)  and (np.shape(self.gamma) != np.shape(x) ):
           raise Exception("gamma' must be either scalar or the same size as 'x'") 
          
            