#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-09-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np


class LogisticLoss :
    """
    This method computes the proximity operator of the function

        f(x)= gamma*tau* log( 1+ exp(x) )
        
    When the input 'x' is an array, the output is computed element-wise.

    'tau*gamma' is the scale factor
    
     INPUTS
    ========
     x     - ND array
     gamma - positive, scalar or ND array with the same size as 'x' (default value: gamma=1)
     tau   - positive, scalar or ND array with the same size as 'x' (default value: tau=1) (only for the method 'prox')
     
    =======
    Examples
    ========
 
     Evaluate the function  f:
    
     >>> LogisticLoss()( 0 )
      0.6931471805599453
     
      Compute the result element-wise for vector inputs :
          
     >>> LogisticLoss(gamma=3)( [-1, 3, np.e] )  
     array([0.93978506, 9.14576205, 8.34655179])
     
     Compute the proximity operator at a given point :
         
     >>> LogisticLoss().prox(  [-2, 3, 4 ])
     array([-2.10829336,  2.10829336,  3.0454158 ]))
     
     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function 'gamma*f'
    
     >>> LogisticLoss( gamma=2.5).prox( [-2, 3, 4, np.e ] )
     array([-1.04044264, -0.68339574, -0.55797119, -0.70389204])
     
      This is equivalent to use the parameter 'tau' of the method 'prox' :
      
     >>> LogisticLoss().prox( [-2, 3, 4, np.e ] , tau=2.5 )
    array([-1.04044264, -0.68339574, -0.55797119, -0.70389204])
    """
        
    def __init__(self, gamma=1):
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma


    def prox(self, x, tau=1):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)):
            tau=np.array(tau)
        self._check( x, tau)
        scale= self.gamma*tau
        
        x= x/scale     # we do this scalling to use the same method as in FermiDiracEntropy.py class
        gamma= 1/scale
          
        limit = 8  
        # RUN HALEY'S METHOD FOR SOLVING w = W_{exp(x/gamma)}(exp(x/gamma)/gamma)
        # where W_r(x) is the generalized Lambert function
        
        w = np.zeros_like(x)
        igamma = 1/gamma
        c = x*igamma - np.log(gamma)
        z = np.exp(c)
        r = np.exp(x*igamma)
         
        # ASYMPTOTIC APPROX
        approx = 1 - np.exp((1-x)*igamma)
        
        #INITIALIZATION
        # Case 1: gamma <= 1/30
        mask= np.logical_and(z>1, gamma<=1/30)
        w[mask] = c[mask]- np.log( c[mask])
        # Case 2: gamma > 1/30
        mask= np.logical_and(z>1, gamma>1/30)
        if np.size(igamma)>1:
            igamma=igamma[mask]
        w[mask]= igamma*approx[mask]
         
        #RUN
        maxiter = 20
        testend = np.zeros_like(x)
        tol = 1e-8
        epsilon= 1e-20 # to avoid dividing by zero
        n=np.size(x)
    
        for it in range(maxiter):
         
            e = np.exp(w)
            y = w*e + r*w - z
            v = e*(1 + w) + r
            u = e*(2 + w)
            wnew = w - y/(v - y*u/(2*v))
            
            mask= np.abs(wnew-w)/( epsilon + np.abs(w) ) < tol
            testend[ np.logical_and(mask, testend == 0)] = 1
            not_mask= np.logical_not(mask)
            idx_update = np.logical_and( not_mask, testend==0)
            w[idx_update] = wnew[idx_update] # the rest stays constant !
            if np.sum(testend)==n : #stop !
                break
            
         
        p = gamma*w
        
        if np.size(p)<=1:
            if c > limit and gamma>1:
                p= approx
            return x- scale*np.minimum(p, 1)

        #ASYMPTOTIC DVP
        mask_approx=  np.logical_and(c>limit, gamma > 1)
        p[mask_approx]= approx[ mask_approx]
        
        #FINAL TRESHOLD TO AVOID NUMERICAL ISSUES FOR SMALL GAMMA
        p = np.minimum(p,1)
        
        return( x - scale*p)
    
    def __call__(self, x):
        return self.gamma*np.log(1 +np.exp(x) )
    
    def _check(self, x, tau, name_name="'tau'"):
          
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
           
       if (np.size(self.gamma) > 1)  and (np.shape(self.gamma) != np.shape(x) ):
           raise Exception("gamma' must be either scalar or the same size as 'x'")