#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-08-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np
from prox.utils.newton import newton_


class NegativeRoot :
    """
    Computes the proximity operator of the funcx^2tion
    
                  /  - gamma* x^(1/q)    if x>=0
           f(x)= | 
                 \  +INF                 otherwise 
                 
                 with q>=1

    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
     x     - scalar or ND array
     q     - scalar >=1
     gamma - positive, scalar or ND array with the same size as 'x' (default value: gamma=1)
     tau   - positive, scalar or ND array with the same size as 'x'   (default value: tau=1) (only for the method 'prox')
    
    =======
    Examples
    ========
 
     Evaluate the function  f:
    
     >>> NegativeRoot(3)( 8 ) 
     array([-2.])
     
      Compute the result element-wise for vector inputs :
          
     >>> NegativeRoot(q=2, gamma=[1, 2, 3])( [-1, 16, 9] )  
     array([ inf, -8., -9.])
     
     Compute the proximity operator at a given point :
         
     >>> NegativeRoot(2).prox(  [-2, 3, 4 ])
     array([0.49278772, 1.3453765 , 1.43519754])
     
     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function 'gamma*f'
    
     >>> NegativeRoot( 2, gamma=[1,2,3,3]).prox( [-2, 3, 4, np.e ] )
     array([0.49278772, 1.37090672, 1.47180413, 1.36953862])
     
      This is equivalent to use the parameter 'tau' of the method 'prox' :
      
     >>> NegativeRoot(2).prox( [-2, 3, 4, np.e ] , tau=[1,2,3, 3])
      array([0.49278772, 1.37090672, 1.47180413, 1.36953862])
    """
        
    def __init__(self, q, gamma=1):
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
        if np.any(q < 1) or np.size(q)>1:
            raise Exception("'q' must be a scalar greater than 1")
        self.q=q
        
    def prox(self, x, tau=1):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)):
            tau=np.array(tau)
        self._check( x, tau)
        scale=self.gamma*tau
        
        abs_x = np.abs(x)
        q= self.q
        if q == 1:
            return np.maximum(0, x + scale)

        def polynom_phi(t):
            return t ** (2 * q - 1) - x * t ** (q - 1) - scale/ q

        def der_polynom_phi(t):
            return (2 * q - 1) * t ** (2 * q - 2) - (q - 1) * x * t ** (q - 2)

        # starting point
        p_init = abs_x 

        # Finding the root of the polynom with Newton method
        p = newton_(polynom_phi, fp=der_polynom_phi, x0=p_init, low=0, high=np.inf)
    
        return p ** (1 / q)

    def __call__(self, x):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(x) <= 1 :
            x=np.reshape(x, (-1))
        
        xx= np.zeros(np.shape(x))
        mask= x>=0
        xx[mask]= - x[mask]**(1/self.q)
        xx[ np.logical_not(mask)]= np.inf
        
        return self.gamma*xx
      
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
           
       if (np.size(self.gamma) > 1)  and (np.shape(self.gamma) != np.shape(x) ):
           raise Exception("gamma' must be either scalar or the same size as 'x'") 