#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-08-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np
from prox.utils.newton import newton_

class Power:
    """
    Computes the proximity operator of the power function :
    
    
              f(x) = gamma*tau* x ^ 2
        
    'gamma*tau' is the scale factor
    
    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
     x     - scalar or ND array
     q     -  scalar >= 1
     gamma - positive, scalar or ND array with the same size as 'x' (default value: gamma=1)
     tau   - positive, scalar or ND array with the same size as 'x'   (default value: tau=1) (only for the method 'prox')
    
    =======
    Examples
    ========
 
     Evaluate the function  f:
    
     >>> Power(3)( 2 ) 
     8
     
      Compute the result element-wise for vector inputs :
          
     >>> Power(q=2, gamma=[1, 2, 3])( [-1, 3, 2] )  
     array([ 1, 18., 12.])
     
     Compute the proximity operator at a given point :
         
     >>> Power(2).prox(  [-2, 3, 4 ])
     array([-0.66666667,  1.        ,  1.33333333])
     
     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function 'gamma*f'
    
     >>> Power( 2, gamma=[1,2,3,3]).prox( [-2, 3, 4, 6 ] )
     array([-0.66666667,  0.6       ,  0.57142857,  0.85714286])
     
      This is equivalent to use the parameter 'tau' of the method 'prox' :
      
     >>> Power(2).prox( [-2, 3, 4, 6 ] , tau=[1,2,3, 3])
     array([-0.66666667,  0.6       ,  0.57142857,  0.85714286])
    """
    def __init__(self, q, gamma=1):
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
        if np.any(q < 1) or np.size(q)>1:
            raise Exception("'q' must be a scalar greater than 1")
        self.q=q

    def prox(self, x, tau=1):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)):
            tau=np.array(tau)
        self._check(x, tau)
        abs_x = np.abs(x)
        sign_x = np.sign(x)
        q=self.q
        gamma= self.gamma*tau
        
        if q == 1:
            p = sign_x * np.maximum(0, abs_x - gamma)

        elif q == 4 / 3:
            ksi = np.sqrt(x ** 2 + 256 * gamma ** 3 / 729)
            ksi_minus = np.power(np.sqrt(ksi - x), 1 / 3)
            ksi_plus = np.power(np.sqrt(ksi + x), 1 / 3)

            p = x + 4 * gamma / (3 * 2 ** (1 / 3)) * (ksi_minus - ksi_plus)

        elif q == 3 / 2:
            gamma2 = np.power(gamma, 2)
            p = x + 9 / 8 * gamma * 2 * sign_x * (
                1 - np.sqrt(1 + 16 * abs_x / (9 * gamma2))
            )

        elif q == 2:
            p = x / (2 * gamma + 1)

        elif q == 3:
            p = sign_x * (np.sqrt(1 + 12 * gamma * abs_x) - 1)

        elif q == 4:
            ksi = np.sqrt(x ** 2 + 1 / (27 * gamma))
            p = np.power((ksi + x) / (8 * gamma), 1 / 3) - np.power(
                (ksi - x) / (8 * gamma), 1 / 3
            )

        else:

            def polynom_phi(t):
                return gamma * q * t ** (q - 1) + t - abs_x

            def derive_polynom_phi(t):
                return gamma * q * (q - 1) * t ** (q - 2) + 1

            p_init = abs_x / 2

            p = newton_(polynom_phi, derive_polynom_phi, p_init, 0, np.inf)
            p = sign_x * p

        return p

        
    def __call__(self, x):
        return self.gamma*np.abs(x)**(self.q)
    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
           
       if (np.size(self.gamma) > 1)  and (np.shape(self.gamma) != np.shape(x) ):
           raise Exception("gamma' must be either scalar or the same size as 'x'") 