#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-10-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np

class SmoothedFermiDiracEntropy :
    """
    This method computes the proximity operator of the 'Smoothed Fermi-Dirac Entropy' function

                      /  gamma*tau*( xlog(x)+ (1-x)log(1-x) -0.5*x^2 ) if 0< x < 1
               f(x)= |   0                                               if x in {0, 1}
                     \  +INF                                             otherwise
    
        
    When the input 'x' is an array, the output is computed element-wise.
    
    'gamma*tau' is the the scale factor
    
     INPUTS
    ========
     x     - scalar or ND array
     
     
    =======
    Examples
    ========
 
     Evaluate the function  f:
    
     >>> SmoothedFermiDiracEntropy()( 0.5 ) 
     array([-0.81814718])
     
      Compute the result element-wise for vector inputs :
          
     >>> SmoothedFermiDiracEntropy()( [-1, .3, 1] )  
     array([       inf, -0.6558643,  0.       ])
     
     Compute the proximity operator at a given point :
     >> SmoothedFermiDiracEntropy().prox(   3 )
     0.9525741268224334
     >>> SmoothedFermiDiracEntropy().prox(  [-2, 3, 4 ])
     array([0.11920292, 0.95257413, 0.98201379])
     
     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function 'gamma*f'
    
     >>> SmoothedFermiDiracEntropy( 2, gamma=[1,2,3,3]).prox( [-2, 3, 4, 6 ] )
     array([-0.66666667,  0.6       ,  0.57142857,  0.85714286])
     
      This is equivalent to use the parameter 'tau' of the method 'prox' :
      
     >>> SmoothedFermiDiracEntropy(2).prox( [-2, 3, 4, 6 ] , tau=[1,2,3, 3])
     array([-0.66666667,  0.6       ,  0.57142857,  0.85714286])
    """

    def __init__(self):
        self.gamma =1
            
    def prox(self, x):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        return 1 / (1 + np.exp(-x))

    def __call__(self, x):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(x) <= 1:
            x=np.reshape(x, (-1))
        result= np.zeros_like(1.0*x)
        mask= np.logical_and( x>0, x<1)
        result[mask] = x[mask]*np.log(x[mask])+ (1-x[mask])*np.log(1-x[mask])-0.5*x[mask]**2
        mask2= np.logical_or( x<0, x>1)
        result[mask2]= np.inf  
        
        return self.gamma*result
       
        