#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-07-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""
import numpy as np

class Square :
    """
    Computes the proximity operator of the function
    
    
            f(x) = gamma*tau * ( 1/2 * x^2 )

    When the input 'x' is an array, the output is computed element-wise.
    
    'gamma*tau' is the the scale factor

     INPUTS
    ========
     x     - ND array
     gamma - positive, scalar or ND array with the same size as 'x' (default value: gamma=1)
     tau   - positive, scalar or ND array with the same size as 'x'   (default value: tau=1) (only for the method 'prox')
    
    =======
    Examples
    ========
 
     Evaluate the function  f:
    
     >>> Square()( 4 ) 
     8.0
     
      Compute the result element-wise for vector inputs :
          
     >>> Square()([-1, 3, 2] )  
     array([0.5, 4.5, 2. ])
     
     Compute the proximity operator at a given point :
         
     >>> Square().prox(  [-2, 3, 4 ])
     array([-1. ,  1.5,  2. ])
     
     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function 'gamma*f'
    
     >>> Square(2).prox( [-2, 3, 4, 6 ] )
     array([-0.66666667,  1.        ,  1.33333333,  2.        ])
     
      This is equivalent to use the parameter 'tau' of the method 'prox' :
      
     >>> Square().prox( [-2, 3, 4, 6 ] , tau=2)
     array([-0.66666667,  1.        ,  1.33333333,  2.        ])
    """
    
    def __init__(self, gamma=1):
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma

    def prox(self, x, tau=1):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)):
            tau=np.array(tau)
        self._check( x, tau)
        scale=tau*self.gamma
        return x / (1 + scale)
        
    def __call__(self, x):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        return self.gamma*0.5*x**2
    
    def _check(self, x, tau, name_name="'tau'"):
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
           
       if (np.size(self.gamma) > 1)  and (np.shape(self.gamma) != np.shape(x) ):
           raise Exception("gamma' must be either scalar or the same size as 'x'") 