#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-07-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np


class SquaredBathtub:
    """
    Computes the proximity operator of the squared w-insensitive loss function
    
            f(x) =  gamma *tau * (max{|x|-w,0} )^2   with w > 0
            
    When the input 'x' is an array, the output is computed element-wise.
    
    'tau*gamma' is the scale factor
    
     INPUTS
    ========
     x     - scalar or ND array
     w     - positive, scalar or ND array with the same size as 'x'
     gamma - positive, scalar or ND array with the same size as 'x' (default value: gamma=1)
     tau   - positive, scalar or ND array with the same size as 'x'   (default value: tau=1) (only for the method 'prox')
    
    =======
    Examples
    ========
 
     Evaluate the function  f:
    
     >>> SquaredBathtub(3)( 2 ) 
     0.0
     
      Compute the result element-wise for vector inputs :
          
     >>> SquaredBathtub(w=2, gamma=[1, 2, 3])( [-1, 3, -5] )  
     array([0. , 0.5, 4.5])
     
     Compute the proximity operator at a given point :
         
     >>> SquaredBathtub(2).prox(  [-2, 3, 4 ] )
     array([-2. ,  2.5,  3. ])
     
     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function 'gamma*f'
    
     >>> SquaredBathtub( 2, gamma=[1,2,3,3]).prox( [-2, 3, 4, 6 ] )
     array([-2.        ,  2.33333333,  2.5       ,  3.        ]
     
      This is equivalent to use the parameter 'tau' of the method 'prox' :
      
     >>> SquaredBathtub(2).prox( [-2, 3, 4, 6 ] , tau=[1,2,3, 3])
     array([-2.        ,  2.33333333,  2.5       ,  3.        ]
    """   

    def __init__(self, w, gamma=1):
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.size(w)>1 and (not isinstance(w, np.ndarray)):
            w=np.array(w)
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma

        if np.any( w <= 0 ):
            raise Exception("'w' (or all of its components if it is an array) must be strictly positive")
        self.w=w
        
    def prox(self, x, tau=1):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)):
            tau=np.array(tau)
        if np.size(x) <= 1:
            x=np.reshape(x, (-1))
            
        self._check(x, tau)
        gamma= self.gamma*tau
        w=self.w
        
        # preliminaries
        abs_x = np.abs(x)
        sign_x = np.sign(x)

        # 2nd branch
        p = (x + gamma * w * sign_x) / (gamma + 1)

        # 1st branch
        mask = abs_x <= w
        p[mask] = x[mask]

        return p
        
    def __call__(self, x, gamma=1):
        
        return gamma*0.5*( np.maximum( np.abs(x)-self.w, 0) )**2
    
        
    def _check(self, x, tau, name_name="'tau'"):
           if np.any( tau <= 0 ):
               raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
           
           if ( np.size(tau) > 1)  and (np.size(tau) != np.size(x) ):
               raise Exception(name_name+" must be either scalar or the same size as 'x'")
               
           if (np.size(self.gamma) > 1)  and (np.size(self.gamma) != np.size(x) ):
               raise Exception("gamma' must be either scalar or the same size as 'x'") 
             
           if (np.size(self.w) > 1)  and (np.size(self.w) != np.size(x) ):
               raise Exception("w' must be either scalar or the same size as 'x'") 
        