#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : 1.0 (06-07-2022)
 
 Author  : Mbaye DIONGUE

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
"""

import numpy as np

class SquaredHinge:
    """
    Computes the proximity operator of the Squared hinge loss function
    
            f(x) = gamma *tau* ( max{ x , 0 } ) ^ 2

    When the input 'x' is an array, the output is computed element-wise.
    
    'gamma*tau' is the scale factor
    
     INPUTS
    ========
     x     - scalar or ND array
     gamma - positive, scalar or ND array with the same size as 'x' (default value: gamma=1)
     tau   - positive, scalar or ND array with the same size as 'x'   (default value: tau=1) (only for the method 'prox')
    
    =======
    Examples
    ========
 
     Evaluate the function  f:
    
     >>> SquaredHinge(2)( 2 )
     4.0
     
      Compute the result element-wise for vector inputs :
          
     >>> SquaredHinge( gamma=[1, 2, 3])( [-1, 3, -5] )  
      array([0., 9., 0.])
     
     Compute the proximity operator at a given point :
         
     >>> SquaredHinge().prox(  [-2, 3, 4 ] )
     array([-2. ,  1.5,  2. ])
     
     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function 'gamma*f'
    
     >>> SquaredHinge(  gamma=[1,2,3,3]).prox( [-2, 3, 4, 6 ] )
     array([-2. ,  1. ,  1. ,  1.5])
     
      This is equivalent to use the parameter 'tau' of the method 'prox' :
      
     >>> SquaredHinge().prox( [-2, 3, 4, 6 ] , tau=[1,2,3, 3])
     array([-2. ,  1. ,  1. ,  1.5])
    """
    
    def __init__(self, gamma=1):
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
        

    def prox(self, x, tau=1):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(tau)>1 and (not isinstance(tau, np.ndarray)):
            tau=np.array(tau)
       
        self._check( x, tau)
        scale=tau*self.gamma
       
        scale= self.gamma*tau
        
        return np.minimum(x, x / (scale + 1))
        
    def __call__(self, x):
        
        return self.gamma*0.5*( np.maximum(x, 0) )**2
    
        
    def _check(self, x, tau, name_name="'tau'"):
        if np.any( tau <= 0 ):
            raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
        
        if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
            raise Exception(name_name+" must be either scalar or the same size as 'x'")
            
        if (np.size(self.gamma) > 1)  and (np.shape(self.gamma) != np.shape(x) ):
            raise Exception("gamma' must be either scalar or the same size as 'x'")
 
