#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Original Version : 1.0 (01-02-2019), 
 
 Author  : MOHAMED KERROUMI
 updated  (06-09-2022) by Mbaye DIONGUE
 Copyright (C) 2019
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).

"""

import numpy as np


class TanhActi :
    """
    This method computes the proximity operator of the 'Hyperbolic tangent activation'  function

                  / x*arctan(x)+(log(1-x^2)-x^2)/2   if |x| < 1
          f(x) = |
                 \ +Inf                              otherwise

        
    When the input 'x' is an array, the output is computed element-wise.
    
     INPUTS
    ========
     x     - scalar or ND array

    ========
    Examples
    ========
 
     Evaluate the function  f:
    
     >>> TanhActi()( 0.5 ) 
     array([0.00581204])
     
      Compute the result element-wise for vector inputs :
          
     >>> TanhActi()([-1, .3, .25, .5] )  
     array([       inf, 0.00070054, 0.00033394, 0.00581204])
     
     Compute the proximity operator at a given point :
         
     >>> TanhActi().prox(  [-2, .3, 4 ])
     array([-0.96402758,  0.29131261,  0.9993293 ])
    """

    def __init__(self):
        self.gamma=1

    def prox(self, x):
        return np.tanh(x)

    def __call__(self, x):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.isscalar(x):
            x = np.reshape(x, (-1))
        p = np.zeros(np.shape(x))
        mask= np.abs(x)<1
        p[mask] = (
            x[mask]* np.arctanh(x[mask]) + 0.5 * (np.log(1 - x[mask]** 2) - x[mask]** 2)
        )
        p[np.abs(x) >= 1] = np.inf
        return p
            

      
        