#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Version : (27-04-2017)
 
 Author  : Giovanni Chierchia
 updated (08-06-2022) by Mbaye Diongue

 Copyright (C) 2022
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).
 """

import numpy as np

class Thresholder :
    """
    This class computes the proximity operator of the function

                    / (gamma *  tau) * a * x   if x < 0
            f(x) = |  0                        if x = 0         with a <= b
                   \  (gamma *tau ) * b * x    otherwise
 
    When the input 'x' is an array, the output is computed element-wise.
    
    'gamma*tau' is the scale factor
    
     INPUTS
    ========
     x     - scalar or ND array
     a     - scalar or ND array with the same size as 'x'
     b     - scalar or ND array with the same size as 'x'
     gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
     tau   - positive, scalar or ND array with the same size as 'x' [default: tau=1] [only for the method 'prox']
     
     ========
     Examples
     ========
     
     Evaluate the 'direct' function:
         
     >>> Thresholder(-2, 2)(3)
     6
     >>> Thresholder(-2, 2, gamma=2)(3)
     12
     
     Compute the proximity operator at a given point :
         
     >>> Thresholder(-2, 2).prox( 3)
     1
     >>> Thresholder(-1, 2).prox([ -3., 1., 6.])
     array([-2.,  0.,  4.])
     
     Use a scale factor 'gamma'>0 to commute the proximity operator of gamma*f
     
     >>> Thresholder(-1, 2, gamma=2).prox([ -3., 1., 6.])
      array([-1.,  0.,  2.])
      
     Or equivalently with the parameter 'tau'
         
     >>> Thresholder(-1, 2).prox([ -3., 1., 6.], tau=2)
     array([-1.,  0.,  2.])
    """

    def __init__(self,  a=-1, b=1, gamma=1):
        if np.size(a)>1 and (not isinstance(a, np.ndarray)):
            a=np.array(a)
        if np.size(b)>1 and (not isinstance(b, np.ndarray)):
            b=np.array(b)
        if np.size(gamma)>1 and (not isinstance(gamma, np.ndarray)):
            gamma=np.array(gamma)
        if np.any( gamma <= 0 ):
            raise Exception("'gamma' (or all of its components if it is an array) must be strictly positive")
        self.gamma = gamma
        if np.any( a > b):
            raise Exception("'b' must be greater than 'a'")
        self.a = a 
        self.b = b

    def prox(self, x, tau=1):
        self._check_gamma( x, tau)
        if not isinstance(x, np.ndarray):
            x= np.array(x)
    
        scale=tau*self.gamma
        self._check_a_b(x)
        
        return np.minimum(0, x - self.a * scale) + np.maximum(0, x - self.b * scale)
    
        
    def __call__(self, x):
        if not isinstance(x, np.ndarray):
            x= np.array(x)
        self._check_a_b(x)
        
        return self.gamma*(self.a * np.minimum(0,x) + self.b * np.maximum(0,x))
        
        
        
    def _check_a_b(self, x):            
        if ( np.size(self.a)> 1) and (np.shape(self.a) != np.shape(x)):
            raise Exception("'a' must be either scalar or the same size as 'x'")

        if ( np.size(self.b) > 1) and (np.shape(self.b) != np.shape(x)):
            raise Exception("'b' must be either scalar or the same size as 'x'")
    
    def _check_gamma(self, x, tau, name_name="'tau'"):
      
       if np.any( tau <= 0 ):
           raise Exception(name_name+" (or all of its components if it is an array) must be strictly positive")
       
       if ( np.size(tau) > 1)  and (np.shape(tau) != np.shape(x) ):
           raise Exception(name_name+" must be either scalar or the same size as 'x'")
           
            
        