#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
 Original Version : 1.0 (01-02-2019), 
 
 Author  : MOHAMED KERROUMI
 updated  (06-09-2022) by Mbaye DIONGUE
 Copyright (C) 2019
 
 This file is part of the codes provided at http://proximity-operator.net
 
 By downloading and/or using any of these files, you implicitly agree to 
 all the terms of the license CeCill-B (available online).

"""

import numpy as np

class UnimodalSigmoid :
    """
    This class computes the proximity operator of the unimodl 'sigmoid activation' function
    
                  /(x+1/2)*log(x+1/2)+(1/2 -x)*log(1/2 -x)-(x^2+1/4)/2  if |x| < 1/2
           f(x) = | -1/4                                                |x| = 1/2
                  \ +Inf                                                otherwise

        
    When the input 'x' is an array, the output is computed element-wise.
    
     INPUTS
    ========
     x     - scalar or ND array

    ========
    Examples
    ========
 
     Evaluate the function  f:
    
     >>> UnimodalSigmoid()( 0.5 ) 
     array([-0.25])
     
      Compute the result element-wise for vector inputs :
          
     >>> UnimodalSigmoid()([-1, .3, .25, .5] )  
     array([        inf, -0.67040242, -0.71858514, -0.25      ])
     
     Compute the proximity operator at a given point :
         
     >>> UnimodalSigmoid().prox(  [-2, .3, 4 ])
     array([-0.38079708,  0.07444252,  0.48201379])
    """

    def __init__(self):
        self.gamma=1

    def prox(self, x):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        return 1 / (1 + np.exp(-x)) - 0.5

    def __call__(self, x):
        if np.size(x)>1 and (not isinstance(x, np.ndarray)):
            x=np.array(x)
        if np.size(x) <=1:
            x=np.reshape(x, (-1))
        p = np.zeros_like(1.0*x)
        mask= np.abs(x) < 0.5
        x_m=x[mask]
        p[mask] = (
            (x_m+ 0.5) * np.log(x_m+ 0.5)
            + (0.5 - x_m) * np.log(0.5 - x_m)
            - 0.5 * (x_m** 2 + 0.25)
        )
        p[np.abs(x) == 0.5] = -0.25
        p[np.abs(x) > 0.5] = np.inf
        return p
            
            

      
        